// This file was automatically generated from records.xml
#include "records.h"
#include <vector>
#include <iomanip>

namespace Swinder {

// ========== BackupRecord ==========

const unsigned BackupRecord::id = 0x0040;

class BackupRecord::Private
{
public:
    bool backupEnabled;
};

BackupRecord::BackupRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setBackupEnabled(false);
}

BackupRecord::~BackupRecord()
{
    delete d;
}

BackupRecord::BackupRecord( const BackupRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BackupRecord& BackupRecord::operator=( const BackupRecord& record )
{
    *d = *record.d;
    return *this;
}

bool BackupRecord::isBackupEnabled() const
{
    return d->backupEnabled;
}

void BackupRecord::setBackupEnabled(bool backupEnabled )
{
    d->backupEnabled = backupEnabled;
}

void BackupRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setBackupEnabled(readU16(data) != 0);
}

void BackupRecord::dump( std::ostream& out ) const
{
    out << "Backup" << std::endl;
    out << "      BackupEnabled : " << isBackupEnabled() << std::endl;
}

static Record* createBackupRecord(Workbook *book)
{
    return new BackupRecord(book);
}

// ========== EOFRecord ==========

const unsigned EOFRecord::id = 0x000A;

class EOFRecord::Private
{
public:
};

EOFRecord::EOFRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

EOFRecord::~EOFRecord()
{
    delete d;
}

EOFRecord::EOFRecord( const EOFRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

EOFRecord& EOFRecord::operator=( const EOFRecord& record )
{
    *d = *record.d;
    return *this;
}

void EOFRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void EOFRecord::dump( std::ostream& out ) const
{
    out << "EOF" << std::endl;
}

static Record* createEOFRecord(Workbook *book)
{
    return new EOFRecord(book);
}

// ========== BlankRecord ==========

const unsigned BlankRecord::id = 0x0201;

class BlankRecord::Private
{
public:
    unsigned column;
    unsigned row;
    unsigned xfIndex;
};

BlankRecord::BlankRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setColumn(0);
    setRow(0);
    setXfIndex(0);
}

BlankRecord::~BlankRecord()
{
    delete d;
}

BlankRecord::BlankRecord( const BlankRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BlankRecord& BlankRecord::operator=( const BlankRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned BlankRecord::column() const
{
    return d->column;
}

void BlankRecord::setColumn(unsigned column )
{
    d->column = column;
}

unsigned BlankRecord::row() const
{
    return d->row;
}

void BlankRecord::setRow(unsigned row )
{
    d->row = row;
}

unsigned BlankRecord::xfIndex() const
{
    return d->xfIndex;
}

void BlankRecord::setXfIndex(unsigned xfIndex )
{
    d->xfIndex = xfIndex;
}

void BlankRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 6) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setColumn(readU16(data + 2));
    setXfIndex(readU16(data + 4));
}

void BlankRecord::dump( std::ostream& out ) const
{
    out << "Blank" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "             Column : " << column() << std::endl;
    out << "            XfIndex : " << xfIndex() << std::endl;
}

static Record* createBlankRecord(Workbook *book)
{
    return new BlankRecord(book);
}

// ========== BoolErrRecord ==========

const unsigned BoolErrRecord::id = 0x0205;

class BoolErrRecord::Private
{
public:
    unsigned column;
    bool error;
    unsigned row;
    unsigned value;
    unsigned xfIndex;
};

BoolErrRecord::BoolErrRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setColumn(0);
    setError(false);
    setRow(0);
    setValue(0);
    setXfIndex(0);
}

BoolErrRecord::~BoolErrRecord()
{
    delete d;
}

BoolErrRecord::BoolErrRecord( const BoolErrRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BoolErrRecord& BoolErrRecord::operator=( const BoolErrRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned BoolErrRecord::column() const
{
    return d->column;
}

void BoolErrRecord::setColumn(unsigned column )
{
    d->column = column;
}

bool BoolErrRecord::isError() const
{
    return d->error;
}

void BoolErrRecord::setError(bool error )
{
    d->error = error;
}

unsigned BoolErrRecord::row() const
{
    return d->row;
}

void BoolErrRecord::setRow(unsigned row )
{
    d->row = row;
}

unsigned BoolErrRecord::value() const
{
    return d->value;
}

void BoolErrRecord::setValue(unsigned value )
{
    d->value = value;
}

unsigned BoolErrRecord::xfIndex() const
{
    return d->xfIndex;
}

void BoolErrRecord::setXfIndex(unsigned xfIndex )
{
    d->xfIndex = xfIndex;
}

Value BoolErrRecord::asValue() const
{
    return isError() ? errorAsValue(value()) : Value(value() ? true : false);
}

void BoolErrRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 8) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setColumn(readU16(data + 2));
    setXfIndex(readU16(data + 4));
    setValue(readU8(data + 6));
    setError(readU8(data + 7) != 0);
}

void BoolErrRecord::dump( std::ostream& out ) const
{
    out << "BoolErr" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "             Column : " << column() << std::endl;
    out << "            XfIndex : " << xfIndex() << std::endl;
    out << "              Value : " << value() << std::endl;
    out << "              Error : " << isError() << std::endl;
}

static Record* createBoolErrRecord(Workbook *book)
{
    return new BoolErrRecord(book);
}

// ========== StartBlockRecord ==========

const unsigned StartBlockRecord::id = 0x0852;

class StartBlockRecord::Private
{
public:
    unsigned frtHeaderOld;
    unsigned iObjectContext;
    unsigned iObjectInstance1;
    unsigned iObjectInstance2;
    unsigned iObjectKind;
};

StartBlockRecord::StartBlockRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFrtHeaderOld(0);
    setIObjectContext(0);
    setIObjectInstance1(0);
    setIObjectInstance2(0);
    setIObjectKind(0);
}

StartBlockRecord::~StartBlockRecord()
{
    delete d;
}

StartBlockRecord::StartBlockRecord( const StartBlockRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

StartBlockRecord& StartBlockRecord::operator=( const StartBlockRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned StartBlockRecord::frtHeaderOld() const
{
    return d->frtHeaderOld;
}

void StartBlockRecord::setFrtHeaderOld(unsigned frtHeaderOld )
{
    d->frtHeaderOld = frtHeaderOld;
}

unsigned StartBlockRecord::iObjectContext() const
{
    return d->iObjectContext;
}

void StartBlockRecord::setIObjectContext(unsigned iObjectContext )
{
    d->iObjectContext = iObjectContext;
}

unsigned StartBlockRecord::iObjectInstance1() const
{
    return d->iObjectInstance1;
}

void StartBlockRecord::setIObjectInstance1(unsigned iObjectInstance1 )
{
    d->iObjectInstance1 = iObjectInstance1;
}

unsigned StartBlockRecord::iObjectInstance2() const
{
    return d->iObjectInstance2;
}

void StartBlockRecord::setIObjectInstance2(unsigned iObjectInstance2 )
{
    d->iObjectInstance2 = iObjectInstance2;
}

unsigned StartBlockRecord::iObjectKind() const
{
    return d->iObjectKind;
}

void StartBlockRecord::setIObjectKind(unsigned iObjectKind )
{
    d->iObjectKind = iObjectKind;
}

void StartBlockRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 12) {
        setIsValid(false);
        return;
    }
    setFrtHeaderOld(readU32(data));
    setIObjectKind(readU16(data + 4));
    setIObjectContext(readU16(data + 6));
    setIObjectInstance1(readU16(data + 8));
    setIObjectInstance2(readU16(data + 10));
}

void StartBlockRecord::dump( std::ostream& out ) const
{
    out << "StartBlock" << std::endl;
    out << "       FrtHeaderOld : " << frtHeaderOld() << std::endl;
    out << "        IObjectKind : " << iObjectKind() << std::endl;
    out << "     IObjectContext : " << iObjectContext() << std::endl;
    out << "   IObjectInstance1 : " << iObjectInstance1() << std::endl;
    out << "   IObjectInstance2 : " << iObjectInstance2() << std::endl;
}

static Record* createStartBlockRecord(Workbook *book)
{
    return new StartBlockRecord(book);
}

// ========== EndBlockRecord ==========

const unsigned EndBlockRecord::id = 0x0853;

class EndBlockRecord::Private
{
public:
    unsigned frtHeaderOld;
    unsigned iObjectKind;
};

EndBlockRecord::EndBlockRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFrtHeaderOld(0);
    setIObjectKind(0);
}

EndBlockRecord::~EndBlockRecord()
{
    delete d;
}

EndBlockRecord::EndBlockRecord( const EndBlockRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

EndBlockRecord& EndBlockRecord::operator=( const EndBlockRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned EndBlockRecord::frtHeaderOld() const
{
    return d->frtHeaderOld;
}

void EndBlockRecord::setFrtHeaderOld(unsigned frtHeaderOld )
{
    d->frtHeaderOld = frtHeaderOld;
}

unsigned EndBlockRecord::iObjectKind() const
{
    return d->iObjectKind;
}

void EndBlockRecord::setIObjectKind(unsigned iObjectKind )
{
    d->iObjectKind = iObjectKind;
}

void EndBlockRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 6) {
        setIsValid(false);
        return;
    }
    setFrtHeaderOld(readU32(data));
    setIObjectKind(readU16(data + 4));
}

void EndBlockRecord::dump( std::ostream& out ) const
{
    out << "EndBlock" << std::endl;
    out << "       FrtHeaderOld : " << frtHeaderOld() << std::endl;
    out << "        IObjectKind : " << iObjectKind() << std::endl;
}

static Record* createEndBlockRecord(Workbook *book)
{
    return new EndBlockRecord(book);
}

// ========== ShapePropsStreamRecord ==========

const unsigned ShapePropsStreamRecord::id = 0x08a4;

class ShapePropsStreamRecord::Private
{
public:
    unsigned dwChecksum;
    unsigned grbitFrt;
    UString rgb;
    unsigned rt;
    unsigned unused;
    unsigned wObjContext;
};

ShapePropsStreamRecord::ShapePropsStreamRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setDwChecksum(0);
    setGrbitFrt(0);
    setRt(0);
    setUnused(0);
    setWObjContext(0);
}

ShapePropsStreamRecord::~ShapePropsStreamRecord()
{
    delete d;
}

ShapePropsStreamRecord::ShapePropsStreamRecord( const ShapePropsStreamRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ShapePropsStreamRecord& ShapePropsStreamRecord::operator=( const ShapePropsStreamRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned ShapePropsStreamRecord::dwChecksum() const
{
    return d->dwChecksum;
}

void ShapePropsStreamRecord::setDwChecksum(unsigned dwChecksum )
{
    d->dwChecksum = dwChecksum;
}

unsigned ShapePropsStreamRecord::grbitFrt() const
{
    return d->grbitFrt;
}

void ShapePropsStreamRecord::setGrbitFrt(unsigned grbitFrt )
{
    d->grbitFrt = grbitFrt;
}

UString ShapePropsStreamRecord::rgb() const
{
    return d->rgb;
}

void ShapePropsStreamRecord::setRgb(UString rgb )
{
    d->rgb = rgb;
}

unsigned ShapePropsStreamRecord::rt() const
{
    return d->rt;
}

void ShapePropsStreamRecord::setRt(unsigned rt )
{
    d->rt = rt;
}

unsigned ShapePropsStreamRecord::unused() const
{
    return d->unused;
}

void ShapePropsStreamRecord::setUnused(unsigned unused )
{
    d->unused = unused;
}

unsigned ShapePropsStreamRecord::wObjContext() const
{
    return d->wObjContext;
}

void ShapePropsStreamRecord::setWObjContext(unsigned wObjContext )
{
    d->wObjContext = wObjContext;
}

void ShapePropsStreamRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 24) {
        setIsValid(false);
        return;
    }
    setRt(readU16(data));
    setGrbitFrt(readU16(data + 2));
    setWObjContext(readU16(data + 12));
    setUnused(readU16(data + 14));
    setDwChecksum(readU32(data + 16));
    unsigned cb = readU32(data + 20);
    curOffset = 24;
    setRgb(readByteString(data + curOffset, cb, size - curOffset, &stringLengthError, &stringSize));
    if (stringLengthError) {
        setIsValid(false);
        return;
    }
    curOffset += stringSize;
}

void ShapePropsStreamRecord::dump( std::ostream& out ) const
{
    out << "ShapePropsStream" << std::endl;
    out << "                 Rt : " << rt() << std::endl;
    out << "           GrbitFrt : " << grbitFrt() << std::endl;
    out << "        WObjContext : " << wObjContext() << std::endl;
    out << "             Unused : " << unused() << std::endl;
    out << "         DwChecksum : " << dwChecksum() << std::endl;
    out << "                Rgb : " << rgb() << std::endl;
}

static Record* createShapePropsStreamRecord(Workbook *book)
{
    return new ShapePropsStreamRecord(book);
}

// ========== TextPropsStreamRecord ==========

const unsigned TextPropsStreamRecord::id = 0x08a5;

class TextPropsStreamRecord::Private
{
public:
    unsigned dwChecksum;
    unsigned grbitFrt;
    UString rgb;
    unsigned rt;
};

TextPropsStreamRecord::TextPropsStreamRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setDwChecksum(0);
    setGrbitFrt(0);
    setRt(0);
}

TextPropsStreamRecord::~TextPropsStreamRecord()
{
    delete d;
}

TextPropsStreamRecord::TextPropsStreamRecord( const TextPropsStreamRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

TextPropsStreamRecord& TextPropsStreamRecord::operator=( const TextPropsStreamRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned TextPropsStreamRecord::dwChecksum() const
{
    return d->dwChecksum;
}

void TextPropsStreamRecord::setDwChecksum(unsigned dwChecksum )
{
    d->dwChecksum = dwChecksum;
}

unsigned TextPropsStreamRecord::grbitFrt() const
{
    return d->grbitFrt;
}

void TextPropsStreamRecord::setGrbitFrt(unsigned grbitFrt )
{
    d->grbitFrt = grbitFrt;
}

UString TextPropsStreamRecord::rgb() const
{
    return d->rgb;
}

void TextPropsStreamRecord::setRgb(UString rgb )
{
    d->rgb = rgb;
}

unsigned TextPropsStreamRecord::rt() const
{
    return d->rt;
}

void TextPropsStreamRecord::setRt(unsigned rt )
{
    d->rt = rt;
}

void TextPropsStreamRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 20) {
        setIsValid(false);
        return;
    }
    setRt(readU16(data));
    setGrbitFrt(readU16(data + 2));
    setDwChecksum(readU32(data + 12));
    unsigned cb = readU32(data + 16);
    curOffset = 20;
    setRgb(readByteString(data + curOffset, cb, size - curOffset, &stringLengthError, &stringSize));
    if (stringLengthError) {
        setIsValid(false);
        return;
    }
    curOffset += stringSize;
}

void TextPropsStreamRecord::dump( std::ostream& out ) const
{
    out << "TextPropsStream" << std::endl;
    out << "                 Rt : " << rt() << std::endl;
    out << "           GrbitFrt : " << grbitFrt() << std::endl;
    out << "         DwChecksum : " << dwChecksum() << std::endl;
    out << "                Rgb : " << rgb() << std::endl;
}

static Record* createTextPropsStreamRecord(Workbook *book)
{
    return new TextPropsStreamRecord(book);
}

// ========== ChartFrtInfoRecord ==========

const unsigned ChartFrtInfoRecord::id = 0x089e;

class ChartFrtInfoRecord::Private
{
public:
};

ChartFrtInfoRecord::ChartFrtInfoRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ChartFrtInfoRecord::~ChartFrtInfoRecord()
{
    delete d;
}

ChartFrtInfoRecord::ChartFrtInfoRecord( const ChartFrtInfoRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ChartFrtInfoRecord& ChartFrtInfoRecord::operator=( const ChartFrtInfoRecord& record )
{
    *d = *record.d;
    return *this;
}

void ChartFrtInfoRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ChartFrtInfoRecord::dump( std::ostream& out ) const
{
    out << "ChartFrtInfo" << std::endl;
}

static Record* createChartFrtInfoRecord(Workbook *book)
{
    return new ChartFrtInfoRecord(book);
}

// ========== ChartLayout12ARecord ==========

const unsigned ChartLayout12ARecord::id = 0x08a7;

class ChartLayout12ARecord::Private
{
public:
    unsigned dwChecksum;
    double dx;
    double dy;
    bool fLayoutTargetInner;
    unsigned grbitFrt;
    unsigned rt;
    unsigned wHeightMode;
    unsigned wWidthMode;
    unsigned wXMode;
    unsigned wYMode;
    double x;
    int xBR;
    int xTL;
    double y;
    int yBR;
    int yTL;
};

ChartLayout12ARecord::ChartLayout12ARecord(Workbook *book)
    : Record(book), d(new Private)
{
    setDwChecksum(0);
    setFLayoutTargetInner(false);
    setGrbitFrt(0);
    setRt(0);
    setWHeightMode(0);
    setWWidthMode(0);
    setWXMode(0);
    setWYMode(0);
    setXBR(0);
    setXTL(0);
    setYBR(0);
    setYTL(0);
}

ChartLayout12ARecord::~ChartLayout12ARecord()
{
    delete d;
}

ChartLayout12ARecord::ChartLayout12ARecord( const ChartLayout12ARecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ChartLayout12ARecord& ChartLayout12ARecord::operator=( const ChartLayout12ARecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned ChartLayout12ARecord::dwChecksum() const
{
    return d->dwChecksum;
}

void ChartLayout12ARecord::setDwChecksum(unsigned dwChecksum )
{
    d->dwChecksum = dwChecksum;
}

double ChartLayout12ARecord::dx() const
{
    return d->dx;
}

void ChartLayout12ARecord::setDx(double dx )
{
    d->dx = dx;
}

double ChartLayout12ARecord::dy() const
{
    return d->dy;
}

void ChartLayout12ARecord::setDy(double dy )
{
    d->dy = dy;
}

bool ChartLayout12ARecord::isFLayoutTargetInner() const
{
    return d->fLayoutTargetInner;
}

void ChartLayout12ARecord::setFLayoutTargetInner(bool fLayoutTargetInner )
{
    d->fLayoutTargetInner = fLayoutTargetInner;
}

unsigned ChartLayout12ARecord::grbitFrt() const
{
    return d->grbitFrt;
}

void ChartLayout12ARecord::setGrbitFrt(unsigned grbitFrt )
{
    d->grbitFrt = grbitFrt;
}

unsigned ChartLayout12ARecord::rt() const
{
    return d->rt;
}

void ChartLayout12ARecord::setRt(unsigned rt )
{
    d->rt = rt;
}

unsigned ChartLayout12ARecord::wHeightMode() const
{
    return d->wHeightMode;
}

void ChartLayout12ARecord::setWHeightMode(unsigned wHeightMode )
{
    d->wHeightMode = wHeightMode;
}

unsigned ChartLayout12ARecord::wWidthMode() const
{
    return d->wWidthMode;
}

void ChartLayout12ARecord::setWWidthMode(unsigned wWidthMode )
{
    d->wWidthMode = wWidthMode;
}

unsigned ChartLayout12ARecord::wXMode() const
{
    return d->wXMode;
}

void ChartLayout12ARecord::setWXMode(unsigned wXMode )
{
    d->wXMode = wXMode;
}

unsigned ChartLayout12ARecord::wYMode() const
{
    return d->wYMode;
}

void ChartLayout12ARecord::setWYMode(unsigned wYMode )
{
    d->wYMode = wYMode;
}

double ChartLayout12ARecord::x() const
{
    return d->x;
}

void ChartLayout12ARecord::setX(double x )
{
    d->x = x;
}

int ChartLayout12ARecord::xBR() const
{
    return d->xBR;
}

void ChartLayout12ARecord::setXBR(int xBR )
{
    d->xBR = xBR;
}

int ChartLayout12ARecord::xTL() const
{
    return d->xTL;
}

void ChartLayout12ARecord::setXTL(int xTL )
{
    d->xTL = xTL;
}

double ChartLayout12ARecord::y() const
{
    return d->y;
}

void ChartLayout12ARecord::setY(double y )
{
    d->y = y;
}

int ChartLayout12ARecord::yBR() const
{
    return d->yBR;
}

void ChartLayout12ARecord::setYBR(int yBR )
{
    d->yBR = yBR;
}

int ChartLayout12ARecord::yTL() const
{
    return d->yTL;
}

void ChartLayout12ARecord::setYTL(int yTL )
{
    d->yTL = yTL;
}

void ChartLayout12ARecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 66) {
        setIsValid(false);
        return;
    }
    setRt(readU16(data));
    setGrbitFrt(readU16(data + 2));
    setDwChecksum(readU32(data + 12));
    setFLayoutTargetInner(((readU8(data + 16)) & 0x1) != 0);
    setXTL(readS16(data + 18));
    setYTL(readS16(data + 20));
    setXBR(readS16(data + 22));
    setYBR(readS16(data + 24));
    setWXMode(readU16(data + 26));
    setWYMode(readU16(data + 28));
    setWWidthMode(readU16(data + 30));
    setWHeightMode(readU16(data + 32));
    setX(readFloat64(data + 34));
    setY(readFloat64(data + 42));
    setDx(readFloat64(data + 50));
    setDy(readFloat64(data + 58));
}

void ChartLayout12ARecord::dump( std::ostream& out ) const
{
    out << "ChartLayout12A" << std::endl;
    out << "                 Rt : " << rt() << std::endl;
    out << "           GrbitFrt : " << grbitFrt() << std::endl;
    out << "         DwChecksum : " << dwChecksum() << std::endl;
    out << " FLayoutTargetInner : " << isFLayoutTargetInner() << std::endl;
    out << "                XTL : " << xTL() << std::endl;
    out << "                YTL : " << yTL() << std::endl;
    out << "                XBR : " << xBR() << std::endl;
    out << "                YBR : " << yBR() << std::endl;
    out << "             WXMode : " << wXMode() << std::endl;
    out << "             WYMode : " << wYMode() << std::endl;
    out << "         WWidthMode : " << wWidthMode() << std::endl;
    out << "        WHeightMode : " << wHeightMode() << std::endl;
    out << "                  X : " << x() << std::endl;
    out << "                  Y : " << y() << std::endl;
    out << "                 Dx : " << dx() << std::endl;
    out << "                 Dy : " << dy() << std::endl;
}

static Record* createChartLayout12ARecord(Workbook *book)
{
    return new ChartLayout12ARecord(book);
}

// ========== LeftMarginRecord ==========

const unsigned LeftMarginRecord::id = 0x0026;

class LeftMarginRecord::Private
{
public:
    double leftMargin;
};

LeftMarginRecord::LeftMarginRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

LeftMarginRecord::~LeftMarginRecord()
{
    delete d;
}

LeftMarginRecord::LeftMarginRecord( const LeftMarginRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LeftMarginRecord& LeftMarginRecord::operator=( const LeftMarginRecord& record )
{
    *d = *record.d;
    return *this;
}

double LeftMarginRecord::leftMargin() const
{
    return d->leftMargin;
}

void LeftMarginRecord::setLeftMargin(double leftMargin )
{
    d->leftMargin = leftMargin;
}

void LeftMarginRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 8) {
        setIsValid(false);
        return;
    }
    setLeftMargin(readFloat64(data));
}

void LeftMarginRecord::dump( std::ostream& out ) const
{
    out << "LeftMargin" << std::endl;
    out << "         LeftMargin : " << leftMargin() << std::endl;
}

static Record* createLeftMarginRecord(Workbook *book)
{
    return new LeftMarginRecord(book);
}

// ========== RightMarginRecord ==========

const unsigned RightMarginRecord::id = 0x0027;

class RightMarginRecord::Private
{
public:
    double rightMargin;
};

RightMarginRecord::RightMarginRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

RightMarginRecord::~RightMarginRecord()
{
    delete d;
}

RightMarginRecord::RightMarginRecord( const RightMarginRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

RightMarginRecord& RightMarginRecord::operator=( const RightMarginRecord& record )
{
    *d = *record.d;
    return *this;
}

double RightMarginRecord::rightMargin() const
{
    return d->rightMargin;
}

void RightMarginRecord::setRightMargin(double rightMargin )
{
    d->rightMargin = rightMargin;
}

void RightMarginRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 8) {
        setIsValid(false);
        return;
    }
    setRightMargin(readFloat64(data));
}

void RightMarginRecord::dump( std::ostream& out ) const
{
    out << "RightMargin" << std::endl;
    out << "        RightMargin : " << rightMargin() << std::endl;
}

static Record* createRightMarginRecord(Workbook *book)
{
    return new RightMarginRecord(book);
}

// ========== TopMarginRecord ==========

const unsigned TopMarginRecord::id = 0x0028;

class TopMarginRecord::Private
{
public:
    double topMargin;
};

TopMarginRecord::TopMarginRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

TopMarginRecord::~TopMarginRecord()
{
    delete d;
}

TopMarginRecord::TopMarginRecord( const TopMarginRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

TopMarginRecord& TopMarginRecord::operator=( const TopMarginRecord& record )
{
    *d = *record.d;
    return *this;
}

double TopMarginRecord::topMargin() const
{
    return d->topMargin;
}

void TopMarginRecord::setTopMargin(double topMargin )
{
    d->topMargin = topMargin;
}

void TopMarginRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 8) {
        setIsValid(false);
        return;
    }
    setTopMargin(readFloat64(data));
}

void TopMarginRecord::dump( std::ostream& out ) const
{
    out << "TopMargin" << std::endl;
    out << "          TopMargin : " << topMargin() << std::endl;
}

static Record* createTopMarginRecord(Workbook *book)
{
    return new TopMarginRecord(book);
}

// ========== BottomMarginRecord ==========

const unsigned BottomMarginRecord::id = 0x0029;

class BottomMarginRecord::Private
{
public:
    double bottomMargin;
};

BottomMarginRecord::BottomMarginRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BottomMarginRecord::~BottomMarginRecord()
{
    delete d;
}

BottomMarginRecord::BottomMarginRecord( const BottomMarginRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BottomMarginRecord& BottomMarginRecord::operator=( const BottomMarginRecord& record )
{
    *d = *record.d;
    return *this;
}

double BottomMarginRecord::bottomMargin() const
{
    return d->bottomMargin;
}

void BottomMarginRecord::setBottomMargin(double bottomMargin )
{
    d->bottomMargin = bottomMargin;
}

void BottomMarginRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 8) {
        setIsValid(false);
        return;
    }
    setBottomMargin(readFloat64(data));
}

void BottomMarginRecord::dump( std::ostream& out ) const
{
    out << "BottomMargin" << std::endl;
    out << "       BottomMargin : " << bottomMargin() << std::endl;
}

static Record* createBottomMarginRecord(Workbook *book)
{
    return new BottomMarginRecord(book);
}

// ========== BoundSheetRecord ==========

const unsigned BoundSheetRecord::id = 0x0085;

class BoundSheetRecord::Private
{
public:
    unsigned bofPosition;
    UString sheetName;
    SheetState sheetState;
    SheetType sheetType;
};

BoundSheetRecord::BoundSheetRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setBofPosition(0);
}

BoundSheetRecord::~BoundSheetRecord()
{
    delete d;
}

BoundSheetRecord::BoundSheetRecord( const BoundSheetRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BoundSheetRecord& BoundSheetRecord::operator=( const BoundSheetRecord& record )
{
    *d = *record.d;
    return *this;
}

UString BoundSheetRecord::sheetStateToString(SheetState sheetState)
{
    switch (sheetState) {
        case Visible: return UString("Visible");
        case Hidden: return UString("Hidden");
        case StrongHidden: return UString("StrongHidden");
        default: return UString("Unknown: ") + UString::from(sheetState);
    }
}

UString BoundSheetRecord::sheetTypeToString(SheetType sheetType)
{
    switch (sheetType) {
        case Worksheet: return UString("Worksheet");
        case Chart: return UString("Chart");
        case VBModule: return UString("VBModule");
        default: return UString("Unknown: ") + UString::from(sheetType);
    }
}

unsigned BoundSheetRecord::bofPosition() const
{
    return d->bofPosition;
}

void BoundSheetRecord::setBofPosition(unsigned bofPosition )
{
    d->bofPosition = bofPosition;
}

UString BoundSheetRecord::sheetName() const
{
    return d->sheetName;
}

void BoundSheetRecord::setSheetName(UString sheetName )
{
    d->sheetName = sheetName;
}

BoundSheetRecord::SheetState BoundSheetRecord::sheetState() const
{
    return d->sheetState;
}

void BoundSheetRecord::setSheetState(SheetState sheetState )
{
    d->sheetState = sheetState;
}

BoundSheetRecord::SheetType BoundSheetRecord::sheetType() const
{
    return d->sheetType;
}

void BoundSheetRecord::setSheetType(SheetType sheetType )
{
    d->sheetType = sheetType;
}

void BoundSheetRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 7) {
        setIsValid(false);
        return;
    }
    setBofPosition(readU32(data));
    setSheetState(static_cast<SheetState>(readU8(data + 4)));
    setSheetType(static_cast<SheetType>(readU8(data + 5)));
    unsigned sheetNameLength = readU8(data + 6);
    curOffset = 7;
    if (version() < Excel97) {
        setSheetName(readByteString(data + curOffset, sheetNameLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
    if (version() >= Excel97) {
        setSheetName(readUnicodeString(data + curOffset, sheetNameLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
}

void BoundSheetRecord::dump( std::ostream& out ) const
{
    out << "BoundSheet" << std::endl;
    out << "        BofPosition : " << bofPosition() << std::endl;
    out << "         SheetState : " << sheetStateToString(sheetState()) << std::endl;
    out << "          SheetType : " << sheetTypeToString(sheetType()) << std::endl;
    if (version() < Excel97) {
        out << "          SheetName : " << sheetName() << std::endl;
    }
    if (version() >= Excel97) {
        out << "          SheetName : " << sheetName() << std::endl;
    }
}

static Record* createBoundSheetRecord(Workbook *book)
{
    return new BoundSheetRecord(book);
}

// ========== CalcModeRecord ==========

const unsigned CalcModeRecord::id = 0x000D;

class CalcModeRecord::Private
{
public:
    CalcMode calcMode;
};

CalcModeRecord::CalcModeRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CalcModeRecord::~CalcModeRecord()
{
    delete d;
}

CalcModeRecord::CalcModeRecord( const CalcModeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CalcModeRecord& CalcModeRecord::operator=( const CalcModeRecord& record )
{
    *d = *record.d;
    return *this;
}

UString CalcModeRecord::calcModeToString(CalcMode calcMode)
{
    switch (calcMode) {
        case Manual: return UString("Manual");
        case Automatic: return UString("Automatic");
        case SemiAutomatic: return UString("SemiAutomatic");
        default: return UString("Unknown: ") + UString::from(calcMode);
    }
}

CalcModeRecord::CalcMode CalcModeRecord::calcMode() const
{
    return d->calcMode;
}

void CalcModeRecord::setCalcMode(CalcMode calcMode )
{
    d->calcMode = calcMode;
}

void CalcModeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setCalcMode(static_cast<CalcMode>(readS16(data)));
}

void CalcModeRecord::dump( std::ostream& out ) const
{
    out << "CalcMode" << std::endl;
    out << "           CalcMode : " << calcModeToString(calcMode()) << std::endl;
}

static Record* createCalcModeRecord(Workbook *book)
{
    return new CalcModeRecord(book);
}

// ========== ColInfoRecord ==========

const unsigned ColInfoRecord::id = 0x007D;

class ColInfoRecord::Private
{
public:
    bool collapsed;
    unsigned firstColumn;
    bool hidden;
    unsigned lastColumn;
    unsigned outlineLevel;
    unsigned width;
    unsigned xfIndex;
};

ColInfoRecord::ColInfoRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCollapsed(false);
    setFirstColumn(0);
    setHidden(false);
    setLastColumn(0);
    setOutlineLevel(0);
    setWidth(0);
    setXfIndex(0);
}

ColInfoRecord::~ColInfoRecord()
{
    delete d;
}

ColInfoRecord::ColInfoRecord( const ColInfoRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ColInfoRecord& ColInfoRecord::operator=( const ColInfoRecord& record )
{
    *d = *record.d;
    return *this;
}

bool ColInfoRecord::isCollapsed() const
{
    return d->collapsed;
}

void ColInfoRecord::setCollapsed(bool collapsed )
{
    d->collapsed = collapsed;
}

unsigned ColInfoRecord::firstColumn() const
{
    return d->firstColumn;
}

void ColInfoRecord::setFirstColumn(unsigned firstColumn )
{
    d->firstColumn = firstColumn;
}

bool ColInfoRecord::isHidden() const
{
    return d->hidden;
}

void ColInfoRecord::setHidden(bool hidden )
{
    d->hidden = hidden;
}

unsigned ColInfoRecord::lastColumn() const
{
    return d->lastColumn;
}

void ColInfoRecord::setLastColumn(unsigned lastColumn )
{
    d->lastColumn = lastColumn;
}

unsigned ColInfoRecord::outlineLevel() const
{
    return d->outlineLevel;
}

void ColInfoRecord::setOutlineLevel(unsigned outlineLevel )
{
    d->outlineLevel = outlineLevel;
}

unsigned ColInfoRecord::width() const
{
    return d->width;
}

void ColInfoRecord::setWidth(unsigned width )
{
    d->width = width;
}

unsigned ColInfoRecord::xfIndex() const
{
    return d->xfIndex;
}

void ColInfoRecord::setXfIndex(unsigned xfIndex )
{
    d->xfIndex = xfIndex;
}

void ColInfoRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 12) {
        setIsValid(false);
        return;
    }
    setFirstColumn(readU16(data));
    setLastColumn(readU16(data + 2));
    setWidth(readU16(data + 4));
    setXfIndex(readU16(data + 6));
    setHidden(((readU8(data + 8)) & 0x1) != 0);
    setOutlineLevel(((readU8(data + 9)) & 0x7));
    setCollapsed(((readU8(data + 9) >> 4) & 0x1) != 0);
}

void ColInfoRecord::dump( std::ostream& out ) const
{
    out << "ColInfo" << std::endl;
    out << "        FirstColumn : " << firstColumn() << std::endl;
    out << "         LastColumn : " << lastColumn() << std::endl;
    out << "              Width : " << width() << std::endl;
    out << "            XfIndex : " << xfIndex() << std::endl;
    out << "             Hidden : " << isHidden() << std::endl;
    out << "       OutlineLevel : " << outlineLevel() << std::endl;
    out << "          Collapsed : " << isCollapsed() << std::endl;
}

static Record* createColInfoRecord(Workbook *book)
{
    return new ColInfoRecord(book);
}

// ========== DataTableRecord ==========

const unsigned DataTableRecord::id = 0x0236;

class DataTableRecord::Private
{
public:
    bool alwaysCalc;
    unsigned firstColumn;
    unsigned firstRow;
    unsigned inputColumn1;
    unsigned inputColumn2;
    unsigned inputRow1;
    unsigned inputRow2;
    unsigned lastColumn;
    unsigned lastRow;
    bool rowColTable;
    bool rowInput;
};

DataTableRecord::DataTableRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setAlwaysCalc(false);
    setFirstColumn(0);
    setFirstRow(0);
    setInputColumn1(0);
    setInputColumn2(0);
    setInputRow1(0);
    setInputRow2(0);
    setLastColumn(0);
    setLastRow(0);
    setRowColTable(false);
    setRowInput(false);
}

DataTableRecord::~DataTableRecord()
{
    delete d;
}

DataTableRecord::DataTableRecord( const DataTableRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DataTableRecord& DataTableRecord::operator=( const DataTableRecord& record )
{
    *d = *record.d;
    return *this;
}

UString DataTableRecord::directionToString(Direction direction)
{
    switch (direction) {
        case InputRow: return UString("InputRow");
        case InputColumn: return UString("InputColumn");
        case Input2D: return UString("Input2D");
        default: return UString("Unknown: ") + UString::from(direction);
    }
}

bool DataTableRecord::isAlwaysCalc() const
{
    return d->alwaysCalc;
}

void DataTableRecord::setAlwaysCalc(bool alwaysCalc )
{
    d->alwaysCalc = alwaysCalc;
}

unsigned DataTableRecord::firstColumn() const
{
    return d->firstColumn;
}

void DataTableRecord::setFirstColumn(unsigned firstColumn )
{
    d->firstColumn = firstColumn;
}

unsigned DataTableRecord::firstRow() const
{
    return d->firstRow;
}

void DataTableRecord::setFirstRow(unsigned firstRow )
{
    d->firstRow = firstRow;
}

unsigned DataTableRecord::inputColumn1() const
{
    return d->inputColumn1;
}

void DataTableRecord::setInputColumn1(unsigned inputColumn1 )
{
    d->inputColumn1 = inputColumn1;
}

unsigned DataTableRecord::inputColumn2() const
{
    return d->inputColumn2;
}

void DataTableRecord::setInputColumn2(unsigned inputColumn2 )
{
    d->inputColumn2 = inputColumn2;
}

unsigned DataTableRecord::inputRow1() const
{
    return d->inputRow1;
}

void DataTableRecord::setInputRow1(unsigned inputRow1 )
{
    d->inputRow1 = inputRow1;
}

unsigned DataTableRecord::inputRow2() const
{
    return d->inputRow2;
}

void DataTableRecord::setInputRow2(unsigned inputRow2 )
{
    d->inputRow2 = inputRow2;
}

unsigned DataTableRecord::lastColumn() const
{
    return d->lastColumn;
}

void DataTableRecord::setLastColumn(unsigned lastColumn )
{
    d->lastColumn = lastColumn;
}

unsigned DataTableRecord::lastRow() const
{
    return d->lastRow;
}

void DataTableRecord::setLastRow(unsigned lastRow )
{
    d->lastRow = lastRow;
}

bool DataTableRecord::isRowColTable() const
{
    return d->rowColTable;
}

void DataTableRecord::setRowColTable(bool rowColTable )
{
    d->rowColTable = rowColTable;
}

bool DataTableRecord::isRowInput() const
{
    return d->rowInput;
}

void DataTableRecord::setRowInput(bool rowInput )
{
    d->rowInput = rowInput;
}

DataTableRecord::Direction DataTableRecord::direction() const
{
    return isRowColTable() ? Input2D : isRowInput() ? InputRow : InputColumn;
}

void DataTableRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 16) {
        setIsValid(false);
        return;
    }
    setFirstRow(readU16(data));
    setLastRow(readU16(data + 2));
    setFirstColumn(readU8(data + 4));
    setLastColumn(readU8(data + 5));
    setAlwaysCalc(((readU8(data + 6)) & 0x1) != 0);
    setRowInput(((readU8(data + 6) >> 2) & 0x1) != 0);
    setRowColTable(((readU8(data + 6) >> 3) & 0x1) != 0);
    setInputRow1(readU16(data + 8));
    setInputColumn1(readU16(data + 10));
    setInputRow2(readU16(data + 12));
    setInputColumn2(readU16(data + 14));
}

void DataTableRecord::dump( std::ostream& out ) const
{
    out << "DataTable" << std::endl;
    out << "           FirstRow : " << firstRow() << std::endl;
    out << "            LastRow : " << lastRow() << std::endl;
    out << "        FirstColumn : " << firstColumn() << std::endl;
    out << "         LastColumn : " << lastColumn() << std::endl;
    out << "         AlwaysCalc : " << isAlwaysCalc() << std::endl;
    out << "           RowInput : " << isRowInput() << std::endl;
    out << "        RowColTable : " << isRowColTable() << std::endl;
    out << "          InputRow1 : " << inputRow1() << std::endl;
    out << "       InputColumn1 : " << inputColumn1() << std::endl;
    out << "          InputRow2 : " << inputRow2() << std::endl;
    out << "       InputColumn2 : " << inputColumn2() << std::endl;
}

static Record* createDataTableRecord(Workbook *book)
{
    return new DataTableRecord(book);
}

// ========== DateModeRecord ==========

const unsigned DateModeRecord::id = 0x0022;

class DateModeRecord::Private
{
public:
    bool base1904;
};

DateModeRecord::DateModeRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setBase1904(false);
}

DateModeRecord::~DateModeRecord()
{
    delete d;
}

DateModeRecord::DateModeRecord( const DateModeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DateModeRecord& DateModeRecord::operator=( const DateModeRecord& record )
{
    *d = *record.d;
    return *this;
}

bool DateModeRecord::isBase1904() const
{
    return d->base1904;
}

void DateModeRecord::setBase1904(bool base1904 )
{
    d->base1904 = base1904;
}

void DateModeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setBase1904(readU16(data) != 0);
}

void DateModeRecord::dump( std::ostream& out ) const
{
    out << "DateMode" << std::endl;
    out << "           Base1904 : " << isBase1904() << std::endl;
}

static Record* createDateModeRecord(Workbook *book)
{
    return new DateModeRecord(book);
}

// ========== DimensionRecord ==========

const unsigned DimensionRecord::id = 0x0200;

class DimensionRecord::Private
{
public:
    unsigned firstColumn;
    unsigned firstRow;
    unsigned lastColumnPlus1;
    unsigned lastRowPlus1;
};

DimensionRecord::DimensionRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFirstColumn(0);
    setFirstRow(0);
    setLastColumnPlus1(0);
    setLastRowPlus1(0);
}

DimensionRecord::~DimensionRecord()
{
    delete d;
}

DimensionRecord::DimensionRecord( const DimensionRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DimensionRecord& DimensionRecord::operator=( const DimensionRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned DimensionRecord::firstColumn() const
{
    return d->firstColumn;
}

void DimensionRecord::setFirstColumn(unsigned firstColumn )
{
    d->firstColumn = firstColumn;
}

unsigned DimensionRecord::firstRow() const
{
    return d->firstRow;
}

void DimensionRecord::setFirstRow(unsigned firstRow )
{
    d->firstRow = firstRow;
}

unsigned DimensionRecord::lastColumnPlus1() const
{
    return d->lastColumnPlus1;
}

void DimensionRecord::setLastColumnPlus1(unsigned lastColumnPlus1 )
{
    d->lastColumnPlus1 = lastColumnPlus1;
}

unsigned DimensionRecord::lastRowPlus1() const
{
    return d->lastRowPlus1;
}

void DimensionRecord::setLastRowPlus1(unsigned lastRowPlus1 )
{
    d->lastRowPlus1 = lastRowPlus1;
}

unsigned DimensionRecord::lastRow() const
{
    return lastRowPlus1() == 0 ? 0 : lastRowPlus1() - 1;
}

unsigned DimensionRecord::lastColumn() const
{
    return lastColumnPlus1() == 0 ? 0 : lastColumnPlus1() - 1;
}

void DimensionRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    curOffset = 0;
    if (version() < Excel97) {
        if (size < curOffset + 4) {
            setIsValid(false);
            return;
        }
        setFirstRow(readU16(data + curOffset));
        setLastRowPlus1(readU16(data + curOffset + 2));
        curOffset += 4;
    }
    if (version() >= Excel97) {
        if (size < curOffset + 8) {
            setIsValid(false);
            return;
        }
        setFirstRow(readU32(data + curOffset));
        setLastRowPlus1(readU32(data + curOffset + 4));
        curOffset += 8;
    }
    if (size < curOffset + 6) {
        setIsValid(false);
        return;
    }
    setFirstColumn(readU16(data + curOffset));
    setLastColumnPlus1(readU16(data + curOffset + 2));
}

void DimensionRecord::dump( std::ostream& out ) const
{
    out << "Dimension" << std::endl;
    if (version() < Excel97) {
        out << "           FirstRow : " << firstRow() << std::endl;
        out << "       LastRowPlus1 : " << lastRowPlus1() << std::endl;
    }
    if (version() >= Excel97) {
        out << "           FirstRow : " << firstRow() << std::endl;
        out << "       LastRowPlus1 : " << lastRowPlus1() << std::endl;
    }
    out << "        FirstColumn : " << firstColumn() << std::endl;
    out << "    LastColumnPlus1 : " << lastColumnPlus1() << std::endl;
}

static Record* createDimensionRecord(Workbook *book)
{
    return new DimensionRecord(book);
}

// ========== ExternSheetRecord ==========

const unsigned ExternSheetRecord::id = 0x0017;

class ExternSheetRecord::Private
{
public:
    std::vector<unsigned> bookRef;
    std::vector<unsigned> firstSheetRef;
    std::vector<unsigned> lastSheetRef;
    unsigned refCount;
};

ExternSheetRecord::ExternSheetRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setRefCount(0);
}

ExternSheetRecord::~ExternSheetRecord()
{
    delete d;
}

ExternSheetRecord::ExternSheetRecord( const ExternSheetRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ExternSheetRecord& ExternSheetRecord::operator=( const ExternSheetRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned ExternSheetRecord::bookRef( unsigned index ) const
{
    return d->bookRef[index];
}

void ExternSheetRecord::setBookRef( unsigned index, unsigned bookRef )
{
    d->bookRef[index] = bookRef;
}

unsigned ExternSheetRecord::firstSheetRef( unsigned index ) const
{
    return d->firstSheetRef[index];
}

void ExternSheetRecord::setFirstSheetRef( unsigned index, unsigned firstSheetRef )
{
    d->firstSheetRef[index] = firstSheetRef;
}

unsigned ExternSheetRecord::lastSheetRef( unsigned index ) const
{
    return d->lastSheetRef[index];
}

void ExternSheetRecord::setLastSheetRef( unsigned index, unsigned lastSheetRef )
{
    d->lastSheetRef[index] = lastSheetRef;
}

unsigned ExternSheetRecord::refCount() const
{
    return d->refCount;
}

void ExternSheetRecord::setRefCount(unsigned refCount )
{
    d->refCount = refCount;
    d->bookRef.resize(refCount);
    d->firstSheetRef.resize(refCount);
    d->lastSheetRef.resize(refCount);
}

void ExternSheetRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    curOffset = 0;
    if (version() >= Excel97) {
        if (size < curOffset + 2) {
            setIsValid(false);
            return;
        }
        setRefCount(readU16(data + curOffset));
        curOffset += 2;
        for (unsigned i = 0, endi = refCount(); i < endi; ++i) {
            if (size < curOffset + 6) {
                setIsValid(false);
                return;
            }
            setBookRef(i, readU16(data + curOffset));
            setFirstSheetRef(i, readU16(data + curOffset + 2));
            setLastSheetRef(i, readU16(data + curOffset + 4));
            curOffset += 6;
        }
    }
}

void ExternSheetRecord::dump( std::ostream& out ) const
{
    out << "ExternSheet" << std::endl;
    if (version() >= Excel97) {
        out << "           RefCount : " << refCount() << std::endl;
        for (unsigned i = 0, endi = refCount(); i < endi; ++i) {
            out << "        BookRef " << std::setw(3) << i <<" : " << bookRef(i) << std::endl;
            out << "  FirstSheetRef " << std::setw(3) << i <<" : " << firstSheetRef(i) << std::endl;
            out << "   LastSheetRef " << std::setw(3) << i <<" : " << lastSheetRef(i) << std::endl;
        }
    }
}

static Record* createExternSheetRecord(Workbook *book)
{
    return new ExternSheetRecord(book);
}

// ========== FontRecord ==========

const unsigned FontRecord::id = 0x0031;

class FontRecord::Private
{
public:
    unsigned characterSet;
    unsigned colorIndex;
    bool condensed;
    Escapement escapement;
    bool extended;
    FontFamily fontFamily;
    UString fontName;
    unsigned fontWeight;
    unsigned height;
    bool italic;
    bool outline;
    bool shadow;
    bool strikeout;
    Underline underline;
};

FontRecord::FontRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCharacterSet(0);
    setColorIndex(0);
    setCondensed(false);
    setExtended(false);
    setFontWeight(0);
    setHeight(0);
    setItalic(false);
    setOutline(false);
    setShadow(false);
    setStrikeout(false);
}

FontRecord::~FontRecord()
{
    delete d;
}

FontRecord::FontRecord( const FontRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

FontRecord& FontRecord::operator=( const FontRecord& record )
{
    *d = *record.d;
    return *this;
}

UString FontRecord::escapementToString(Escapement escapement)
{
    switch (escapement) {
        case Normal: return UString("Normal");
        case Superscript: return UString("Superscript");
        case Subscript: return UString("Subscript");
        default: return UString("Unknown: ") + UString::from(escapement);
    }
}

UString FontRecord::underlineToString(Underline underline)
{
    switch (underline) {
        case None: return UString("None");
        case Single: return UString("Single");
        case Double: return UString("Double");
        case SingleAccounting: return UString("SingleAccounting");
        case DoubleAccounting: return UString("DoubleAccounting");
        default: return UString("Unknown: ") + UString::from(underline);
    }
}

UString FontRecord::fontFamilyToString(FontFamily fontFamily)
{
    switch (fontFamily) {
        case Unknown: return UString("Unknown");
        case Roman: return UString("Roman");
        case Swiss: return UString("Swiss");
        case Modern: return UString("Modern");
        case Script: return UString("Script");
        case Decorative: return UString("Decorative");
        default: return UString("Unknown: ") + UString::from(fontFamily);
    }
}

unsigned FontRecord::characterSet() const
{
    return d->characterSet;
}

void FontRecord::setCharacterSet(unsigned characterSet )
{
    d->characterSet = characterSet;
}

unsigned FontRecord::colorIndex() const
{
    return d->colorIndex;
}

void FontRecord::setColorIndex(unsigned colorIndex )
{
    d->colorIndex = colorIndex;
}

bool FontRecord::isCondensed() const
{
    return d->condensed;
}

void FontRecord::setCondensed(bool condensed )
{
    d->condensed = condensed;
}

FontRecord::Escapement FontRecord::escapement() const
{
    return d->escapement;
}

void FontRecord::setEscapement(Escapement escapement )
{
    d->escapement = escapement;
}

bool FontRecord::isExtended() const
{
    return d->extended;
}

void FontRecord::setExtended(bool extended )
{
    d->extended = extended;
}

FontRecord::FontFamily FontRecord::fontFamily() const
{
    return d->fontFamily;
}

void FontRecord::setFontFamily(FontFamily fontFamily )
{
    d->fontFamily = fontFamily;
}

UString FontRecord::fontName() const
{
    return d->fontName;
}

void FontRecord::setFontName(UString fontName )
{
    d->fontName = fontName;
}

unsigned FontRecord::fontWeight() const
{
    return d->fontWeight;
}

void FontRecord::setFontWeight(unsigned fontWeight )
{
    d->fontWeight = fontWeight;
}

unsigned FontRecord::height() const
{
    return d->height;
}

void FontRecord::setHeight(unsigned height )
{
    d->height = height;
}

bool FontRecord::isItalic() const
{
    return d->italic;
}

void FontRecord::setItalic(bool italic )
{
    d->italic = italic;
}

bool FontRecord::isOutline() const
{
    return d->outline;
}

void FontRecord::setOutline(bool outline )
{
    d->outline = outline;
}

bool FontRecord::isShadow() const
{
    return d->shadow;
}

void FontRecord::setShadow(bool shadow )
{
    d->shadow = shadow;
}

bool FontRecord::isStrikeout() const
{
    return d->strikeout;
}

void FontRecord::setStrikeout(bool strikeout )
{
    d->strikeout = strikeout;
}

FontRecord::Underline FontRecord::underline() const
{
    return d->underline;
}

void FontRecord::setUnderline(Underline underline )
{
    d->underline = underline;
}

void FontRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 15) {
        setIsValid(false);
        return;
    }
    setHeight(readU16(data));
    setItalic(((readU8(data + 2) >> 1) & 0x1) != 0);
    setStrikeout(((readU8(data + 2) >> 3) & 0x1) != 0);
    setOutline(((readU8(data + 2) >> 4) & 0x1) != 0);
    setShadow(((readU8(data + 2) >> 5) & 0x1) != 0);
    setCondensed(((readU8(data + 2) >> 6) & 0x1) != 0);
    setExtended(((readU8(data + 2) >> 7) & 0x1) != 0);
    setColorIndex(readU16(data + 4));
    setFontWeight(readU16(data + 6));
    setEscapement(static_cast<Escapement>(readU16(data + 8)));
    setUnderline(static_cast<Underline>(readU8(data + 10)));
    setFontFamily(static_cast<FontFamily>(readU8(data + 11)));
    setCharacterSet(readU8(data + 12));
    unsigned fontNameLength = readU8(data + 14);
    curOffset = 15;
    if (version() < Excel97) {
        setFontName(readByteString(data + curOffset, fontNameLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
    if (version() >= Excel97) {
        setFontName(readUnicodeString(data + curOffset, fontNameLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
}

void FontRecord::dump( std::ostream& out ) const
{
    out << "Font" << std::endl;
    out << "             Height : " << height() << std::endl;
    out << "             Italic : " << isItalic() << std::endl;
    out << "          Strikeout : " << isStrikeout() << std::endl;
    out << "            Outline : " << isOutline() << std::endl;
    out << "             Shadow : " << isShadow() << std::endl;
    out << "          Condensed : " << isCondensed() << std::endl;
    out << "           Extended : " << isExtended() << std::endl;
    out << "         ColorIndex : " << colorIndex() << std::endl;
    out << "         FontWeight : " << fontWeight() << std::endl;
    out << "         Escapement : " << escapementToString(escapement()) << std::endl;
    out << "          Underline : " << underlineToString(underline()) << std::endl;
    out << "         FontFamily : " << fontFamilyToString(fontFamily()) << std::endl;
    out << "       CharacterSet : " << characterSet() << std::endl;
    if (version() < Excel97) {
        out << "           FontName : " << fontName() << std::endl;
    }
    if (version() >= Excel97) {
        out << "           FontName : " << fontName() << std::endl;
    }
}

static Record* createFontRecord(Workbook *book)
{
    return new FontRecord(book);
}

// ========== HeaderRecord ==========

const unsigned HeaderRecord::id = 0x0014;

class HeaderRecord::Private
{
public:
    UString header;
};

HeaderRecord::HeaderRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

HeaderRecord::~HeaderRecord()
{
    delete d;
}

HeaderRecord::HeaderRecord( const HeaderRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

HeaderRecord& HeaderRecord::operator=( const HeaderRecord& record )
{
    *d = *record.d;
    return *this;
}

UString HeaderRecord::header() const
{
    return d->header;
}

void HeaderRecord::setHeader(UString header )
{
    d->header = header;
}

void HeaderRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    curOffset = 0;
    if (version() < Excel97) {
        if (size < curOffset + 1) {
            setIsValid(false);
            return;
        }
        unsigned headerLength = readU8(data + curOffset);
        curOffset += 1;
        setHeader(readByteString(data + curOffset, headerLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
    if (version() >= Excel97) {
        if (size < curOffset + 2) {
            setIsValid(false);
            return;
        }
        unsigned headerLength = readU16(data + curOffset);
        curOffset += 2;
        setHeader(readUnicodeString(data + curOffset, headerLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
}

void HeaderRecord::dump( std::ostream& out ) const
{
    out << "Header" << std::endl;
    if (version() < Excel97) {
        out << "             Header : " << header() << std::endl;
    }
    if (version() >= Excel97) {
        out << "             Header : " << header() << std::endl;
    }
}

static Record* createHeaderRecord(Workbook *book)
{
    return new HeaderRecord(book);
}

// ========== FooterRecord ==========

const unsigned FooterRecord::id = 0x0015;

class FooterRecord::Private
{
public:
    UString footer;
};

FooterRecord::FooterRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

FooterRecord::~FooterRecord()
{
    delete d;
}

FooterRecord::FooterRecord( const FooterRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

FooterRecord& FooterRecord::operator=( const FooterRecord& record )
{
    *d = *record.d;
    return *this;
}

UString FooterRecord::footer() const
{
    return d->footer;
}

void FooterRecord::setFooter(UString footer )
{
    d->footer = footer;
}

void FooterRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    curOffset = 0;
    if (version() < Excel97) {
        if (size < curOffset + 1) {
            setIsValid(false);
            return;
        }
        unsigned footerLength = readU8(data + curOffset);
        curOffset += 1;
        setFooter(readByteString(data + curOffset, footerLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
    if (version() >= Excel97) {
        if (size < curOffset + 2) {
            setIsValid(false);
            return;
        }
        unsigned footerLength = readU16(data + curOffset);
        curOffset += 2;
        setFooter(readUnicodeString(data + curOffset, footerLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
}

void FooterRecord::dump( std::ostream& out ) const
{
    out << "Footer" << std::endl;
    if (version() < Excel97) {
        out << "             Footer : " << footer() << std::endl;
    }
    if (version() >= Excel97) {
        out << "             Footer : " << footer() << std::endl;
    }
}

static Record* createFooterRecord(Workbook *book)
{
    return new FooterRecord(book);
}

// ========== FormatRecord ==========

const unsigned FormatRecord::id = 0x041E;

class FormatRecord::Private
{
public:
    UString formatString;
    unsigned index;
};

FormatRecord::FormatRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setIndex(0);
}

FormatRecord::~FormatRecord()
{
    delete d;
}

FormatRecord::FormatRecord( const FormatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

FormatRecord& FormatRecord::operator=( const FormatRecord& record )
{
    *d = *record.d;
    return *this;
}

UString FormatRecord::formatString() const
{
    return d->formatString;
}

void FormatRecord::setFormatString(UString formatString )
{
    d->formatString = formatString;
}

unsigned FormatRecord::index() const
{
    return d->index;
}

void FormatRecord::setIndex(unsigned index )
{
    d->index = index;
}

void FormatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setIndex(readU16(data));
    curOffset = 2;
    if (version() < Excel97) {
        if (size < curOffset + 1) {
            setIsValid(false);
            return;
        }
        unsigned formatStringLength = readU8(data + curOffset);
        curOffset += 1;
        setFormatString(readByteString(data + curOffset, formatStringLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
    if (version() >= Excel97) {
        if (size < curOffset + 2) {
            setIsValid(false);
            return;
        }
        unsigned formatStringLength = readU16(data + curOffset);
        curOffset += 2;
        setFormatString(readUnicodeString(data + curOffset, formatStringLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
}

void FormatRecord::dump( std::ostream& out ) const
{
    out << "Format" << std::endl;
    out << "              Index : " << index() << std::endl;
    if (version() < Excel97) {
        out << "       FormatString : " << formatString() << std::endl;
    }
    if (version() >= Excel97) {
        out << "       FormatString : " << formatString() << std::endl;
    }
}

static Record* createFormatRecord(Workbook *book)
{
    return new FormatRecord(book);
}

// ========== LabelRecord ==========

const unsigned LabelRecord::id = 0x0204;

class LabelRecord::Private
{
public:
    unsigned column;
    UString label;
    unsigned row;
    unsigned xfIndex;
};

LabelRecord::LabelRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setColumn(0);
    setRow(0);
    setXfIndex(0);
}

LabelRecord::~LabelRecord()
{
    delete d;
}

LabelRecord::LabelRecord( const LabelRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LabelRecord& LabelRecord::operator=( const LabelRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned LabelRecord::column() const
{
    return d->column;
}

void LabelRecord::setColumn(unsigned column )
{
    d->column = column;
}

UString LabelRecord::label() const
{
    return d->label;
}

void LabelRecord::setLabel(UString label )
{
    d->label = label;
}

unsigned LabelRecord::row() const
{
    return d->row;
}

void LabelRecord::setRow(unsigned row )
{
    d->row = row;
}

unsigned LabelRecord::xfIndex() const
{
    return d->xfIndex;
}

void LabelRecord::setXfIndex(unsigned xfIndex )
{
    d->xfIndex = xfIndex;
}

void LabelRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 8) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setColumn(readU16(data + 2));
    setXfIndex(readU16(data + 4));
    unsigned labelLength = readU16(data + 6);
    curOffset = 8;
    if (version() < Excel97) {
        setLabel(readByteString(data + curOffset, labelLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
    if (version() >= Excel97) {
        setLabel(readUnicodeString(data + curOffset, labelLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
}

void LabelRecord::dump( std::ostream& out ) const
{
    out << "Label" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "             Column : " << column() << std::endl;
    out << "            XfIndex : " << xfIndex() << std::endl;
    if (version() < Excel97) {
        out << "              Label : " << label() << std::endl;
    }
    if (version() >= Excel97) {
        out << "              Label : " << label() << std::endl;
    }
}

static Record* createLabelRecord(Workbook *book)
{
    return new LabelRecord(book);
}

// ========== LabelSSTRecord ==========

const unsigned LabelSSTRecord::id = 0x00FD;

class LabelSSTRecord::Private
{
public:
    unsigned column;
    unsigned row;
    unsigned sstIndex;
    unsigned xfIndex;
};

LabelSSTRecord::LabelSSTRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setColumn(0);
    setRow(0);
    setSstIndex(0);
    setXfIndex(0);
}

LabelSSTRecord::~LabelSSTRecord()
{
    delete d;
}

LabelSSTRecord::LabelSSTRecord( const LabelSSTRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LabelSSTRecord& LabelSSTRecord::operator=( const LabelSSTRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned LabelSSTRecord::column() const
{
    return d->column;
}

void LabelSSTRecord::setColumn(unsigned column )
{
    d->column = column;
}

unsigned LabelSSTRecord::row() const
{
    return d->row;
}

void LabelSSTRecord::setRow(unsigned row )
{
    d->row = row;
}

unsigned LabelSSTRecord::sstIndex() const
{
    return d->sstIndex;
}

void LabelSSTRecord::setSstIndex(unsigned sstIndex )
{
    d->sstIndex = sstIndex;
}

unsigned LabelSSTRecord::xfIndex() const
{
    return d->xfIndex;
}

void LabelSSTRecord::setXfIndex(unsigned xfIndex )
{
    d->xfIndex = xfIndex;
}

void LabelSSTRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 10) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setColumn(readU16(data + 2));
    setXfIndex(readU16(data + 4));
    setSstIndex(readU32(data + 6));
}

void LabelSSTRecord::dump( std::ostream& out ) const
{
    out << "LabelSST" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "             Column : " << column() << std::endl;
    out << "            XfIndex : " << xfIndex() << std::endl;
    out << "           SstIndex : " << sstIndex() << std::endl;
}

static Record* createLabelSSTRecord(Workbook *book)
{
    return new LabelSSTRecord(book);
}

// ========== MergedCellsRecord ==========

const unsigned MergedCellsRecord::id = 0x00E5;

class MergedCellsRecord::Private
{
public:
    unsigned count;
    std::vector<unsigned> firstColumn;
    std::vector<unsigned> firstRow;
    std::vector<unsigned> lastColumn;
    std::vector<unsigned> lastRow;
};

MergedCellsRecord::MergedCellsRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCount(0);
}

MergedCellsRecord::~MergedCellsRecord()
{
    delete d;
}

MergedCellsRecord::MergedCellsRecord( const MergedCellsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

MergedCellsRecord& MergedCellsRecord::operator=( const MergedCellsRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned MergedCellsRecord::count() const
{
    return d->count;
}

void MergedCellsRecord::setCount(unsigned count )
{
    d->count = count;
    d->firstRow.resize(count);
    d->lastRow.resize(count);
    d->firstColumn.resize(count);
    d->lastColumn.resize(count);
}

unsigned MergedCellsRecord::firstColumn( unsigned index ) const
{
    return d->firstColumn[index];
}

void MergedCellsRecord::setFirstColumn( unsigned index, unsigned firstColumn )
{
    d->firstColumn[index] = firstColumn;
}

unsigned MergedCellsRecord::firstRow( unsigned index ) const
{
    return d->firstRow[index];
}

void MergedCellsRecord::setFirstRow( unsigned index, unsigned firstRow )
{
    d->firstRow[index] = firstRow;
}

unsigned MergedCellsRecord::lastColumn( unsigned index ) const
{
    return d->lastColumn[index];
}

void MergedCellsRecord::setLastColumn( unsigned index, unsigned lastColumn )
{
    d->lastColumn[index] = lastColumn;
}

unsigned MergedCellsRecord::lastRow( unsigned index ) const
{
    return d->lastRow[index];
}

void MergedCellsRecord::setLastRow( unsigned index, unsigned lastRow )
{
    d->lastRow[index] = lastRow;
}

void MergedCellsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setCount(readU16(data));
    curOffset = 2;
    for (unsigned i = 0, endi = count(); i < endi; ++i) {
        if (size < curOffset + 8) {
            setIsValid(false);
            return;
        }
        setFirstRow(i, readU16(data + curOffset));
        setLastRow(i, readU16(data + curOffset + 2));
        setFirstColumn(i, readU16(data + curOffset + 4));
        setLastColumn(i, readU16(data + curOffset + 6));
        curOffset += 8;
    }
}

void MergedCellsRecord::dump( std::ostream& out ) const
{
    out << "MergedCells" << std::endl;
    out << "              Count : " << count() << std::endl;
    for (unsigned i = 0, endi = count(); i < endi; ++i) {
        out << "       FirstRow " << std::setw(3) << i <<" : " << firstRow(i) << std::endl;
        out << "        LastRow " << std::setw(3) << i <<" : " << lastRow(i) << std::endl;
        out << "    FirstColumn " << std::setw(3) << i <<" : " << firstColumn(i) << std::endl;
        out << "     LastColumn " << std::setw(3) << i <<" : " << lastColumn(i) << std::endl;
    }
}

static Record* createMergedCellsRecord(Workbook *book)
{
    return new MergedCellsRecord(book);
}

// ========== MulBlankRecord ==========

const unsigned MulBlankRecord::id = 0x00BE;

class MulBlankRecord::Private
{
public:
    unsigned firstColumn;
    unsigned lastColumn;
    unsigned row;
    std::vector<unsigned> xfIndex;
};

MulBlankRecord::MulBlankRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFirstColumn(0);
    setLastColumn(0);
    setRow(0);
}

MulBlankRecord::~MulBlankRecord()
{
    delete d;
}

MulBlankRecord::MulBlankRecord( const MulBlankRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

MulBlankRecord& MulBlankRecord::operator=( const MulBlankRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned MulBlankRecord::firstColumn() const
{
    return d->firstColumn;
}

void MulBlankRecord::setFirstColumn(unsigned firstColumn )
{
    d->firstColumn = firstColumn;
}

unsigned MulBlankRecord::lastColumn() const
{
    return d->lastColumn;
}

void MulBlankRecord::setLastColumn(unsigned lastColumn )
{
    d->lastColumn = lastColumn;
}

unsigned MulBlankRecord::row() const
{
    return d->row;
}

void MulBlankRecord::setRow(unsigned row )
{
    d->row = row;
}

unsigned MulBlankRecord::xfIndex( unsigned index ) const
{
    return d->xfIndex[index];
}

void MulBlankRecord::setXfIndex( unsigned index, unsigned xfIndex )
{
    d->xfIndex[index] = xfIndex;
}

void MulBlankRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    if (size < 4) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setFirstColumn(readU16(data + 2));
    curOffset = 4;
    d->xfIndex.resize((size-6)/2);
    for (unsigned i = 0, endi = (size-6)/2; i < endi; ++i) {
        if (size < curOffset + 2) {
            setIsValid(false);
            return;
        }
        setXfIndex(i, readU16(data + curOffset));
        curOffset += 2;
    }
    if (size < curOffset + 2) {
        setIsValid(false);
        return;
    }
    setLastColumn(readU16(data + curOffset));
}

void MulBlankRecord::dump( std::ostream& out ) const
{
    out << "MulBlank" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "        FirstColumn : " << firstColumn() << std::endl;
    for (unsigned i = 0, endi = d->xfIndex.size(); i < endi; ++i) {
        out << "        XfIndex " << std::setw(3) << i <<" : " << xfIndex(i) << std::endl;
    }
    out << "         LastColumn : " << lastColumn() << std::endl;
}

static Record* createMulBlankRecord(Workbook *book)
{
    return new MulBlankRecord(book);
}

// ========== NumberRecord ==========

const unsigned NumberRecord::id = 0x0203;

class NumberRecord::Private
{
public:
    unsigned column;
    double number;
    unsigned row;
    unsigned xfIndex;
};

NumberRecord::NumberRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setColumn(0);
    setRow(0);
    setXfIndex(0);
}

NumberRecord::~NumberRecord()
{
    delete d;
}

NumberRecord::NumberRecord( const NumberRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

NumberRecord& NumberRecord::operator=( const NumberRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned NumberRecord::column() const
{
    return d->column;
}

void NumberRecord::setColumn(unsigned column )
{
    d->column = column;
}

double NumberRecord::number() const
{
    return d->number;
}

void NumberRecord::setNumber(double number )
{
    d->number = number;
}

unsigned NumberRecord::row() const
{
    return d->row;
}

void NumberRecord::setRow(unsigned row )
{
    d->row = row;
}

unsigned NumberRecord::xfIndex() const
{
    return d->xfIndex;
}

void NumberRecord::setXfIndex(unsigned xfIndex )
{
    d->xfIndex = xfIndex;
}

void NumberRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 14) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setColumn(readU16(data + 2));
    setXfIndex(readU16(data + 4));
    setNumber(readFloat64(data + 6));
}

void NumberRecord::dump( std::ostream& out ) const
{
    out << "Number" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "             Column : " << column() << std::endl;
    out << "            XfIndex : " << xfIndex() << std::endl;
    out << "             Number : " << number() << std::endl;
}

static Record* createNumberRecord(Workbook *book)
{
    return new NumberRecord(book);
}

// ========== PaletteRecord ==========

const unsigned PaletteRecord::id = 0x0092;

class PaletteRecord::Private
{
public:
    std::vector<unsigned> blue;
    unsigned count;
    std::vector<unsigned> green;
    std::vector<unsigned> red;
};

PaletteRecord::PaletteRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCount(0);
}

PaletteRecord::~PaletteRecord()
{
    delete d;
}

PaletteRecord::PaletteRecord( const PaletteRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PaletteRecord& PaletteRecord::operator=( const PaletteRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned PaletteRecord::blue( unsigned index ) const
{
    return d->blue[index];
}

void PaletteRecord::setBlue( unsigned index, unsigned blue )
{
    d->blue[index] = blue;
}

unsigned PaletteRecord::count() const
{
    return d->count;
}

void PaletteRecord::setCount(unsigned count )
{
    d->count = count;
    d->red.resize(count);
    d->green.resize(count);
    d->blue.resize(count);
}

unsigned PaletteRecord::green( unsigned index ) const
{
    return d->green[index];
}

void PaletteRecord::setGreen( unsigned index, unsigned green )
{
    d->green[index] = green;
}

unsigned PaletteRecord::red( unsigned index ) const
{
    return d->red[index];
}

void PaletteRecord::setRed( unsigned index, unsigned red )
{
    d->red[index] = red;
}

void PaletteRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setCount(readU16(data));
    curOffset = 2;
    for (unsigned i = 0, endi = count(); i < endi; ++i) {
        if (size < curOffset + 4) {
            setIsValid(false);
            return;
        }
        setRed(i, readU8(data + curOffset));
        setGreen(i, readU8(data + curOffset + 1));
        setBlue(i, readU8(data + curOffset + 2));
        curOffset += 4;
    }
}

void PaletteRecord::dump( std::ostream& out ) const
{
    out << "Palette" << std::endl;
    out << "              Count : " << count() << std::endl;
    for (unsigned i = 0, endi = count(); i < endi; ++i) {
        out << "            Red " << std::setw(3) << i <<" : " << red(i) << std::endl;
        out << "          Green " << std::setw(3) << i <<" : " << green(i) << std::endl;
        out << "           Blue " << std::setw(3) << i <<" : " << blue(i) << std::endl;
    }
}

static Record* createPaletteRecord(Workbook *book)
{
    return new PaletteRecord(book);
}

// ========== RowRecord ==========

const unsigned RowRecord::id = 0x0208;

class RowRecord::Private
{
public:
    bool collapsed;
    unsigned firstColumn;
    unsigned height;
    bool hidden;
    unsigned lastColumnPlus1;
    unsigned outlineLevel;
    unsigned row;
    unsigned xfIndex;
};

RowRecord::RowRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCollapsed(false);
    setFirstColumn(0);
    setHeight(0);
    setHidden(false);
    setLastColumnPlus1(0);
    setOutlineLevel(0);
    setRow(0);
    setXfIndex(0);
}

RowRecord::~RowRecord()
{
    delete d;
}

RowRecord::RowRecord( const RowRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

RowRecord& RowRecord::operator=( const RowRecord& record )
{
    *d = *record.d;
    return *this;
}

bool RowRecord::isCollapsed() const
{
    return d->collapsed;
}

void RowRecord::setCollapsed(bool collapsed )
{
    d->collapsed = collapsed;
}

unsigned RowRecord::firstColumn() const
{
    return d->firstColumn;
}

void RowRecord::setFirstColumn(unsigned firstColumn )
{
    d->firstColumn = firstColumn;
}

unsigned RowRecord::height() const
{
    return d->height;
}

void RowRecord::setHeight(unsigned height )
{
    d->height = height;
}

bool RowRecord::isHidden() const
{
    return d->hidden;
}

void RowRecord::setHidden(bool hidden )
{
    d->hidden = hidden;
}

unsigned RowRecord::lastColumnPlus1() const
{
    return d->lastColumnPlus1;
}

void RowRecord::setLastColumnPlus1(unsigned lastColumnPlus1 )
{
    d->lastColumnPlus1 = lastColumnPlus1;
}

unsigned RowRecord::outlineLevel() const
{
    return d->outlineLevel;
}

void RowRecord::setOutlineLevel(unsigned outlineLevel )
{
    d->outlineLevel = outlineLevel;
}

unsigned RowRecord::row() const
{
    return d->row;
}

void RowRecord::setRow(unsigned row )
{
    d->row = row;
}

unsigned RowRecord::xfIndex() const
{
    return d->xfIndex;
}

void RowRecord::setXfIndex(unsigned xfIndex )
{
    d->xfIndex = xfIndex;
}

void RowRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 16) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setFirstColumn(readU16(data + 2));
    setLastColumnPlus1(readU16(data + 4));
    setHeight(readU16(data + 6));
    setOutlineLevel(((readU8(data + 12)) & 0x7));
    setCollapsed(((readU8(data + 12) >> 4) & 0x1) != 0);
    setHidden(((readU8(data + 12) >> 5) & 0x1) != 0);
    setXfIndex(((readU16(data + 14)) & 0xfff));
}

void RowRecord::dump( std::ostream& out ) const
{
    out << "Row" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "        FirstColumn : " << firstColumn() << std::endl;
    out << "    LastColumnPlus1 : " << lastColumnPlus1() << std::endl;
    out << "             Height : " << height() << std::endl;
    out << "       OutlineLevel : " << outlineLevel() << std::endl;
    out << "          Collapsed : " << isCollapsed() << std::endl;
    out << "             Hidden : " << isHidden() << std::endl;
    out << "            XfIndex : " << xfIndex() << std::endl;
}

static Record* createRowRecord(Workbook *book)
{
    return new RowRecord(book);
}

// ========== StringRecord ==========

const unsigned StringRecord::id = 0x0207;

class StringRecord::Private
{
public:
    UString ustring;
};

StringRecord::StringRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

StringRecord::~StringRecord()
{
    delete d;
}

StringRecord::StringRecord( const StringRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

StringRecord& StringRecord::operator=( const StringRecord& record )
{
    *d = *record.d;
    return *this;
}

UString StringRecord::ustring() const
{
    return d->ustring;
}

void StringRecord::setUstring(UString ustring )
{
    d->ustring = ustring;
}

Value StringRecord::value() const
{
    return Value(ustring());
}

void StringRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 2) {
        setIsValid(false);
        return;
    }
    unsigned stringLength = readU16(data);
    curOffset = 2;
    if (version() < Excel97) {
        setUstring(readByteString(data + curOffset, stringLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
    if (version() >= Excel97) {
        setUstring(readUnicodeString(data + curOffset, stringLength, size - curOffset, &stringLengthError, &stringSize));
        if (stringLengthError) {
            setIsValid(false);
            return;
        }
        curOffset += stringSize;
    }
}

void StringRecord::dump( std::ostream& out ) const
{
    out << "String" << std::endl;
    if (version() < Excel97) {
        out << "            Ustring : " << ustring() << std::endl;
    }
    if (version() >= Excel97) {
        out << "            Ustring : " << ustring() << std::endl;
    }
}

static Record* createStringRecord(Workbook *book)
{
    return new StringRecord(book);
}

// ========== NoteRecord ==========

const unsigned NoteRecord::id = 0x1C;

class NoteRecord::Private
{
public:
    unsigned column;
    unsigned idObj;
    unsigned opts;
    unsigned row;
};

NoteRecord::NoteRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setColumn(0);
    setIdObj(0);
    setOpts(0);
    setRow(0);
}

NoteRecord::~NoteRecord()
{
    delete d;
}

NoteRecord::NoteRecord( const NoteRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

NoteRecord& NoteRecord::operator=( const NoteRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned NoteRecord::column() const
{
    return d->column;
}

void NoteRecord::setColumn(unsigned column )
{
    d->column = column;
}

unsigned NoteRecord::idObj() const
{
    return d->idObj;
}

void NoteRecord::setIdObj(unsigned idObj )
{
    d->idObj = idObj;
}

unsigned NoteRecord::opts() const
{
    return d->opts;
}

void NoteRecord::setOpts(unsigned opts )
{
    d->opts = opts;
}

unsigned NoteRecord::row() const
{
    return d->row;
}

void NoteRecord::setRow(unsigned row )
{
    d->row = row;
}

void NoteRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 8) {
        setIsValid(false);
        return;
    }
    setRow(readU16(data));
    setColumn(readU16(data + 2));
    setOpts(readU16(data + 4));
    setIdObj(readU16(data + 6));
}

void NoteRecord::dump( std::ostream& out ) const
{
    out << "Note" << std::endl;
    out << "                Row : " << row() << std::endl;
    out << "             Column : " << column() << std::endl;
    out << "               Opts : " << opts() << std::endl;
    out << "              IdObj : " << idObj() << std::endl;
}

static Record* createNoteRecord(Workbook *book)
{
    return new NoteRecord(book);
}

// ========== ProtectRecord ==========

const unsigned ProtectRecord::id = 0x0012;

class ProtectRecord::Private
{
public:
    bool locked;
};

ProtectRecord::ProtectRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setLocked(false);
}

ProtectRecord::~ProtectRecord()
{
    delete d;
}

ProtectRecord::ProtectRecord( const ProtectRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ProtectRecord& ProtectRecord::operator=( const ProtectRecord& record )
{
    *d = *record.d;
    return *this;
}

bool ProtectRecord::isLocked() const
{
    return d->locked;
}

void ProtectRecord::setLocked(bool locked )
{
    d->locked = locked;
}

void ProtectRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setLocked(readU16(data) != 0);
}

void ProtectRecord::dump( std::ostream& out ) const
{
    out << "Protect" << std::endl;
    out << "             Locked : " << isLocked() << std::endl;
}

static Record* createProtectRecord(Workbook *book)
{
    return new ProtectRecord(book);
}

// ========== DefaultRowHeightRecord ==========

const unsigned DefaultRowHeightRecord::id = 0x225;

class DefaultRowHeightRecord::Private
{
public:
    int miyRw;
    int miyRwHidden;
    unsigned opts;
};

DefaultRowHeightRecord::DefaultRowHeightRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setMiyRw(0);
    setMiyRwHidden(0);
    setOpts(0);
}

DefaultRowHeightRecord::~DefaultRowHeightRecord()
{
    delete d;
}

DefaultRowHeightRecord::DefaultRowHeightRecord( const DefaultRowHeightRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DefaultRowHeightRecord& DefaultRowHeightRecord::operator=( const DefaultRowHeightRecord& record )
{
    *d = *record.d;
    return *this;
}

int DefaultRowHeightRecord::miyRw() const
{
    return d->miyRw;
}

void DefaultRowHeightRecord::setMiyRw(int miyRw )
{
    d->miyRw = miyRw;
}

int DefaultRowHeightRecord::miyRwHidden() const
{
    return d->miyRwHidden;
}

void DefaultRowHeightRecord::setMiyRwHidden(int miyRwHidden )
{
    d->miyRwHidden = miyRwHidden;
}

unsigned DefaultRowHeightRecord::opts() const
{
    return d->opts;
}

void DefaultRowHeightRecord::setOpts(unsigned opts )
{
    d->opts = opts;
}

void DefaultRowHeightRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 6) {
        setIsValid(false);
        return;
    }
    setOpts(readU16(data));
    setMiyRw(readS16(data + 2));
    setMiyRwHidden(readS16(data + 4));
}

void DefaultRowHeightRecord::dump( std::ostream& out ) const
{
    out << "DefaultRowHeight" << std::endl;
    out << "               Opts : " << opts() << std::endl;
    out << "              MiyRw : " << miyRw() << std::endl;
    out << "        MiyRwHidden : " << miyRwHidden() << std::endl;
}

static Record* createDefaultRowHeightRecord(Workbook *book)
{
    return new DefaultRowHeightRecord(book);
}

// ========== DefaultColWidthRecord ==========

const unsigned DefaultColWidthRecord::id = 0x55;

class DefaultColWidthRecord::Private
{
public:
    unsigned cchdefColWidth;
};

DefaultColWidthRecord::DefaultColWidthRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCchdefColWidth(0);
}

DefaultColWidthRecord::~DefaultColWidthRecord()
{
    delete d;
}

DefaultColWidthRecord::DefaultColWidthRecord( const DefaultColWidthRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DefaultColWidthRecord& DefaultColWidthRecord::operator=( const DefaultColWidthRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned DefaultColWidthRecord::cchdefColWidth() const
{
    return d->cchdefColWidth;
}

void DefaultColWidthRecord::setCchdefColWidth(unsigned cchdefColWidth )
{
    d->cchdefColWidth = cchdefColWidth;
}

void DefaultColWidthRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setCchdefColWidth(readU16(data));
}

void DefaultColWidthRecord::dump( std::ostream& out ) const
{
    out << "DefaultColWidth" << std::endl;
    out << "     CchdefColWidth : " << cchdefColWidth() << std::endl;
}

static Record* createDefaultColWidthRecord(Workbook *book)
{
    return new DefaultColWidthRecord(book);
}

// ========== HCenterRecord ==========

const unsigned HCenterRecord::id = 0x0083;

class HCenterRecord::Private
{
public:
    bool centered;
};

HCenterRecord::HCenterRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCentered(false);
}

HCenterRecord::~HCenterRecord()
{
    delete d;
}

HCenterRecord::HCenterRecord( const HCenterRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

HCenterRecord& HCenterRecord::operator=( const HCenterRecord& record )
{
    *d = *record.d;
    return *this;
}

bool HCenterRecord::isCentered() const
{
    return d->centered;
}

void HCenterRecord::setCentered(bool centered )
{
    d->centered = centered;
}

void HCenterRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setCentered(readU16(data) != 0);
}

void HCenterRecord::dump( std::ostream& out ) const
{
    out << "HCenter" << std::endl;
    out << "           Centered : " << isCentered() << std::endl;
}

static Record* createHCenterRecord(Workbook *book)
{
    return new HCenterRecord(book);
}

// ========== VCenterRecord ==========

const unsigned VCenterRecord::id = 0x0084;

class VCenterRecord::Private
{
public:
    bool centered;
};

VCenterRecord::VCenterRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCentered(false);
}

VCenterRecord::~VCenterRecord()
{
    delete d;
}

VCenterRecord::VCenterRecord( const VCenterRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

VCenterRecord& VCenterRecord::operator=( const VCenterRecord& record )
{
    *d = *record.d;
    return *this;
}

bool VCenterRecord::isCentered() const
{
    return d->centered;
}

void VCenterRecord::setCentered(bool centered )
{
    d->centered = centered;
}

void VCenterRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setCentered(readU16(data) != 0);
}

void VCenterRecord::dump( std::ostream& out ) const
{
    out << "VCenter" << std::endl;
    out << "           Centered : " << isCentered() << std::endl;
}

static Record* createVCenterRecord(Workbook *book)
{
    return new VCenterRecord(book);
}

// ========== SetupRecord ==========

const unsigned SetupRecord::id = 0x00a1;

class SetupRecord::Private
{
public:
    bool commentsAsEndNotes;
    bool custumStartPageNumber;
    bool draftQuality;
    ErrorPrintMode errorPrintMode;
    unsigned fitHeightToPageCount;
    unsigned fitWidthToPageCount;
    double footerMargin;
    bool greyscale;
    double headerMargin;
    bool leftToRight;
    bool noOrientationSet;
    bool noPaperSizeSet;
    unsigned numCopies;
    unsigned paperSize;
    bool portrait;
    bool printNotes;
    unsigned printResolution;
    unsigned scalePercentage;
    int startingPageNumber;
    unsigned verticalPrintResolution;
};

SetupRecord::SetupRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCommentsAsEndNotes(false);
    setCustumStartPageNumber(false);
    setDraftQuality(false);
    setFitHeightToPageCount(0);
    setFitWidthToPageCount(0);
    setGreyscale(false);
    setLeftToRight(false);
    setNoOrientationSet(false);
    setNoPaperSizeSet(false);
    setNumCopies(0);
    setPaperSize(0);
    setPortrait(false);
    setPrintNotes(false);
    setPrintResolution(0);
    setScalePercentage(0);
    setStartingPageNumber(0);
    setVerticalPrintResolution(0);
}

SetupRecord::~SetupRecord()
{
    delete d;
}

SetupRecord::SetupRecord( const SetupRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SetupRecord& SetupRecord::operator=( const SetupRecord& record )
{
    *d = *record.d;
    return *this;
}

UString SetupRecord::errorPrintModeToString(ErrorPrintMode errorPrintMode)
{
    switch (errorPrintMode) {
        case ErrorsAsDisplayed: return UString("ErrorsAsDisplayed");
        case ErrorsAsBlank: return UString("ErrorsAsBlank");
        case ErrorsAsDashes: return UString("ErrorsAsDashes");
        case ErrorsAsNA: return UString("ErrorsAsNA");
        default: return UString("Unknown: ") + UString::from(errorPrintMode);
    }
}

bool SetupRecord::isCommentsAsEndNotes() const
{
    return d->commentsAsEndNotes;
}

void SetupRecord::setCommentsAsEndNotes(bool commentsAsEndNotes )
{
    d->commentsAsEndNotes = commentsAsEndNotes;
}

bool SetupRecord::isCustumStartPageNumber() const
{
    return d->custumStartPageNumber;
}

void SetupRecord::setCustumStartPageNumber(bool custumStartPageNumber )
{
    d->custumStartPageNumber = custumStartPageNumber;
}

bool SetupRecord::isDraftQuality() const
{
    return d->draftQuality;
}

void SetupRecord::setDraftQuality(bool draftQuality )
{
    d->draftQuality = draftQuality;
}

SetupRecord::ErrorPrintMode SetupRecord::errorPrintMode() const
{
    return d->errorPrintMode;
}

void SetupRecord::setErrorPrintMode(ErrorPrintMode errorPrintMode )
{
    d->errorPrintMode = errorPrintMode;
}

unsigned SetupRecord::fitHeightToPageCount() const
{
    return d->fitHeightToPageCount;
}

void SetupRecord::setFitHeightToPageCount(unsigned fitHeightToPageCount )
{
    d->fitHeightToPageCount = fitHeightToPageCount;
}

unsigned SetupRecord::fitWidthToPageCount() const
{
    return d->fitWidthToPageCount;
}

void SetupRecord::setFitWidthToPageCount(unsigned fitWidthToPageCount )
{
    d->fitWidthToPageCount = fitWidthToPageCount;
}

double SetupRecord::footerMargin() const
{
    return d->footerMargin;
}

void SetupRecord::setFooterMargin(double footerMargin )
{
    d->footerMargin = footerMargin;
}

bool SetupRecord::isGreyscale() const
{
    return d->greyscale;
}

void SetupRecord::setGreyscale(bool greyscale )
{
    d->greyscale = greyscale;
}

double SetupRecord::headerMargin() const
{
    return d->headerMargin;
}

void SetupRecord::setHeaderMargin(double headerMargin )
{
    d->headerMargin = headerMargin;
}

bool SetupRecord::isLeftToRight() const
{
    return d->leftToRight;
}

void SetupRecord::setLeftToRight(bool leftToRight )
{
    d->leftToRight = leftToRight;
}

bool SetupRecord::isNoOrientationSet() const
{
    return d->noOrientationSet;
}

void SetupRecord::setNoOrientationSet(bool noOrientationSet )
{
    d->noOrientationSet = noOrientationSet;
}

bool SetupRecord::isNoPaperSizeSet() const
{
    return d->noPaperSizeSet;
}

void SetupRecord::setNoPaperSizeSet(bool noPaperSizeSet )
{
    d->noPaperSizeSet = noPaperSizeSet;
}

unsigned SetupRecord::numCopies() const
{
    return d->numCopies;
}

void SetupRecord::setNumCopies(unsigned numCopies )
{
    d->numCopies = numCopies;
}

unsigned SetupRecord::paperSize() const
{
    return d->paperSize;
}

void SetupRecord::setPaperSize(unsigned paperSize )
{
    d->paperSize = paperSize;
}

bool SetupRecord::isPortrait() const
{
    return d->portrait;
}

void SetupRecord::setPortrait(bool portrait )
{
    d->portrait = portrait;
}

bool SetupRecord::isPrintNotes() const
{
    return d->printNotes;
}

void SetupRecord::setPrintNotes(bool printNotes )
{
    d->printNotes = printNotes;
}

unsigned SetupRecord::printResolution() const
{
    return d->printResolution;
}

void SetupRecord::setPrintResolution(unsigned printResolution )
{
    d->printResolution = printResolution;
}

unsigned SetupRecord::scalePercentage() const
{
    return d->scalePercentage;
}

void SetupRecord::setScalePercentage(unsigned scalePercentage )
{
    d->scalePercentage = scalePercentage;
}

int SetupRecord::startingPageNumber() const
{
    return d->startingPageNumber;
}

void SetupRecord::setStartingPageNumber(int startingPageNumber )
{
    d->startingPageNumber = startingPageNumber;
}

unsigned SetupRecord::verticalPrintResolution() const
{
    return d->verticalPrintResolution;
}

void SetupRecord::setVerticalPrintResolution(unsigned verticalPrintResolution )
{
    d->verticalPrintResolution = verticalPrintResolution;
}

void SetupRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 34) {
        setIsValid(false);
        return;
    }
    setPaperSize(readU16(data));
    setScalePercentage(readU16(data + 2));
    setStartingPageNumber(readS16(data + 4));
    setFitWidthToPageCount(readU16(data + 6));
    setFitHeightToPageCount(readU16(data + 8));
    setLeftToRight(((readU8(data + 10)) & 0x1) != 0);
    setPortrait(((readU8(data + 10) >> 1) & 0x1) != 0);
    setNoPaperSizeSet(((readU8(data + 10) >> 2) & 0x1) != 0);
    setGreyscale(((readU8(data + 10) >> 3) & 0x1) != 0);
    setDraftQuality(((readU8(data + 10) >> 4) & 0x1) != 0);
    setPrintNotes(((readU8(data + 10) >> 5) & 0x1) != 0);
    setNoOrientationSet(((readU8(data + 10) >> 6) & 0x1) != 0);
    setCustumStartPageNumber(((readU8(data + 10) >> 7) & 0x1) != 0);
    setCommentsAsEndNotes(((readU8(data + 11) >> 1) & 0x1) != 0);
    setErrorPrintMode(static_cast<ErrorPrintMode>(((readU8(data + 11) >> 2) & 0x3)));
    setPrintResolution(readU16(data + 12));
    setVerticalPrintResolution(readU16(data + 14));
    setHeaderMargin(readFloat64(data + 16));
    setFooterMargin(readFloat64(data + 24));
    setNumCopies(readU16(data + 32));
}

void SetupRecord::dump( std::ostream& out ) const
{
    out << "Setup" << std::endl;
    out << "          PaperSize : " << paperSize() << std::endl;
    out << "    ScalePercentage : " << scalePercentage() << std::endl;
    out << " StartingPageNumber : " << startingPageNumber() << std::endl;
    out << "FitWidthToPageCount : " << fitWidthToPageCount() << std::endl;
    out << "FitHeightToPageCount : " << fitHeightToPageCount() << std::endl;
    out << "        LeftToRight : " << isLeftToRight() << std::endl;
    out << "           Portrait : " << isPortrait() << std::endl;
    out << "     NoPaperSizeSet : " << isNoPaperSizeSet() << std::endl;
    out << "          Greyscale : " << isGreyscale() << std::endl;
    out << "       DraftQuality : " << isDraftQuality() << std::endl;
    out << "         PrintNotes : " << isPrintNotes() << std::endl;
    out << "   NoOrientationSet : " << isNoOrientationSet() << std::endl;
    out << "CustumStartPageNumber : " << isCustumStartPageNumber() << std::endl;
    out << " CommentsAsEndNotes : " << isCommentsAsEndNotes() << std::endl;
    out << "     ErrorPrintMode : " << errorPrintModeToString(errorPrintMode()) << std::endl;
    out << "    PrintResolution : " << printResolution() << std::endl;
    out << "VerticalPrintResolution : " << verticalPrintResolution() << std::endl;
    out << "       HeaderMargin : " << headerMargin() << std::endl;
    out << "       FooterMargin : " << footerMargin() << std::endl;
    out << "          NumCopies : " << numCopies() << std::endl;
}

static Record* createSetupRecord(Workbook *book)
{
    return new SetupRecord(book);
}

// ========== CrtLineRecord ==========

const unsigned CrtLineRecord::id = 0x101c;

class CrtLineRecord::Private
{
public:
};

CrtLineRecord::CrtLineRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CrtLineRecord::~CrtLineRecord()
{
    delete d;
}

CrtLineRecord::CrtLineRecord( const CrtLineRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CrtLineRecord& CrtLineRecord::operator=( const CrtLineRecord& record )
{
    *d = *record.d;
    return *this;
}

void CrtLineRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CrtLineRecord::dump( std::ostream& out ) const
{
    out << "CrtLine" << std::endl;
}

static Record* createCrtLineRecord(Workbook *book)
{
    return new CrtLineRecord(book);
}

// ========== UnitsRecord ==========

const unsigned UnitsRecord::id = 0x1001;

class UnitsRecord::Private
{
public:
};

UnitsRecord::UnitsRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

UnitsRecord::~UnitsRecord()
{
    delete d;
}

UnitsRecord::UnitsRecord( const UnitsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

UnitsRecord& UnitsRecord::operator=( const UnitsRecord& record )
{
    *d = *record.d;
    return *this;
}

void UnitsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void UnitsRecord::dump( std::ostream& out ) const
{
    out << "Units" << std::endl;
}

static Record* createUnitsRecord(Workbook *book)
{
    return new UnitsRecord(book);
}

// ========== PrintSizeRecord ==========

const unsigned PrintSizeRecord::id = 0x0033;

class PrintSizeRecord::Private
{
public:
    PrintSize printSize;
};

PrintSizeRecord::PrintSizeRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PrintSizeRecord::~PrintSizeRecord()
{
    delete d;
}

PrintSizeRecord::PrintSizeRecord( const PrintSizeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PrintSizeRecord& PrintSizeRecord::operator=( const PrintSizeRecord& record )
{
    *d = *record.d;
    return *this;
}

UString PrintSizeRecord::printSizeToString(PrintSize printSize)
{
    switch (printSize) {
        case WorkbookDefaults: return UString("WorkbookDefaults");
        case FillPage: return UString("FillPage");
        case FillPageProportionally: return UString("FillPageProportionally");
        case SizeFromChartRecord: return UString("SizeFromChartRecord");
        default: return UString("Unknown: ") + UString::from(printSize);
    }
}

PrintSizeRecord::PrintSize PrintSizeRecord::printSize() const
{
    return d->printSize;
}

void PrintSizeRecord::setPrintSize(PrintSize printSize )
{
    d->printSize = printSize;
}

void PrintSizeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setPrintSize(static_cast<PrintSize>(readU16(data)));
}

void PrintSizeRecord::dump( std::ostream& out ) const
{
    out << "PrintSize" << std::endl;
    out << "          PrintSize : " << printSizeToString(printSize()) << std::endl;
}

static Record* createPrintSizeRecord(Workbook *book)
{
    return new PrintSizeRecord(book);
}

// ========== ChartRecord ==========

const unsigned ChartRecord::id = 0x1002;

class ChartRecord::Private
{
public:
    double height;
    double width;
    double x;
    double y;
};

ChartRecord::ChartRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ChartRecord::~ChartRecord()
{
    delete d;
}

ChartRecord::ChartRecord( const ChartRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ChartRecord& ChartRecord::operator=( const ChartRecord& record )
{
    *d = *record.d;
    return *this;
}

double ChartRecord::height() const
{
    return d->height;
}

void ChartRecord::setHeight(double height )
{
    d->height = height;
}

double ChartRecord::width() const
{
    return d->width;
}

void ChartRecord::setWidth(double width )
{
    d->width = width;
}

double ChartRecord::x() const
{
    return d->x;
}

void ChartRecord::setX(double x )
{
    d->x = x;
}

double ChartRecord::y() const
{
    return d->y;
}

void ChartRecord::setY(double y )
{
    d->y = y;
}

void ChartRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 16) {
        setIsValid(false);
        return;
    }
    setX(readFixed32(data));
    setY(readFixed32(data + 4));
    setWidth(readFixed32(data + 8));
    setHeight(readFixed32(data + 12));
}

void ChartRecord::dump( std::ostream& out ) const
{
    out << "Chart" << std::endl;
    out << "                  X : " << x() << std::endl;
    out << "                  Y : " << y() << std::endl;
    out << "              Width : " << width() << std::endl;
    out << "             Height : " << height() << std::endl;
}

static Record* createChartRecord(Workbook *book)
{
    return new ChartRecord(book);
}

// ========== SheetExtRecord ==========

const unsigned SheetExtRecord::id = 0x0862;

class SheetExtRecord::Private
{
public:
    unsigned cb;
    unsigned color;
    bool fCondFmtCalc;
    bool fNotPublished;
    unsigned icvPlain;
    unsigned icvPlain12;
};

SheetExtRecord::SheetExtRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setCb(0);
    setColor(0);
    setFCondFmtCalc(false);
    setFNotPublished(false);
    setIcvPlain(0);
    setIcvPlain12(0);
}

SheetExtRecord::~SheetExtRecord()
{
    delete d;
}

SheetExtRecord::SheetExtRecord( const SheetExtRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SheetExtRecord& SheetExtRecord::operator=( const SheetExtRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned SheetExtRecord::cb() const
{
    return d->cb;
}

void SheetExtRecord::setCb(unsigned cb )
{
    d->cb = cb;
}

unsigned SheetExtRecord::color() const
{
    return d->color;
}

void SheetExtRecord::setColor(unsigned color )
{
    d->color = color;
}

bool SheetExtRecord::isFCondFmtCalc() const
{
    return d->fCondFmtCalc;
}

void SheetExtRecord::setFCondFmtCalc(bool fCondFmtCalc )
{
    d->fCondFmtCalc = fCondFmtCalc;
}

bool SheetExtRecord::isFNotPublished() const
{
    return d->fNotPublished;
}

void SheetExtRecord::setFNotPublished(bool fNotPublished )
{
    d->fNotPublished = fNotPublished;
}

unsigned SheetExtRecord::icvPlain() const
{
    return d->icvPlain;
}

void SheetExtRecord::setIcvPlain(unsigned icvPlain )
{
    d->icvPlain = icvPlain;
}

unsigned SheetExtRecord::icvPlain12() const
{
    return d->icvPlain12;
}

void SheetExtRecord::setIcvPlain12(unsigned icvPlain12 )
{
    d->icvPlain12 = icvPlain12;
}

void SheetExtRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 14) {
        setIsValid(false);
        return;
    }
    setCb(readU32(data));
    setIcvPlain(((readU8(data + 4)) & 0x7f));
    setIcvPlain12(((readU8(data + 8)) & 0x7f));
    setFCondFmtCalc(((readU8(data + 8) >> 7) & 0x1) != 0);
    setFNotPublished(((readU8(data + 9)) & 0x1) != 0);
    setColor(readU16(data + 12));
}

void SheetExtRecord::dump( std::ostream& out ) const
{
    out << "SheetExt" << std::endl;
    out << "                 Cb : " << cb() << std::endl;
    out << "           IcvPlain : " << icvPlain() << std::endl;
    out << "         IcvPlain12 : " << icvPlain12() << std::endl;
    out << "       FCondFmtCalc : " << isFCondFmtCalc() << std::endl;
    out << "      FNotPublished : " << isFNotPublished() << std::endl;
    out << "              Color : " << color() << std::endl;
}

static Record* createSheetExtRecord(Workbook *book)
{
    return new SheetExtRecord(book);
}

// ========== BeginRecord ==========

const unsigned BeginRecord::id = 0x1033;

class BeginRecord::Private
{
public:
};

BeginRecord::BeginRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BeginRecord::~BeginRecord()
{
    delete d;
}

BeginRecord::BeginRecord( const BeginRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BeginRecord& BeginRecord::operator=( const BeginRecord& record )
{
    *d = *record.d;
    return *this;
}

void BeginRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void BeginRecord::dump( std::ostream& out ) const
{
    out << "Begin" << std::endl;
}

static Record* createBeginRecord(Workbook *book)
{
    return new BeginRecord(book);
}

// ========== EndRecord ==========

const unsigned EndRecord::id = 0x1034;

class EndRecord::Private
{
public:
};

EndRecord::EndRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

EndRecord::~EndRecord()
{
    delete d;
}

EndRecord::EndRecord( const EndRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

EndRecord& EndRecord::operator=( const EndRecord& record )
{
    *d = *record.d;
    return *this;
}

void EndRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void EndRecord::dump( std::ostream& out ) const
{
    out << "End" << std::endl;
}

static Record* createEndRecord(Workbook *book)
{
    return new EndRecord(book);
}

// ========== DataFormatRecord ==========

const unsigned DataFormatRecord::id = 0x1006;

class DataFormatRecord::Private
{
public:
    unsigned iss;
    unsigned xi;
    unsigned yi;
};

DataFormatRecord::DataFormatRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setIss(0);
    setXi(0);
    setYi(0);
}

DataFormatRecord::~DataFormatRecord()
{
    delete d;
}

DataFormatRecord::DataFormatRecord( const DataFormatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DataFormatRecord& DataFormatRecord::operator=( const DataFormatRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned DataFormatRecord::iss() const
{
    return d->iss;
}

void DataFormatRecord::setIss(unsigned iss )
{
    d->iss = iss;
}

unsigned DataFormatRecord::xi() const
{
    return d->xi;
}

void DataFormatRecord::setXi(unsigned xi )
{
    d->xi = xi;
}

unsigned DataFormatRecord::yi() const
{
    return d->yi;
}

void DataFormatRecord::setYi(unsigned yi )
{
    d->yi = yi;
}

void DataFormatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 6) {
        setIsValid(false);
        return;
    }
    setXi(readU16(data));
    setYi(readU16(data + 2));
    setIss(readU16(data + 4));
}

void DataFormatRecord::dump( std::ostream& out ) const
{
    out << "DataFormat" << std::endl;
    out << "                 Xi : " << xi() << std::endl;
    out << "                 Yi : " << yi() << std::endl;
    out << "                Iss : " << iss() << std::endl;
}

static Record* createDataFormatRecord(Workbook *book)
{
    return new DataFormatRecord(book);
}

// ========== LineFormatRecord ==========

const unsigned LineFormatRecord::id = 0x1007;

class LineFormatRecord::Private
{
public:
};

LineFormatRecord::LineFormatRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

LineFormatRecord::~LineFormatRecord()
{
    delete d;
}

LineFormatRecord::LineFormatRecord( const LineFormatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LineFormatRecord& LineFormatRecord::operator=( const LineFormatRecord& record )
{
    *d = *record.d;
    return *this;
}

void LineFormatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void LineFormatRecord::dump( std::ostream& out ) const
{
    out << "LineFormat" << std::endl;
}

static Record* createLineFormatRecord(Workbook *book)
{
    return new LineFormatRecord(book);
}

// ========== MarkerFormatRecord ==========

const unsigned MarkerFormatRecord::id = 0x1009;

class MarkerFormatRecord::Private
{
public:
};

MarkerFormatRecord::MarkerFormatRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

MarkerFormatRecord::~MarkerFormatRecord()
{
    delete d;
}

MarkerFormatRecord::MarkerFormatRecord( const MarkerFormatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

MarkerFormatRecord& MarkerFormatRecord::operator=( const MarkerFormatRecord& record )
{
    *d = *record.d;
    return *this;
}

void MarkerFormatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void MarkerFormatRecord::dump( std::ostream& out ) const
{
    out << "MarkerFormat" << std::endl;
}

static Record* createMarkerFormatRecord(Workbook *book)
{
    return new MarkerFormatRecord(book);
}

// ========== AreaFormatRecord ==========

const unsigned AreaFormatRecord::id = 0x100a;

class AreaFormatRecord::Private
{
public:
};

AreaFormatRecord::AreaFormatRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AreaFormatRecord::~AreaFormatRecord()
{
    delete d;
}

AreaFormatRecord::AreaFormatRecord( const AreaFormatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AreaFormatRecord& AreaFormatRecord::operator=( const AreaFormatRecord& record )
{
    *d = *record.d;
    return *this;
}

void AreaFormatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AreaFormatRecord::dump( std::ostream& out ) const
{
    out << "AreaFormat" << std::endl;
}

static Record* createAreaFormatRecord(Workbook *book)
{
    return new AreaFormatRecord(book);
}

// ========== PieRecord ==========

const unsigned PieRecord::id = 0x1019;

class PieRecord::Private
{
public:
    unsigned anStart;
    bool fHasShadow;
    bool fShowLdrLines;
    unsigned pcDonut;
};

PieRecord::PieRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setAnStart(0);
    setFHasShadow(false);
    setFShowLdrLines(false);
    setPcDonut(0);
}

PieRecord::~PieRecord()
{
    delete d;
}

PieRecord::PieRecord( const PieRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PieRecord& PieRecord::operator=( const PieRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned PieRecord::anStart() const
{
    return d->anStart;
}

void PieRecord::setAnStart(unsigned anStart )
{
    d->anStart = anStart;
}

bool PieRecord::isFHasShadow() const
{
    return d->fHasShadow;
}

void PieRecord::setFHasShadow(bool fHasShadow )
{
    d->fHasShadow = fHasShadow;
}

bool PieRecord::isFShowLdrLines() const
{
    return d->fShowLdrLines;
}

void PieRecord::setFShowLdrLines(bool fShowLdrLines )
{
    d->fShowLdrLines = fShowLdrLines;
}

unsigned PieRecord::pcDonut() const
{
    return d->pcDonut;
}

void PieRecord::setPcDonut(unsigned pcDonut )
{
    d->pcDonut = pcDonut;
}

void PieRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 6) {
        setIsValid(false);
        return;
    }
    setAnStart(readU16(data));
    setPcDonut(readU16(data + 2));
    setFHasShadow(((readU8(data + 4)) & 0x1) != 0);
    setFShowLdrLines(((readU8(data + 4) >> 1) & 0x1) != 0);
}

void PieRecord::dump( std::ostream& out ) const
{
    out << "Pie" << std::endl;
    out << "            AnStart : " << anStart() << std::endl;
    out << "            PcDonut : " << pcDonut() << std::endl;
    out << "         FHasShadow : " << isFHasShadow() << std::endl;
    out << "      FShowLdrLines : " << isFShowLdrLines() << std::endl;
}

static Record* createPieRecord(Workbook *book)
{
    return new PieRecord(book);
}

// ========== PieFormatRecord ==========

const unsigned PieFormatRecord::id = 0x100b;

class PieFormatRecord::Private
{
public:
    int pcExplode;
};

PieFormatRecord::PieFormatRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setPcExplode(0);
}

PieFormatRecord::~PieFormatRecord()
{
    delete d;
}

PieFormatRecord::PieFormatRecord( const PieFormatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PieFormatRecord& PieFormatRecord::operator=( const PieFormatRecord& record )
{
    *d = *record.d;
    return *this;
}

int PieFormatRecord::pcExplode() const
{
    return d->pcExplode;
}

void PieFormatRecord::setPcExplode(int pcExplode )
{
    d->pcExplode = pcExplode;
}

void PieFormatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setPcExplode(readS16(data));
}

void PieFormatRecord::dump( std::ostream& out ) const
{
    out << "PieFormat" << std::endl;
    out << "          PcExplode : " << pcExplode() << std::endl;
}

static Record* createPieFormatRecord(Workbook *book)
{
    return new PieFormatRecord(book);
}

// ========== AttachedLabelRecord ==========

const unsigned AttachedLabelRecord::id = 0x100c;

class AttachedLabelRecord::Private
{
public:
};

AttachedLabelRecord::AttachedLabelRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AttachedLabelRecord::~AttachedLabelRecord()
{
    delete d;
}

AttachedLabelRecord::AttachedLabelRecord( const AttachedLabelRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AttachedLabelRecord& AttachedLabelRecord::operator=( const AttachedLabelRecord& record )
{
    *d = *record.d;
    return *this;
}

void AttachedLabelRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AttachedLabelRecord::dump( std::ostream& out ) const
{
    out << "AttachedLabel" << std::endl;
}

static Record* createAttachedLabelRecord(Workbook *book)
{
    return new AttachedLabelRecord(book);
}

// ========== ChartFormatRecord ==========

const unsigned ChartFormatRecord::id = 0x1014;

class ChartFormatRecord::Private
{
public:
};

ChartFormatRecord::ChartFormatRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ChartFormatRecord::~ChartFormatRecord()
{
    delete d;
}

ChartFormatRecord::ChartFormatRecord( const ChartFormatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ChartFormatRecord& ChartFormatRecord::operator=( const ChartFormatRecord& record )
{
    *d = *record.d;
    return *this;
}

void ChartFormatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ChartFormatRecord::dump( std::ostream& out ) const
{
    out << "ChartFormat" << std::endl;
}

static Record* createChartFormatRecord(Workbook *book)
{
    return new ChartFormatRecord(book);
}

// ========== LegendRecord ==========

const unsigned LegendRecord::id = 0x1015;

class LegendRecord::Private
{
public:
    unsigned dx;
    unsigned dy;
    bool fAutoPosX;
    bool fAutoPosY;
    bool fAutoPosition;
    bool fVert;
    bool fWasDataTable;
    unsigned unused;
    unsigned wSpace;
    unsigned x;
    unsigned y;
};

LegendRecord::LegendRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setDx(0);
    setDy(0);
    setFAutoPosX(false);
    setFAutoPosY(false);
    setFAutoPosition(false);
    setFVert(false);
    setFWasDataTable(false);
    setUnused(0);
    setWSpace(0);
    setX(0);
    setY(0);
}

LegendRecord::~LegendRecord()
{
    delete d;
}

LegendRecord::LegendRecord( const LegendRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LegendRecord& LegendRecord::operator=( const LegendRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned LegendRecord::dx() const
{
    return d->dx;
}

void LegendRecord::setDx(unsigned dx )
{
    d->dx = dx;
}

unsigned LegendRecord::dy() const
{
    return d->dy;
}

void LegendRecord::setDy(unsigned dy )
{
    d->dy = dy;
}

bool LegendRecord::isFAutoPosX() const
{
    return d->fAutoPosX;
}

void LegendRecord::setFAutoPosX(bool fAutoPosX )
{
    d->fAutoPosX = fAutoPosX;
}

bool LegendRecord::isFAutoPosY() const
{
    return d->fAutoPosY;
}

void LegendRecord::setFAutoPosY(bool fAutoPosY )
{
    d->fAutoPosY = fAutoPosY;
}

bool LegendRecord::isFAutoPosition() const
{
    return d->fAutoPosition;
}

void LegendRecord::setFAutoPosition(bool fAutoPosition )
{
    d->fAutoPosition = fAutoPosition;
}

bool LegendRecord::isFVert() const
{
    return d->fVert;
}

void LegendRecord::setFVert(bool fVert )
{
    d->fVert = fVert;
}

bool LegendRecord::isFWasDataTable() const
{
    return d->fWasDataTable;
}

void LegendRecord::setFWasDataTable(bool fWasDataTable )
{
    d->fWasDataTable = fWasDataTable;
}

unsigned LegendRecord::unused() const
{
    return d->unused;
}

void LegendRecord::setUnused(unsigned unused )
{
    d->unused = unused;
}

unsigned LegendRecord::wSpace() const
{
    return d->wSpace;
}

void LegendRecord::setWSpace(unsigned wSpace )
{
    d->wSpace = wSpace;
}

unsigned LegendRecord::x() const
{
    return d->x;
}

void LegendRecord::setX(unsigned x )
{
    d->x = x;
}

unsigned LegendRecord::y() const
{
    return d->y;
}

void LegendRecord::setY(unsigned y )
{
    d->y = y;
}

void LegendRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 20) {
        setIsValid(false);
        return;
    }
    setX(readU32(data));
    setY(readU32(data + 4));
    setDx(readU32(data + 8));
    setDy(readU32(data + 12));
    setUnused(readU8(data + 16));
    setWSpace(readU8(data + 17));
    setFAutoPosition(((readU8(data + 18)) & 0x1) != 0);
    setFAutoPosX(((readU8(data + 18) >> 2) & 0x1) != 0);
    setFAutoPosY(((readU8(data + 18) >> 3) & 0x1) != 0);
    setFVert(((readU8(data + 18) >> 4) & 0x1) != 0);
    setFWasDataTable(((readU8(data + 18) >> 5) & 0x1) != 0);
}

void LegendRecord::dump( std::ostream& out ) const
{
    out << "Legend" << std::endl;
    out << "                  X : " << x() << std::endl;
    out << "                  Y : " << y() << std::endl;
    out << "                 Dx : " << dx() << std::endl;
    out << "                 Dy : " << dy() << std::endl;
    out << "             Unused : " << unused() << std::endl;
    out << "             WSpace : " << wSpace() << std::endl;
    out << "      FAutoPosition : " << isFAutoPosition() << std::endl;
    out << "          FAutoPosX : " << isFAutoPosX() << std::endl;
    out << "          FAutoPosY : " << isFAutoPosY() << std::endl;
    out << "              FVert : " << isFVert() << std::endl;
    out << "      FWasDataTable : " << isFWasDataTable() << std::endl;
}

static Record* createLegendRecord(Workbook *book)
{
    return new LegendRecord(book);
}

// ========== SeriesListRecord ==========

const unsigned SeriesListRecord::id = 0x1016;

class SeriesListRecord::Private
{
public:
};

SeriesListRecord::SeriesListRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SeriesListRecord::~SeriesListRecord()
{
    delete d;
}

SeriesListRecord::SeriesListRecord( const SeriesListRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SeriesListRecord& SeriesListRecord::operator=( const SeriesListRecord& record )
{
    *d = *record.d;
    return *this;
}

void SeriesListRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void SeriesListRecord::dump( std::ostream& out ) const
{
    out << "SeriesList" << std::endl;
}

static Record* createSeriesListRecord(Workbook *book)
{
    return new SeriesListRecord(book);
}

// ========== BarRecord ==========

const unsigned BarRecord::id = 0x1017;

class BarRecord::Private
{
public:
};

BarRecord::BarRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BarRecord::~BarRecord()
{
    delete d;
}

BarRecord::BarRecord( const BarRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BarRecord& BarRecord::operator=( const BarRecord& record )
{
    *d = *record.d;
    return *this;
}

void BarRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void BarRecord::dump( std::ostream& out ) const
{
    out << "Bar" << std::endl;
}

static Record* createBarRecord(Workbook *book)
{
    return new BarRecord(book);
}

// ========== LineRecord ==========

const unsigned LineRecord::id = 0x1018;

class LineRecord::Private
{
public:
};

LineRecord::LineRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

LineRecord::~LineRecord()
{
    delete d;
}

LineRecord::LineRecord( const LineRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LineRecord& LineRecord::operator=( const LineRecord& record )
{
    *d = *record.d;
    return *this;
}

void LineRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void LineRecord::dump( std::ostream& out ) const
{
    out << "Line" << std::endl;
}

static Record* createLineRecord(Workbook *book)
{
    return new LineRecord(book);
}

// ========== AreaRecord ==========

const unsigned AreaRecord::id = 0x101a;

class AreaRecord::Private
{
public:
};

AreaRecord::AreaRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AreaRecord::~AreaRecord()
{
    delete d;
}

AreaRecord::AreaRecord( const AreaRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AreaRecord& AreaRecord::operator=( const AreaRecord& record )
{
    *d = *record.d;
    return *this;
}

void AreaRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AreaRecord::dump( std::ostream& out ) const
{
    out << "Area" << std::endl;
}

static Record* createAreaRecord(Workbook *book)
{
    return new AreaRecord(book);
}

// ========== ScatterRecord ==========

const unsigned ScatterRecord::id = 0x101b;

class ScatterRecord::Private
{
public:
};

ScatterRecord::ScatterRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ScatterRecord::~ScatterRecord()
{
    delete d;
}

ScatterRecord::ScatterRecord( const ScatterRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ScatterRecord& ScatterRecord::operator=( const ScatterRecord& record )
{
    *d = *record.d;
    return *this;
}

void ScatterRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ScatterRecord::dump( std::ostream& out ) const
{
    out << "Scatter" << std::endl;
}

static Record* createScatterRecord(Workbook *book)
{
    return new ScatterRecord(book);
}

// ========== AxisRecord ==========

const unsigned AxisRecord::id = 0x101d;

class AxisRecord::Private
{
public:
};

AxisRecord::AxisRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AxisRecord::~AxisRecord()
{
    delete d;
}

AxisRecord::AxisRecord( const AxisRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AxisRecord& AxisRecord::operator=( const AxisRecord& record )
{
    *d = *record.d;
    return *this;
}

void AxisRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AxisRecord::dump( std::ostream& out ) const
{
    out << "Axis" << std::endl;
}

static Record* createAxisRecord(Workbook *book)
{
    return new AxisRecord(book);
}

// ========== TickRecord ==========

const unsigned TickRecord::id = 0x101e;

class TickRecord::Private
{
public:
};

TickRecord::TickRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

TickRecord::~TickRecord()
{
    delete d;
}

TickRecord::TickRecord( const TickRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

TickRecord& TickRecord::operator=( const TickRecord& record )
{
    *d = *record.d;
    return *this;
}

void TickRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void TickRecord::dump( std::ostream& out ) const
{
    out << "Tick" << std::endl;
}

static Record* createTickRecord(Workbook *book)
{
    return new TickRecord(book);
}

// ========== ValueRangeRecord ==========

const unsigned ValueRangeRecord::id = 0x101f;

class ValueRangeRecord::Private
{
public:
};

ValueRangeRecord::ValueRangeRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ValueRangeRecord::~ValueRangeRecord()
{
    delete d;
}

ValueRangeRecord::ValueRangeRecord( const ValueRangeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ValueRangeRecord& ValueRangeRecord::operator=( const ValueRangeRecord& record )
{
    *d = *record.d;
    return *this;
}

void ValueRangeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ValueRangeRecord::dump( std::ostream& out ) const
{
    out << "ValueRange" << std::endl;
}

static Record* createValueRangeRecord(Workbook *book)
{
    return new ValueRangeRecord(book);
}

// ========== CatSerRangeRecord ==========

const unsigned CatSerRangeRecord::id = 0x1020;

class CatSerRangeRecord::Private
{
public:
};

CatSerRangeRecord::CatSerRangeRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CatSerRangeRecord::~CatSerRangeRecord()
{
    delete d;
}

CatSerRangeRecord::CatSerRangeRecord( const CatSerRangeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CatSerRangeRecord& CatSerRangeRecord::operator=( const CatSerRangeRecord& record )
{
    *d = *record.d;
    return *this;
}

void CatSerRangeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CatSerRangeRecord::dump( std::ostream& out ) const
{
    out << "CatSerRange" << std::endl;
}

static Record* createCatSerRangeRecord(Workbook *book)
{
    return new CatSerRangeRecord(book);
}

// ========== AxisLineRecord ==========

const unsigned AxisLineRecord::id = 0x1021;

class AxisLineRecord::Private
{
public:
};

AxisLineRecord::AxisLineRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AxisLineRecord::~AxisLineRecord()
{
    delete d;
}

AxisLineRecord::AxisLineRecord( const AxisLineRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AxisLineRecord& AxisLineRecord::operator=( const AxisLineRecord& record )
{
    *d = *record.d;
    return *this;
}

void AxisLineRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AxisLineRecord::dump( std::ostream& out ) const
{
    out << "AxisLine" << std::endl;
}

static Record* createAxisLineRecord(Workbook *book)
{
    return new AxisLineRecord(book);
}

// ========== CrtLinkRecord ==========

const unsigned CrtLinkRecord::id = 0x1022;

class CrtLinkRecord::Private
{
public:
};

CrtLinkRecord::CrtLinkRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CrtLinkRecord::~CrtLinkRecord()
{
    delete d;
}

CrtLinkRecord::CrtLinkRecord( const CrtLinkRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CrtLinkRecord& CrtLinkRecord::operator=( const CrtLinkRecord& record )
{
    *d = *record.d;
    return *this;
}

void CrtLinkRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CrtLinkRecord::dump( std::ostream& out ) const
{
    out << "CrtLink" << std::endl;
}

static Record* createCrtLinkRecord(Workbook *book)
{
    return new CrtLinkRecord(book);
}

// ========== DefaultTextRecord ==========

const unsigned DefaultTextRecord::id = 0x1024;

class DefaultTextRecord::Private
{
public:
    Identifier identifier;
};

DefaultTextRecord::DefaultTextRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

DefaultTextRecord::~DefaultTextRecord()
{
    delete d;
}

DefaultTextRecord::DefaultTextRecord( const DefaultTextRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DefaultTextRecord& DefaultTextRecord::operator=( const DefaultTextRecord& record )
{
    *d = *record.d;
    return *this;
}

UString DefaultTextRecord::identifierToString(Identifier identifier)
{
    switch (identifier) {
        case ShowPercent0_or_ShowValue0: return UString("ShowPercent0_or_ShowValue0");
        case ShowPercent1_or_ShowValue1: return UString("ShowPercent1_or_ShowValue1");
        case Scalable0: return UString("Scalable0");
        case Scalable1: return UString("Scalable1");
        default: return UString("Unknown: ") + UString::from(identifier);
    }
}

DefaultTextRecord::Identifier DefaultTextRecord::identifier() const
{
    return d->identifier;
}

void DefaultTextRecord::setIdentifier(Identifier identifier )
{
    d->identifier = identifier;
}

void DefaultTextRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setIdentifier(static_cast<Identifier>(readU16(data)));
}

void DefaultTextRecord::dump( std::ostream& out ) const
{
    out << "DefaultText" << std::endl;
    out << "         Identifier : " << identifierToString(identifier()) << std::endl;
}

static Record* createDefaultTextRecord(Workbook *book)
{
    return new DefaultTextRecord(book);
}

// ========== TextRecord ==========

const unsigned TextRecord::id = 0x1025;

class TextRecord::Private
{
public:
    unsigned at;
    unsigned dx;
    unsigned dy;
    bool fAutoColor;
    bool fAutoMode;
    bool fAutoText;
    bool fDeleted;
    bool fGenerated;
    bool fShowBubbleSizes;
    bool fShowKey;
    bool fShowLabel;
    bool fShowLabelAndPerc;
    bool fShowPercent;
    bool fShowValue;
    unsigned icvText;
    unsigned rgbText;
    unsigned unused1;
    unsigned unused2;
    unsigned vat;
    unsigned wBkgMode;
    unsigned x;
    unsigned y;
};

TextRecord::TextRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setAt(0);
    setDx(0);
    setDy(0);
    setFAutoColor(false);
    setFAutoMode(false);
    setFAutoText(false);
    setFDeleted(false);
    setFGenerated(false);
    setFShowBubbleSizes(false);
    setFShowKey(false);
    setFShowLabel(false);
    setFShowLabelAndPerc(false);
    setFShowPercent(false);
    setFShowValue(false);
    setIcvText(0);
    setRgbText(0);
    setUnused1(0);
    setUnused2(0);
    setVat(0);
    setWBkgMode(0);
    setX(0);
    setY(0);
}

TextRecord::~TextRecord()
{
    delete d;
}

TextRecord::TextRecord( const TextRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

TextRecord& TextRecord::operator=( const TextRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned TextRecord::at() const
{
    return d->at;
}

void TextRecord::setAt(unsigned at )
{
    d->at = at;
}

unsigned TextRecord::dx() const
{
    return d->dx;
}

void TextRecord::setDx(unsigned dx )
{
    d->dx = dx;
}

unsigned TextRecord::dy() const
{
    return d->dy;
}

void TextRecord::setDy(unsigned dy )
{
    d->dy = dy;
}

bool TextRecord::isFAutoColor() const
{
    return d->fAutoColor;
}

void TextRecord::setFAutoColor(bool fAutoColor )
{
    d->fAutoColor = fAutoColor;
}

bool TextRecord::isFAutoMode() const
{
    return d->fAutoMode;
}

void TextRecord::setFAutoMode(bool fAutoMode )
{
    d->fAutoMode = fAutoMode;
}

bool TextRecord::isFAutoText() const
{
    return d->fAutoText;
}

void TextRecord::setFAutoText(bool fAutoText )
{
    d->fAutoText = fAutoText;
}

bool TextRecord::isFDeleted() const
{
    return d->fDeleted;
}

void TextRecord::setFDeleted(bool fDeleted )
{
    d->fDeleted = fDeleted;
}

bool TextRecord::isFGenerated() const
{
    return d->fGenerated;
}

void TextRecord::setFGenerated(bool fGenerated )
{
    d->fGenerated = fGenerated;
}

bool TextRecord::isFShowBubbleSizes() const
{
    return d->fShowBubbleSizes;
}

void TextRecord::setFShowBubbleSizes(bool fShowBubbleSizes )
{
    d->fShowBubbleSizes = fShowBubbleSizes;
}

bool TextRecord::isFShowKey() const
{
    return d->fShowKey;
}

void TextRecord::setFShowKey(bool fShowKey )
{
    d->fShowKey = fShowKey;
}

bool TextRecord::isFShowLabel() const
{
    return d->fShowLabel;
}

void TextRecord::setFShowLabel(bool fShowLabel )
{
    d->fShowLabel = fShowLabel;
}

bool TextRecord::isFShowLabelAndPerc() const
{
    return d->fShowLabelAndPerc;
}

void TextRecord::setFShowLabelAndPerc(bool fShowLabelAndPerc )
{
    d->fShowLabelAndPerc = fShowLabelAndPerc;
}

bool TextRecord::isFShowPercent() const
{
    return d->fShowPercent;
}

void TextRecord::setFShowPercent(bool fShowPercent )
{
    d->fShowPercent = fShowPercent;
}

bool TextRecord::isFShowValue() const
{
    return d->fShowValue;
}

void TextRecord::setFShowValue(bool fShowValue )
{
    d->fShowValue = fShowValue;
}

unsigned TextRecord::icvText() const
{
    return d->icvText;
}

void TextRecord::setIcvText(unsigned icvText )
{
    d->icvText = icvText;
}

unsigned TextRecord::rgbText() const
{
    return d->rgbText;
}

void TextRecord::setRgbText(unsigned rgbText )
{
    d->rgbText = rgbText;
}

unsigned TextRecord::unused1() const
{
    return d->unused1;
}

void TextRecord::setUnused1(unsigned unused1 )
{
    d->unused1 = unused1;
}

unsigned TextRecord::unused2() const
{
    return d->unused2;
}

void TextRecord::setUnused2(unsigned unused2 )
{
    d->unused2 = unused2;
}

unsigned TextRecord::vat() const
{
    return d->vat;
}

void TextRecord::setVat(unsigned vat )
{
    d->vat = vat;
}

unsigned TextRecord::wBkgMode() const
{
    return d->wBkgMode;
}

void TextRecord::setWBkgMode(unsigned wBkgMode )
{
    d->wBkgMode = wBkgMode;
}

unsigned TextRecord::x() const
{
    return d->x;
}

void TextRecord::setX(unsigned x )
{
    d->x = x;
}

unsigned TextRecord::y() const
{
    return d->y;
}

void TextRecord::setY(unsigned y )
{
    d->y = y;
}

void TextRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 28) {
        setIsValid(false);
        return;
    }
    setAt(readU8(data));
    setVat(readU8(data + 1));
    setWBkgMode(readU16(data + 2));
    setRgbText(readU32(data + 4));
    setX(readU32(data + 8));
    setY(readU32(data + 12));
    setDx(readU32(data + 16));
    setDy(readU32(data + 20));
    setFAutoColor(((readU8(data + 24)) & 0x1) != 0);
    setFShowKey(((readU8(data + 24) >> 1) & 0x1) != 0);
    setFShowValue(((readU8(data + 24) >> 2) & 0x1) != 0);
    setUnused1(((readU8(data + 24) >> 3) & 0x1));
    setFAutoText(((readU8(data + 24) >> 4) & 0x1) != 0);
    setFGenerated(((readU8(data + 24) >> 5) & 0x1) != 0);
    setFDeleted(((readU8(data + 24) >> 6) & 0x1) != 0);
    setFAutoMode(((readU8(data + 24) >> 7) & 0x1) != 0);
    setUnused2(((readU8(data + 25)) & 0x7));
    setFShowLabelAndPerc(((readU8(data + 25) >> 3) & 0x1) != 0);
    setFShowPercent(((readU8(data + 25) >> 4) & 0x1) != 0);
    setFShowBubbleSizes(((readU8(data + 25) >> 5) & 0x1) != 0);
    setFShowLabel(((readU8(data + 25) >> 6) & 0x1) != 0);
    setIcvText(readU16(data + 26));
}

void TextRecord::dump( std::ostream& out ) const
{
    out << "Text" << std::endl;
    out << "                 At : " << at() << std::endl;
    out << "                Vat : " << vat() << std::endl;
    out << "           WBkgMode : " << wBkgMode() << std::endl;
    out << "            RgbText : " << rgbText() << std::endl;
    out << "                  X : " << x() << std::endl;
    out << "                  Y : " << y() << std::endl;
    out << "                 Dx : " << dx() << std::endl;
    out << "                 Dy : " << dy() << std::endl;
    out << "         FAutoColor : " << isFAutoColor() << std::endl;
    out << "           FShowKey : " << isFShowKey() << std::endl;
    out << "         FShowValue : " << isFShowValue() << std::endl;
    out << "            Unused1 : " << unused1() << std::endl;
    out << "          FAutoText : " << isFAutoText() << std::endl;
    out << "         FGenerated : " << isFGenerated() << std::endl;
    out << "           FDeleted : " << isFDeleted() << std::endl;
    out << "          FAutoMode : " << isFAutoMode() << std::endl;
    out << "            Unused2 : " << unused2() << std::endl;
    out << "  FShowLabelAndPerc : " << isFShowLabelAndPerc() << std::endl;
    out << "       FShowPercent : " << isFShowPercent() << std::endl;
    out << "   FShowBubbleSizes : " << isFShowBubbleSizes() << std::endl;
    out << "         FShowLabel : " << isFShowLabel() << std::endl;
    out << "            IcvText : " << icvText() << std::endl;
}

static Record* createTextRecord(Workbook *book)
{
    return new TextRecord(book);
}

// ========== FontXRecord ==========

const unsigned FontXRecord::id = 0x1026;

class FontXRecord::Private
{
public:
};

FontXRecord::FontXRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

FontXRecord::~FontXRecord()
{
    delete d;
}

FontXRecord::FontXRecord( const FontXRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

FontXRecord& FontXRecord::operator=( const FontXRecord& record )
{
    *d = *record.d;
    return *this;
}

void FontXRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void FontXRecord::dump( std::ostream& out ) const
{
    out << "FontX" << std::endl;
}

static Record* createFontXRecord(Workbook *book)
{
    return new FontXRecord(book);
}

// ========== ObjectLinkRecord ==========

const unsigned ObjectLinkRecord::id = 0x1027;

class ObjectLinkRecord::Private
{
public:
    WLinkObj wLinkObj;
    unsigned wLinkVar1;
    unsigned wLinkVar2;
};

ObjectLinkRecord::ObjectLinkRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setWLinkVar1(0);
    setWLinkVar2(0);
}

ObjectLinkRecord::~ObjectLinkRecord()
{
    delete d;
}

ObjectLinkRecord::ObjectLinkRecord( const ObjectLinkRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ObjectLinkRecord& ObjectLinkRecord::operator=( const ObjectLinkRecord& record )
{
    *d = *record.d;
    return *this;
}

UString ObjectLinkRecord::wLinkObjToString(WLinkObj wLinkObj)
{
    switch (wLinkObj) {
        case EntireChart: return UString("EntireChart");
        case ValueOrVerticalAxis: return UString("ValueOrVerticalAxis");
        case CategoryOrHorizontalAxis: return UString("CategoryOrHorizontalAxis");
        case SeriesOrDatapoints: return UString("SeriesOrDatapoints");
        case SeriesAxis: return UString("SeriesAxis");
        case DisplayUnitsLabelsOfAxis: return UString("DisplayUnitsLabelsOfAxis");
        default: return UString("Unknown: ") + UString::from(wLinkObj);
    }
}

ObjectLinkRecord::WLinkObj ObjectLinkRecord::wLinkObj() const
{
    return d->wLinkObj;
}

void ObjectLinkRecord::setWLinkObj(WLinkObj wLinkObj )
{
    d->wLinkObj = wLinkObj;
}

unsigned ObjectLinkRecord::wLinkVar1() const
{
    return d->wLinkVar1;
}

void ObjectLinkRecord::setWLinkVar1(unsigned wLinkVar1 )
{
    d->wLinkVar1 = wLinkVar1;
}

unsigned ObjectLinkRecord::wLinkVar2() const
{
    return d->wLinkVar2;
}

void ObjectLinkRecord::setWLinkVar2(unsigned wLinkVar2 )
{
    d->wLinkVar2 = wLinkVar2;
}

void ObjectLinkRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 6) {
        setIsValid(false);
        return;
    }
    setWLinkObj(static_cast<WLinkObj>(readU16(data)));
    setWLinkVar1(readU16(data + 2));
    setWLinkVar2(readU16(data + 4));
}

void ObjectLinkRecord::dump( std::ostream& out ) const
{
    out << "ObjectLink" << std::endl;
    out << "           WLinkObj : " << wLinkObjToString(wLinkObj()) << std::endl;
    out << "          WLinkVar1 : " << wLinkVar1() << std::endl;
    out << "          WLinkVar2 : " << wLinkVar2() << std::endl;
}

static Record* createObjectLinkRecord(Workbook *book)
{
    return new ObjectLinkRecord(book);
}

// ========== PlotAreaRecord ==========

const unsigned PlotAreaRecord::id = 0x1035;

class PlotAreaRecord::Private
{
public:
};

PlotAreaRecord::PlotAreaRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PlotAreaRecord::~PlotAreaRecord()
{
    delete d;
}

PlotAreaRecord::PlotAreaRecord( const PlotAreaRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PlotAreaRecord& PlotAreaRecord::operator=( const PlotAreaRecord& record )
{
    *d = *record.d;
    return *this;
}

void PlotAreaRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PlotAreaRecord::dump( std::ostream& out ) const
{
    out << "PlotArea" << std::endl;
}

static Record* createPlotAreaRecord(Workbook *book)
{
    return new PlotAreaRecord(book);
}

// ========== MsoDrawingSelectionRecord ==========

const unsigned MsoDrawingSelectionRecord::id = 0x00ed;

class MsoDrawingSelectionRecord::Private
{
public:
};

MsoDrawingSelectionRecord::MsoDrawingSelectionRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

MsoDrawingSelectionRecord::~MsoDrawingSelectionRecord()
{
    delete d;
}

MsoDrawingSelectionRecord::MsoDrawingSelectionRecord( const MsoDrawingSelectionRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

MsoDrawingSelectionRecord& MsoDrawingSelectionRecord::operator=( const MsoDrawingSelectionRecord& record )
{
    *d = *record.d;
    return *this;
}

void MsoDrawingSelectionRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void MsoDrawingSelectionRecord::dump( std::ostream& out ) const
{
    out << "MsoDrawingSelection" << std::endl;
}

static Record* createMsoDrawingSelectionRecord(Workbook *book)
{
    return new MsoDrawingSelectionRecord(book);
}

// ========== Chart3dRecord ==========

const unsigned Chart3dRecord::id = 0x103a;

class Chart3dRecord::Private
{
public:
    int anElev;
    int anRot;
    bool f3DScaling;
    bool fCluster;
    bool fNotPieChart;
    bool fPerspective;
    bool fWalls2D;
    int pcDepth;
    int pcDist;
    unsigned pcGap;
    unsigned pcHeight;
};

Chart3dRecord::Chart3dRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setAnElev(0);
    setAnRot(0);
    setF3DScaling(false);
    setFCluster(false);
    setFNotPieChart(false);
    setFPerspective(false);
    setFWalls2D(false);
    setPcDepth(0);
    setPcDist(0);
    setPcGap(0);
    setPcHeight(0);
}

Chart3dRecord::~Chart3dRecord()
{
    delete d;
}

Chart3dRecord::Chart3dRecord( const Chart3dRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Chart3dRecord& Chart3dRecord::operator=( const Chart3dRecord& record )
{
    *d = *record.d;
    return *this;
}

int Chart3dRecord::anElev() const
{
    return d->anElev;
}

void Chart3dRecord::setAnElev(int anElev )
{
    d->anElev = anElev;
}

int Chart3dRecord::anRot() const
{
    return d->anRot;
}

void Chart3dRecord::setAnRot(int anRot )
{
    d->anRot = anRot;
}

bool Chart3dRecord::isF3DScaling() const
{
    return d->f3DScaling;
}

void Chart3dRecord::setF3DScaling(bool f3DScaling )
{
    d->f3DScaling = f3DScaling;
}

bool Chart3dRecord::isFCluster() const
{
    return d->fCluster;
}

void Chart3dRecord::setFCluster(bool fCluster )
{
    d->fCluster = fCluster;
}

bool Chart3dRecord::isFNotPieChart() const
{
    return d->fNotPieChart;
}

void Chart3dRecord::setFNotPieChart(bool fNotPieChart )
{
    d->fNotPieChart = fNotPieChart;
}

bool Chart3dRecord::isFPerspective() const
{
    return d->fPerspective;
}

void Chart3dRecord::setFPerspective(bool fPerspective )
{
    d->fPerspective = fPerspective;
}

bool Chart3dRecord::isFWalls2D() const
{
    return d->fWalls2D;
}

void Chart3dRecord::setFWalls2D(bool fWalls2D )
{
    d->fWalls2D = fWalls2D;
}

int Chart3dRecord::pcDepth() const
{
    return d->pcDepth;
}

void Chart3dRecord::setPcDepth(int pcDepth )
{
    d->pcDepth = pcDepth;
}

int Chart3dRecord::pcDist() const
{
    return d->pcDist;
}

void Chart3dRecord::setPcDist(int pcDist )
{
    d->pcDist = pcDist;
}

unsigned Chart3dRecord::pcGap() const
{
    return d->pcGap;
}

void Chart3dRecord::setPcGap(unsigned pcGap )
{
    d->pcGap = pcGap;
}

unsigned Chart3dRecord::pcHeight() const
{
    return d->pcHeight;
}

void Chart3dRecord::setPcHeight(unsigned pcHeight )
{
    d->pcHeight = pcHeight;
}

void Chart3dRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 14) {
        setIsValid(false);
        return;
    }
    setAnRot(readS16(data));
    setAnElev(readS16(data + 2));
    setPcDist(readS16(data + 4));
    setPcHeight(readU16(data + 6));
    setPcDepth(readS16(data + 8));
    setPcGap(readU16(data + 10));
    setFPerspective(((readU8(data + 12)) & 0x1) != 0);
    setFCluster(((readU8(data + 12) >> 1) & 0x1) != 0);
    setF3DScaling(((readU8(data + 12) >> 2) & 0x1) != 0);
    setFNotPieChart(((readU8(data + 12) >> 4) & 0x1) != 0);
    setFWalls2D(((readU8(data + 12) >> 5) & 0x1) != 0);
}

void Chart3dRecord::dump( std::ostream& out ) const
{
    out << "Chart3d" << std::endl;
    out << "              AnRot : " << anRot() << std::endl;
    out << "             AnElev : " << anElev() << std::endl;
    out << "             PcDist : " << pcDist() << std::endl;
    out << "           PcHeight : " << pcHeight() << std::endl;
    out << "            PcDepth : " << pcDepth() << std::endl;
    out << "              PcGap : " << pcGap() << std::endl;
    out << "       FPerspective : " << isFPerspective() << std::endl;
    out << "           FCluster : " << isFCluster() << std::endl;
    out << "         F3DScaling : " << isF3DScaling() << std::endl;
    out << "       FNotPieChart : " << isFNotPieChart() << std::endl;
    out << "           FWalls2D : " << isFWalls2D() << std::endl;
}

static Record* createChart3dRecord(Workbook *book)
{
    return new Chart3dRecord(book);
}

// ========== PicFRecord ==========

const unsigned PicFRecord::id = 0x103c;

class PicFRecord::Private
{
public:
};

PicFRecord::PicFRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PicFRecord::~PicFRecord()
{
    delete d;
}

PicFRecord::PicFRecord( const PicFRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PicFRecord& PicFRecord::operator=( const PicFRecord& record )
{
    *d = *record.d;
    return *this;
}

void PicFRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PicFRecord::dump( std::ostream& out ) const
{
    out << "PicF" << std::endl;
}

static Record* createPicFRecord(Workbook *book)
{
    return new PicFRecord(book);
}

// ========== DropBarRecord ==========

const unsigned DropBarRecord::id = 0x103d;

class DropBarRecord::Private
{
public:
};

DropBarRecord::DropBarRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

DropBarRecord::~DropBarRecord()
{
    delete d;
}

DropBarRecord::DropBarRecord( const DropBarRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DropBarRecord& DropBarRecord::operator=( const DropBarRecord& record )
{
    *d = *record.d;
    return *this;
}

void DropBarRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void DropBarRecord::dump( std::ostream& out ) const
{
    out << "DropBar" << std::endl;
}

static Record* createDropBarRecord(Workbook *book)
{
    return new DropBarRecord(book);
}

// ========== RadarRecord ==========

const unsigned RadarRecord::id = 0x103e;

class RadarRecord::Private
{
public:
};

RadarRecord::RadarRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

RadarRecord::~RadarRecord()
{
    delete d;
}

RadarRecord::RadarRecord( const RadarRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

RadarRecord& RadarRecord::operator=( const RadarRecord& record )
{
    *d = *record.d;
    return *this;
}

void RadarRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void RadarRecord::dump( std::ostream& out ) const
{
    out << "Radar" << std::endl;
}

static Record* createRadarRecord(Workbook *book)
{
    return new RadarRecord(book);
}

// ========== SurfRecord ==========

const unsigned SurfRecord::id = 0x103f;

class SurfRecord::Private
{
public:
};

SurfRecord::SurfRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SurfRecord::~SurfRecord()
{
    delete d;
}

SurfRecord::SurfRecord( const SurfRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SurfRecord& SurfRecord::operator=( const SurfRecord& record )
{
    *d = *record.d;
    return *this;
}

void SurfRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void SurfRecord::dump( std::ostream& out ) const
{
    out << "Surf" << std::endl;
}

static Record* createSurfRecord(Workbook *book)
{
    return new SurfRecord(book);
}

// ========== RadarAreaRecord ==========

const unsigned RadarAreaRecord::id = 0x1040;

class RadarAreaRecord::Private
{
public:
};

RadarAreaRecord::RadarAreaRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

RadarAreaRecord::~RadarAreaRecord()
{
    delete d;
}

RadarAreaRecord::RadarAreaRecord( const RadarAreaRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

RadarAreaRecord& RadarAreaRecord::operator=( const RadarAreaRecord& record )
{
    *d = *record.d;
    return *this;
}

void RadarAreaRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void RadarAreaRecord::dump( std::ostream& out ) const
{
    out << "RadarArea" << std::endl;
}

static Record* createRadarAreaRecord(Workbook *book)
{
    return new RadarAreaRecord(book);
}

// ========== AxisParentRecord ==========

const unsigned AxisParentRecord::id = 0x1041;

class AxisParentRecord::Private
{
public:
};

AxisParentRecord::AxisParentRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AxisParentRecord::~AxisParentRecord()
{
    delete d;
}

AxisParentRecord::AxisParentRecord( const AxisParentRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AxisParentRecord& AxisParentRecord::operator=( const AxisParentRecord& record )
{
    *d = *record.d;
    return *this;
}

void AxisParentRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AxisParentRecord::dump( std::ostream& out ) const
{
    out << "AxisParent" << std::endl;
}

static Record* createAxisParentRecord(Workbook *book)
{
    return new AxisParentRecord(book);
}

// ========== LegendExceptionRecord ==========

const unsigned LegendExceptionRecord::id = 0x1043;

class LegendExceptionRecord::Private
{
public:
};

LegendExceptionRecord::LegendExceptionRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

LegendExceptionRecord::~LegendExceptionRecord()
{
    delete d;
}

LegendExceptionRecord::LegendExceptionRecord( const LegendExceptionRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LegendExceptionRecord& LegendExceptionRecord::operator=( const LegendExceptionRecord& record )
{
    *d = *record.d;
    return *this;
}

void LegendExceptionRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void LegendExceptionRecord::dump( std::ostream& out ) const
{
    out << "LegendException" << std::endl;
}

static Record* createLegendExceptionRecord(Workbook *book)
{
    return new LegendExceptionRecord(book);
}

// ========== ShtPropsRecord ==========

const unsigned ShtPropsRecord::id = 0x1044;

class ShtPropsRecord::Private
{
public:
    bool fAlwaysAutoPlotArea;
    bool fManPlotArea;
    bool fManSerAlloc;
    bool fNotSizeWIth;
    bool fPlotVisOnly;
    unsigned mdBlank;
};

ShtPropsRecord::ShtPropsRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFAlwaysAutoPlotArea(false);
    setFManPlotArea(false);
    setFManSerAlloc(false);
    setFNotSizeWIth(false);
    setFPlotVisOnly(false);
    setMdBlank(0);
}

ShtPropsRecord::~ShtPropsRecord()
{
    delete d;
}

ShtPropsRecord::ShtPropsRecord( const ShtPropsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ShtPropsRecord& ShtPropsRecord::operator=( const ShtPropsRecord& record )
{
    *d = *record.d;
    return *this;
}

bool ShtPropsRecord::isFAlwaysAutoPlotArea() const
{
    return d->fAlwaysAutoPlotArea;
}

void ShtPropsRecord::setFAlwaysAutoPlotArea(bool fAlwaysAutoPlotArea )
{
    d->fAlwaysAutoPlotArea = fAlwaysAutoPlotArea;
}

bool ShtPropsRecord::isFManPlotArea() const
{
    return d->fManPlotArea;
}

void ShtPropsRecord::setFManPlotArea(bool fManPlotArea )
{
    d->fManPlotArea = fManPlotArea;
}

bool ShtPropsRecord::isFManSerAlloc() const
{
    return d->fManSerAlloc;
}

void ShtPropsRecord::setFManSerAlloc(bool fManSerAlloc )
{
    d->fManSerAlloc = fManSerAlloc;
}

bool ShtPropsRecord::isFNotSizeWIth() const
{
    return d->fNotSizeWIth;
}

void ShtPropsRecord::setFNotSizeWIth(bool fNotSizeWIth )
{
    d->fNotSizeWIth = fNotSizeWIth;
}

bool ShtPropsRecord::isFPlotVisOnly() const
{
    return d->fPlotVisOnly;
}

void ShtPropsRecord::setFPlotVisOnly(bool fPlotVisOnly )
{
    d->fPlotVisOnly = fPlotVisOnly;
}

unsigned ShtPropsRecord::mdBlank() const
{
    return d->mdBlank;
}

void ShtPropsRecord::setMdBlank(unsigned mdBlank )
{
    d->mdBlank = mdBlank;
}

void ShtPropsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 4) {
        setIsValid(false);
        return;
    }
    setFManSerAlloc(((readU8(data)) & 0x1) != 0);
    setFPlotVisOnly(((readU8(data) >> 1) & 0x1) != 0);
    setFNotSizeWIth(((readU8(data) >> 2) & 0x1) != 0);
    setFManPlotArea(((readU8(data) >> 3) & 0x1) != 0);
    setFAlwaysAutoPlotArea(((readU8(data) >> 4) & 0x1) != 0);
    setMdBlank(readU8(data + 2));
}

void ShtPropsRecord::dump( std::ostream& out ) const
{
    out << "ShtProps" << std::endl;
    out << "       FManSerAlloc : " << isFManSerAlloc() << std::endl;
    out << "       FPlotVisOnly : " << isFPlotVisOnly() << std::endl;
    out << "       FNotSizeWIth : " << isFNotSizeWIth() << std::endl;
    out << "       FManPlotArea : " << isFManPlotArea() << std::endl;
    out << "FAlwaysAutoPlotArea : " << isFAlwaysAutoPlotArea() << std::endl;
    out << "            MdBlank : " << mdBlank() << std::endl;
}

static Record* createShtPropsRecord(Workbook *book)
{
    return new ShtPropsRecord(book);
}

// ========== SerToCrtRecord ==========

const unsigned SerToCrtRecord::id = 0x1045;

class SerToCrtRecord::Private
{
public:
    unsigned identifier;
};

SerToCrtRecord::SerToCrtRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setIdentifier(0);
}

SerToCrtRecord::~SerToCrtRecord()
{
    delete d;
}

SerToCrtRecord::SerToCrtRecord( const SerToCrtRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SerToCrtRecord& SerToCrtRecord::operator=( const SerToCrtRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned SerToCrtRecord::identifier() const
{
    return d->identifier;
}

void SerToCrtRecord::setIdentifier(unsigned identifier )
{
    d->identifier = identifier;
}

void SerToCrtRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setIdentifier(readU16(data));
}

void SerToCrtRecord::dump( std::ostream& out ) const
{
    out << "SerToCrt" << std::endl;
    out << "         Identifier : " << identifier() << std::endl;
}

static Record* createSerToCrtRecord(Workbook *book)
{
    return new SerToCrtRecord(book);
}

// ========== AxesUsedRecord ==========

const unsigned AxesUsedRecord::id = 0x1046;

class AxesUsedRecord::Private
{
public:
};

AxesUsedRecord::AxesUsedRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AxesUsedRecord::~AxesUsedRecord()
{
    delete d;
}

AxesUsedRecord::AxesUsedRecord( const AxesUsedRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AxesUsedRecord& AxesUsedRecord::operator=( const AxesUsedRecord& record )
{
    *d = *record.d;
    return *this;
}

void AxesUsedRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AxesUsedRecord::dump( std::ostream& out ) const
{
    out << "AxesUsed" << std::endl;
}

static Record* createAxesUsedRecord(Workbook *book)
{
    return new AxesUsedRecord(book);
}

// ========== SBaseRefRecord ==========

const unsigned SBaseRefRecord::id = 0x1048;

class SBaseRefRecord::Private
{
public:
};

SBaseRefRecord::SBaseRefRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SBaseRefRecord::~SBaseRefRecord()
{
    delete d;
}

SBaseRefRecord::SBaseRefRecord( const SBaseRefRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SBaseRefRecord& SBaseRefRecord::operator=( const SBaseRefRecord& record )
{
    *d = *record.d;
    return *this;
}

void SBaseRefRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void SBaseRefRecord::dump( std::ostream& out ) const
{
    out << "SBaseRef" << std::endl;
}

static Record* createSBaseRefRecord(Workbook *book)
{
    return new SBaseRefRecord(book);
}

// ========== SerParentRecord ==========

const unsigned SerParentRecord::id = 0x104a;

class SerParentRecord::Private
{
public:
};

SerParentRecord::SerParentRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SerParentRecord::~SerParentRecord()
{
    delete d;
}

SerParentRecord::SerParentRecord( const SerParentRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SerParentRecord& SerParentRecord::operator=( const SerParentRecord& record )
{
    *d = *record.d;
    return *this;
}

void SerParentRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void SerParentRecord::dump( std::ostream& out ) const
{
    out << "SerParent" << std::endl;
}

static Record* createSerParentRecord(Workbook *book)
{
    return new SerParentRecord(book);
}

// ========== SerAuxTrendRecord ==========

const unsigned SerAuxTrendRecord::id = 0x104b;

class SerAuxTrendRecord::Private
{
public:
};

SerAuxTrendRecord::SerAuxTrendRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SerAuxTrendRecord::~SerAuxTrendRecord()
{
    delete d;
}

SerAuxTrendRecord::SerAuxTrendRecord( const SerAuxTrendRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SerAuxTrendRecord& SerAuxTrendRecord::operator=( const SerAuxTrendRecord& record )
{
    *d = *record.d;
    return *this;
}

void SerAuxTrendRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void SerAuxTrendRecord::dump( std::ostream& out ) const
{
    out << "SerAuxTrend" << std::endl;
}

static Record* createSerAuxTrendRecord(Workbook *book)
{
    return new SerAuxTrendRecord(book);
}

// ========== IFmtRecordRecord ==========

const unsigned IFmtRecordRecord::id = 0x104e;

class IFmtRecordRecord::Private
{
public:
};

IFmtRecordRecord::IFmtRecordRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

IFmtRecordRecord::~IFmtRecordRecord()
{
    delete d;
}

IFmtRecordRecord::IFmtRecordRecord( const IFmtRecordRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

IFmtRecordRecord& IFmtRecordRecord::operator=( const IFmtRecordRecord& record )
{
    *d = *record.d;
    return *this;
}

void IFmtRecordRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void IFmtRecordRecord::dump( std::ostream& out ) const
{
    out << "IFmtRecord" << std::endl;
}

static Record* createIFmtRecordRecord(Workbook *book)
{
    return new IFmtRecordRecord(book);
}

// ========== PosRecord ==========

const unsigned PosRecord::id = 0x104f;

class PosRecord::Private
{
public:
    unsigned mdBotRt;
    unsigned mdTopLt;
    unsigned unused1;
    unsigned unused2;
    unsigned unused3;
    unsigned unused4;
    unsigned x1;
    unsigned x2;
    unsigned y1;
    unsigned y2;
};

PosRecord::PosRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setMdBotRt(0);
    setMdTopLt(0);
    setUnused1(0);
    setUnused2(0);
    setUnused3(0);
    setUnused4(0);
    setX1(0);
    setX2(0);
    setY1(0);
    setY2(0);
}

PosRecord::~PosRecord()
{
    delete d;
}

PosRecord::PosRecord( const PosRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PosRecord& PosRecord::operator=( const PosRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned PosRecord::mdBotRt() const
{
    return d->mdBotRt;
}

void PosRecord::setMdBotRt(unsigned mdBotRt )
{
    d->mdBotRt = mdBotRt;
}

unsigned PosRecord::mdTopLt() const
{
    return d->mdTopLt;
}

void PosRecord::setMdTopLt(unsigned mdTopLt )
{
    d->mdTopLt = mdTopLt;
}

unsigned PosRecord::unused1() const
{
    return d->unused1;
}

void PosRecord::setUnused1(unsigned unused1 )
{
    d->unused1 = unused1;
}

unsigned PosRecord::unused2() const
{
    return d->unused2;
}

void PosRecord::setUnused2(unsigned unused2 )
{
    d->unused2 = unused2;
}

unsigned PosRecord::unused3() const
{
    return d->unused3;
}

void PosRecord::setUnused3(unsigned unused3 )
{
    d->unused3 = unused3;
}

unsigned PosRecord::unused4() const
{
    return d->unused4;
}

void PosRecord::setUnused4(unsigned unused4 )
{
    d->unused4 = unused4;
}

unsigned PosRecord::x1() const
{
    return d->x1;
}

void PosRecord::setX1(unsigned x1 )
{
    d->x1 = x1;
}

unsigned PosRecord::x2() const
{
    return d->x2;
}

void PosRecord::setX2(unsigned x2 )
{
    d->x2 = x2;
}

unsigned PosRecord::y1() const
{
    return d->y1;
}

void PosRecord::setY1(unsigned y1 )
{
    d->y1 = y1;
}

unsigned PosRecord::y2() const
{
    return d->y2;
}

void PosRecord::setY2(unsigned y2 )
{
    d->y2 = y2;
}

void PosRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 20) {
        setIsValid(false);
        return;
    }
    setMdTopLt(readU16(data));
    setMdBotRt(readU16(data + 2));
    setX1(readU16(data + 4));
    setUnused1(readU16(data + 6));
    setY1(readU16(data + 8));
    setUnused2(readU16(data + 10));
    setX2(readU16(data + 12));
    setUnused3(readU16(data + 14));
    setY2(readU16(data + 16));
    setUnused4(readU16(data + 18));
}

void PosRecord::dump( std::ostream& out ) const
{
    out << "Pos" << std::endl;
    out << "            MdTopLt : " << mdTopLt() << std::endl;
    out << "            MdBotRt : " << mdBotRt() << std::endl;
    out << "                 X1 : " << x1() << std::endl;
    out << "            Unused1 : " << unused1() << std::endl;
    out << "                 Y1 : " << y1() << std::endl;
    out << "            Unused2 : " << unused2() << std::endl;
    out << "                 X2 : " << x2() << std::endl;
    out << "            Unused3 : " << unused3() << std::endl;
    out << "                 Y2 : " << y2() << std::endl;
    out << "            Unused4 : " << unused4() << std::endl;
}

static Record* createPosRecord(Workbook *book)
{
    return new PosRecord(book);
}

// ========== AIRunsRecord ==========

const unsigned AIRunsRecord::id = 0x1050;

class AIRunsRecord::Private
{
public:
};

AIRunsRecord::AIRunsRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AIRunsRecord::~AIRunsRecord()
{
    delete d;
}

AIRunsRecord::AIRunsRecord( const AIRunsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AIRunsRecord& AIRunsRecord::operator=( const AIRunsRecord& record )
{
    *d = *record.d;
    return *this;
}

void AIRunsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AIRunsRecord::dump( std::ostream& out ) const
{
    out << "AIRuns" << std::endl;
}

static Record* createAIRunsRecord(Workbook *book)
{
    return new AIRunsRecord(book);
}

// ========== SerAuxErrBarRecord ==========

const unsigned SerAuxErrBarRecord::id = 0x105b;

class SerAuxErrBarRecord::Private
{
public:
};

SerAuxErrBarRecord::SerAuxErrBarRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SerAuxErrBarRecord::~SerAuxErrBarRecord()
{
    delete d;
}

SerAuxErrBarRecord::SerAuxErrBarRecord( const SerAuxErrBarRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SerAuxErrBarRecord& SerAuxErrBarRecord::operator=( const SerAuxErrBarRecord& record )
{
    *d = *record.d;
    return *this;
}

void SerAuxErrBarRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void SerAuxErrBarRecord::dump( std::ostream& out ) const
{
    out << "SerAuxErrBar" << std::endl;
}

static Record* createSerAuxErrBarRecord(Workbook *book)
{
    return new SerAuxErrBarRecord(book);
}

// ========== ClrtClientRecord ==========

const unsigned ClrtClientRecord::id = 0x105c;

class ClrtClientRecord::Private
{
public:
};

ClrtClientRecord::ClrtClientRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ClrtClientRecord::~ClrtClientRecord()
{
    delete d;
}

ClrtClientRecord::ClrtClientRecord( const ClrtClientRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ClrtClientRecord& ClrtClientRecord::operator=( const ClrtClientRecord& record )
{
    *d = *record.d;
    return *this;
}

void ClrtClientRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ClrtClientRecord::dump( std::ostream& out ) const
{
    out << "ClrtClient" << std::endl;
}

static Record* createClrtClientRecord(Workbook *book)
{
    return new ClrtClientRecord(book);
}

// ========== SerFmtRecord ==========

const unsigned SerFmtRecord::id = 0x105d;

class SerFmtRecord::Private
{
public:
};

SerFmtRecord::SerFmtRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SerFmtRecord::~SerFmtRecord()
{
    delete d;
}

SerFmtRecord::SerFmtRecord( const SerFmtRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SerFmtRecord& SerFmtRecord::operator=( const SerFmtRecord& record )
{
    *d = *record.d;
    return *this;
}

void SerFmtRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void SerFmtRecord::dump( std::ostream& out ) const
{
    out << "SerFmt" << std::endl;
}

static Record* createSerFmtRecord(Workbook *book)
{
    return new SerFmtRecord(book);
}

// ========== Chart3DBarShapeRecord ==========

const unsigned Chart3DBarShapeRecord::id = 0x105f;

class Chart3DBarShapeRecord::Private
{
public:
    unsigned riser;
    unsigned taper;
};

Chart3DBarShapeRecord::Chart3DBarShapeRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setRiser(0);
    setTaper(0);
}

Chart3DBarShapeRecord::~Chart3DBarShapeRecord()
{
    delete d;
}

Chart3DBarShapeRecord::Chart3DBarShapeRecord( const Chart3DBarShapeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Chart3DBarShapeRecord& Chart3DBarShapeRecord::operator=( const Chart3DBarShapeRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned Chart3DBarShapeRecord::riser() const
{
    return d->riser;
}

void Chart3DBarShapeRecord::setRiser(unsigned riser )
{
    d->riser = riser;
}

unsigned Chart3DBarShapeRecord::taper() const
{
    return d->taper;
}

void Chart3DBarShapeRecord::setTaper(unsigned taper )
{
    d->taper = taper;
}

void Chart3DBarShapeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setRiser(readU8(data));
    setTaper(readU8(data + 1));
}

void Chart3DBarShapeRecord::dump( std::ostream& out ) const
{
    out << "Chart3DBarShape" << std::endl;
    out << "              Riser : " << riser() << std::endl;
    out << "              Taper : " << taper() << std::endl;
}

static Record* createChart3DBarShapeRecord(Workbook *book)
{
    return new Chart3DBarShapeRecord(book);
}

// ========== FbiRecord ==========

const unsigned FbiRecord::id = 0x1060;

class FbiRecord::Private
{
public:
};

FbiRecord::FbiRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

FbiRecord::~FbiRecord()
{
    delete d;
}

FbiRecord::FbiRecord( const FbiRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

FbiRecord& FbiRecord::operator=( const FbiRecord& record )
{
    *d = *record.d;
    return *this;
}

void FbiRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void FbiRecord::dump( std::ostream& out ) const
{
    out << "Fbi" << std::endl;
}

static Record* createFbiRecord(Workbook *book)
{
    return new FbiRecord(book);
}

// ========== BopPopRecord ==========

const unsigned BopPopRecord::id = 0x1061;

class BopPopRecord::Private
{
public:
};

BopPopRecord::BopPopRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BopPopRecord::~BopPopRecord()
{
    delete d;
}

BopPopRecord::BopPopRecord( const BopPopRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BopPopRecord& BopPopRecord::operator=( const BopPopRecord& record )
{
    *d = *record.d;
    return *this;
}

void BopPopRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void BopPopRecord::dump( std::ostream& out ) const
{
    out << "BopPop" << std::endl;
}

static Record* createBopPopRecord(Workbook *book)
{
    return new BopPopRecord(book);
}

// ========== AxcExtRecord ==========

const unsigned AxcExtRecord::id = 0x1062;

class AxcExtRecord::Private
{
public:
};

AxcExtRecord::AxcExtRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

AxcExtRecord::~AxcExtRecord()
{
    delete d;
}

AxcExtRecord::AxcExtRecord( const AxcExtRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AxcExtRecord& AxcExtRecord::operator=( const AxcExtRecord& record )
{
    *d = *record.d;
    return *this;
}

void AxcExtRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AxcExtRecord::dump( std::ostream& out ) const
{
    out << "AxcExt" << std::endl;
}

static Record* createAxcExtRecord(Workbook *book)
{
    return new AxcExtRecord(book);
}

// ========== DatRecord ==========

const unsigned DatRecord::id = 0x1063;

class DatRecord::Private
{
public:
};

DatRecord::DatRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

DatRecord::~DatRecord()
{
    delete d;
}

DatRecord::DatRecord( const DatRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DatRecord& DatRecord::operator=( const DatRecord& record )
{
    *d = *record.d;
    return *this;
}

void DatRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void DatRecord::dump( std::ostream& out ) const
{
    out << "Dat" << std::endl;
}

static Record* createDatRecord(Workbook *book)
{
    return new DatRecord(book);
}

// ========== PlotGrowthRecord ==========

const unsigned PlotGrowthRecord::id = 0x1064;

class PlotGrowthRecord::Private
{
public:
};

PlotGrowthRecord::PlotGrowthRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PlotGrowthRecord::~PlotGrowthRecord()
{
    delete d;
}

PlotGrowthRecord::PlotGrowthRecord( const PlotGrowthRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PlotGrowthRecord& PlotGrowthRecord::operator=( const PlotGrowthRecord& record )
{
    *d = *record.d;
    return *this;
}

void PlotGrowthRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PlotGrowthRecord::dump( std::ostream& out ) const
{
    out << "PlotGrowth" << std::endl;
}

static Record* createPlotGrowthRecord(Workbook *book)
{
    return new PlotGrowthRecord(book);
}

// ========== SIIndexRecord ==========

const unsigned SIIndexRecord::id = 0x1065;

class SIIndexRecord::Private
{
public:
    unsigned numIndex;
};

SIIndexRecord::SIIndexRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setNumIndex(0);
}

SIIndexRecord::~SIIndexRecord()
{
    delete d;
}

SIIndexRecord::SIIndexRecord( const SIIndexRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SIIndexRecord& SIIndexRecord::operator=( const SIIndexRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned SIIndexRecord::numIndex() const
{
    return d->numIndex;
}

void SIIndexRecord::setNumIndex(unsigned numIndex )
{
    d->numIndex = numIndex;
}

void SIIndexRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setNumIndex(readU16(data));
}

void SIIndexRecord::dump( std::ostream& out ) const
{
    out << "SIIndex" << std::endl;
    out << "           NumIndex : " << numIndex() << std::endl;
}

static Record* createSIIndexRecord(Workbook *book)
{
    return new SIIndexRecord(book);
}

// ========== GelFrameRecord ==========

const unsigned GelFrameRecord::id = 0x1066;

class GelFrameRecord::Private
{
public:
};

GelFrameRecord::GelFrameRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

GelFrameRecord::~GelFrameRecord()
{
    delete d;
}

GelFrameRecord::GelFrameRecord( const GelFrameRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

GelFrameRecord& GelFrameRecord::operator=( const GelFrameRecord& record )
{
    *d = *record.d;
    return *this;
}

void GelFrameRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void GelFrameRecord::dump( std::ostream& out ) const
{
    out << "GelFrame" << std::endl;
}

static Record* createGelFrameRecord(Workbook *book)
{
    return new GelFrameRecord(book);
}

// ========== BobPopCustomRecord ==========

const unsigned BobPopCustomRecord::id = 0x1067;

class BobPopCustomRecord::Private
{
public:
};

BobPopCustomRecord::BobPopCustomRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BobPopCustomRecord::~BobPopCustomRecord()
{
    delete d;
}

BobPopCustomRecord::BobPopCustomRecord( const BobPopCustomRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BobPopCustomRecord& BobPopCustomRecord::operator=( const BobPopCustomRecord& record )
{
    *d = *record.d;
    return *this;
}

void BobPopCustomRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void BobPopCustomRecord::dump( std::ostream& out ) const
{
    out << "BobPopCustom" << std::endl;
}

static Record* createBobPopCustomRecord(Workbook *book)
{
    return new BobPopCustomRecord(book);
}

// ========== Fbi2Record ==========

const unsigned Fbi2Record::id = 0x1068;

class Fbi2Record::Private
{
public:
};

Fbi2Record::Fbi2Record(Workbook *book)
    : Record(book), d(new Private)
{
}

Fbi2Record::~Fbi2Record()
{
    delete d;
}

Fbi2Record::Fbi2Record( const Fbi2Record& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Fbi2Record& Fbi2Record::operator=( const Fbi2Record& record )
{
    *d = *record.d;
    return *this;
}

void Fbi2Record::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void Fbi2Record::dump( std::ostream& out ) const
{
    out << "Fbi2" << std::endl;
}

static Record* createFbi2Record(Workbook *book)
{
    return new Fbi2Record(book);
}

// ========== ZoomLevelRecord ==========

const unsigned ZoomLevelRecord::id = 0x00a0;

class ZoomLevelRecord::Private
{
public:
    int denominator;
    int numerator;
};

ZoomLevelRecord::ZoomLevelRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setDenominator(0);
    setNumerator(0);
}

ZoomLevelRecord::~ZoomLevelRecord()
{
    delete d;
}

ZoomLevelRecord::ZoomLevelRecord( const ZoomLevelRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ZoomLevelRecord& ZoomLevelRecord::operator=( const ZoomLevelRecord& record )
{
    *d = *record.d;
    return *this;
}

int ZoomLevelRecord::denominator() const
{
    return d->denominator;
}

void ZoomLevelRecord::setDenominator(int denominator )
{
    d->denominator = denominator;
}

int ZoomLevelRecord::numerator() const
{
    return d->numerator;
}

void ZoomLevelRecord::setNumerator(int numerator )
{
    d->numerator = numerator;
}

void ZoomLevelRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 4) {
        setIsValid(false);
        return;
    }
    setNumerator(readS16(data));
    setDenominator(readS16(data + 2));
}

void ZoomLevelRecord::dump( std::ostream& out ) const
{
    out << "ZoomLevel" << std::endl;
    out << "          Numerator : " << numerator() << std::endl;
    out << "        Denominator : " << denominator() << std::endl;
}

static Record* createZoomLevelRecord(Workbook *book)
{
    return new ZoomLevelRecord(book);
}

// ========== FrameRecord ==========

const unsigned FrameRecord::id = 0x1032;

class FrameRecord::Private
{
public:
    bool autoPosition;
    bool autoSize;
    FrameType frameType;
};

FrameRecord::FrameRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setAutoPosition(false);
    setAutoSize(false);
}

FrameRecord::~FrameRecord()
{
    delete d;
}

FrameRecord::FrameRecord( const FrameRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

FrameRecord& FrameRecord::operator=( const FrameRecord& record )
{
    *d = *record.d;
    return *this;
}

UString FrameRecord::frameTypeToString(FrameType frameType)
{
    switch (frameType) {
        case SimpleFrame: return UString("SimpleFrame");
        case ShadowedFrame: return UString("ShadowedFrame");
        default: return UString("Unknown: ") + UString::from(frameType);
    }
}

bool FrameRecord::isAutoPosition() const
{
    return d->autoPosition;
}

void FrameRecord::setAutoPosition(bool autoPosition )
{
    d->autoPosition = autoPosition;
}

bool FrameRecord::isAutoSize() const
{
    return d->autoSize;
}

void FrameRecord::setAutoSize(bool autoSize )
{
    d->autoSize = autoSize;
}

FrameRecord::FrameType FrameRecord::frameType() const
{
    return d->frameType;
}

void FrameRecord::setFrameType(FrameType frameType )
{
    d->frameType = frameType;
}

void FrameRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 4) {
        setIsValid(false);
        return;
    }
    setFrameType(static_cast<FrameType>(readU16(data)));
    setAutoSize(((readU8(data + 2)) & 0x1) != 0);
    setAutoPosition(((readU8(data + 2) >> 1) & 0x1) != 0);
}

void FrameRecord::dump( std::ostream& out ) const
{
    out << "Frame" << std::endl;
    out << "          FrameType : " << frameTypeToString(frameType()) << std::endl;
    out << "           AutoSize : " << isAutoSize() << std::endl;
    out << "       AutoPosition : " << isAutoPosition() << std::endl;
}

static Record* createFrameRecord(Workbook *book)
{
    return new FrameRecord(book);
}

// ========== SeriesRecord ==========

const unsigned SeriesRecord::id = 0x1003;

class SeriesRecord::Private
{
public:
    unsigned bubbleSizeDataType;
    unsigned countBubbleSizeValues;
    unsigned countXValues;
    unsigned countYValues;
    DataTypeX dataTypeX;
    unsigned dataTypeY;
};

SeriesRecord::SeriesRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setBubbleSizeDataType(0);
    setCountBubbleSizeValues(0);
    setCountXValues(0);
    setCountYValues(0);
    setDataTypeY(0);
}

SeriesRecord::~SeriesRecord()
{
    delete d;
}

SeriesRecord::SeriesRecord( const SeriesRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SeriesRecord& SeriesRecord::operator=( const SeriesRecord& record )
{
    *d = *record.d;
    return *this;
}

UString SeriesRecord::dataTypeXToString(DataTypeX dataTypeX)
{
    switch (dataTypeX) {
        case Numeric: return UString("Numeric");
        case Textual: return UString("Textual");
        default: return UString("Unknown: ") + UString::from(dataTypeX);
    }
}

unsigned SeriesRecord::bubbleSizeDataType() const
{
    return d->bubbleSizeDataType;
}

void SeriesRecord::setBubbleSizeDataType(unsigned bubbleSizeDataType )
{
    d->bubbleSizeDataType = bubbleSizeDataType;
}

unsigned SeriesRecord::countBubbleSizeValues() const
{
    return d->countBubbleSizeValues;
}

void SeriesRecord::setCountBubbleSizeValues(unsigned countBubbleSizeValues )
{
    d->countBubbleSizeValues = countBubbleSizeValues;
}

unsigned SeriesRecord::countXValues() const
{
    return d->countXValues;
}

void SeriesRecord::setCountXValues(unsigned countXValues )
{
    d->countXValues = countXValues;
}

unsigned SeriesRecord::countYValues() const
{
    return d->countYValues;
}

void SeriesRecord::setCountYValues(unsigned countYValues )
{
    d->countYValues = countYValues;
}

SeriesRecord::DataTypeX SeriesRecord::dataTypeX() const
{
    return d->dataTypeX;
}

void SeriesRecord::setDataTypeX(DataTypeX dataTypeX )
{
    d->dataTypeX = dataTypeX;
}

unsigned SeriesRecord::dataTypeY() const
{
    return d->dataTypeY;
}

void SeriesRecord::setDataTypeY(unsigned dataTypeY )
{
    d->dataTypeY = dataTypeY;
}

void SeriesRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 12) {
        setIsValid(false);
        return;
    }
    setDataTypeX(static_cast<DataTypeX>(readU16(data)));
    setDataTypeY(readU16(data + 2));
    setCountXValues(readU16(data + 4));
    setCountYValues(readU16(data + 6));
    setBubbleSizeDataType(readU16(data + 8));
    setCountBubbleSizeValues(readU16(data + 10));
}

void SeriesRecord::dump( std::ostream& out ) const
{
    out << "Series" << std::endl;
    out << "          DataTypeX : " << dataTypeXToString(dataTypeX()) << std::endl;
    out << "          DataTypeY : " << dataTypeY() << std::endl;
    out << "       CountXValues : " << countXValues() << std::endl;
    out << "       CountYValues : " << countYValues() << std::endl;
    out << " BubbleSizeDataType : " << bubbleSizeDataType() << std::endl;
    out << "CountBubbleSizeValues : " << countBubbleSizeValues() << std::endl;
}

static Record* createSeriesRecord(Workbook *book)
{
    return new SeriesRecord(book);
}

// ========== SeriesTextRecord ==========

const unsigned SeriesTextRecord::id = 0x100d;

class SeriesTextRecord::Private
{
public:
    UString text;
};

SeriesTextRecord::SeriesTextRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

SeriesTextRecord::~SeriesTextRecord()
{
    delete d;
}

SeriesTextRecord::SeriesTextRecord( const SeriesTextRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SeriesTextRecord& SeriesTextRecord::operator=( const SeriesTextRecord& record )
{
    *d = *record.d;
    return *this;
}

UString SeriesTextRecord::text() const
{
    return d->text;
}

void SeriesTextRecord::setText(UString text )
{
    d->text = text;
}

void SeriesTextRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    unsigned curOffset;
    bool stringLengthError = false;
    unsigned stringSize;
    if (size < 3) {
        setIsValid(false);
        return;
    }
    unsigned stringLength = readU8(data + 2);
    curOffset = 3;
    setText(readUnicodeString(data + curOffset, stringLength, size - curOffset, &stringLengthError, &stringSize));
    if (stringLengthError) {
        setIsValid(false);
        return;
    }
    curOffset += stringSize;
}

void SeriesTextRecord::dump( std::ostream& out ) const
{
    out << "SeriesText" << std::endl;
    out << "               Text : " << text() << std::endl;
}

static Record* createSeriesTextRecord(Workbook *book)
{
    return new SeriesTextRecord(book);
}

// ========== GridSetReservedRecord ==========

const unsigned GridSetReservedRecord::id = 0x0082;

class GridSetReservedRecord::Private
{
public:
};

GridSetReservedRecord::GridSetReservedRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

GridSetReservedRecord::~GridSetReservedRecord()
{
    delete d;
}

GridSetReservedRecord::GridSetReservedRecord( const GridSetReservedRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

GridSetReservedRecord& GridSetReservedRecord::operator=( const GridSetReservedRecord& record )
{
    *d = *record.d;
    return *this;
}

void GridSetReservedRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void GridSetReservedRecord::dump( std::ostream& out ) const
{
    out << "GridSetReserved" << std::endl;
}

static Record* createGridSetReservedRecord(Workbook *book)
{
    return new GridSetReservedRecord(book);
}

// ========== DSFReservedRecord ==========

const unsigned DSFReservedRecord::id = 0x0161;

class DSFReservedRecord::Private
{
public:
};

DSFReservedRecord::DSFReservedRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

DSFReservedRecord::~DSFReservedRecord()
{
    delete d;
}

DSFReservedRecord::DSFReservedRecord( const DSFReservedRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DSFReservedRecord& DSFReservedRecord::operator=( const DSFReservedRecord& record )
{
    *d = *record.d;
    return *this;
}

void DSFReservedRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void DSFReservedRecord::dump( std::ostream& out ) const
{
    out << "DSFReserved" << std::endl;
}

static Record* createDSFReservedRecord(Workbook *book)
{
    return new DSFReservedRecord(book);
}

// ========== Excel9FileRecord ==========

const unsigned Excel9FileRecord::id = 0x01c0;

class Excel9FileRecord::Private
{
public:
};

Excel9FileRecord::Excel9FileRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

Excel9FileRecord::~Excel9FileRecord()
{
    delete d;
}

Excel9FileRecord::Excel9FileRecord( const Excel9FileRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Excel9FileRecord& Excel9FileRecord::operator=( const Excel9FileRecord& record )
{
    *d = *record.d;
    return *this;
}

void Excel9FileRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void Excel9FileRecord::dump( std::ostream& out ) const
{
    out << "Excel9File" << std::endl;
}

static Record* createExcel9FileRecord(Workbook *book)
{
    return new Excel9FileRecord(book);
}

// ========== MmsReservedRecord ==========

const unsigned MmsReservedRecord::id = 0x00c1;

class MmsReservedRecord::Private
{
public:
};

MmsReservedRecord::MmsReservedRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

MmsReservedRecord::~MmsReservedRecord()
{
    delete d;
}

MmsReservedRecord::MmsReservedRecord( const MmsReservedRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

MmsReservedRecord& MmsReservedRecord::operator=( const MmsReservedRecord& record )
{
    *d = *record.d;
    return *this;
}

void MmsReservedRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void MmsReservedRecord::dump( std::ostream& out ) const
{
    out << "MmsReserved" << std::endl;
}

static Record* createMmsReservedRecord(Workbook *book)
{
    return new MmsReservedRecord(book);
}

// ========== Compat12Record ==========

const unsigned Compat12Record::id = 0x088c;

class Compat12Record::Private
{
public:
};

Compat12Record::Compat12Record(Workbook *book)
    : Record(book), d(new Private)
{
}

Compat12Record::~Compat12Record()
{
    delete d;
}

Compat12Record::Compat12Record( const Compat12Record& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Compat12Record& Compat12Record::operator=( const Compat12Record& record )
{
    *d = *record.d;
    return *this;
}

void Compat12Record::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void Compat12Record::dump( std::ostream& out ) const
{
    out << "Compat12" << std::endl;
}

static Record* createCompat12Record(Workbook *book)
{
    return new Compat12Record(book);
}

// ========== Window1Record ==========

const unsigned Window1Record::id = 0x003D;

class Window1Record::Private
{
public:
    unsigned cTabSel;
    int dxWn;
    int dyWn;
    bool fBotAdornment;
    bool fDspHScroll;
    bool fDspVScroll;
    bool fHidden;
    bool fIconic;
    bool fNoAFDateGroup;
    bool fVeryHidden;
    unsigned itabCur;
    unsigned itabFirst;
    unsigned wTabRatio;
    int xWn;
    int yWn;
};

Window1Record::Window1Record(Workbook *book)
    : Record(book), d(new Private)
{
    setCTabSel(0);
    setDxWn(0);
    setDyWn(0);
    setFBotAdornment(false);
    setFDspHScroll(false);
    setFDspVScroll(false);
    setFHidden(false);
    setFIconic(false);
    setFNoAFDateGroup(false);
    setFVeryHidden(false);
    setItabCur(0);
    setItabFirst(0);
    setWTabRatio(0);
    setXWn(0);
    setYWn(0);
}

Window1Record::~Window1Record()
{
    delete d;
}

Window1Record::Window1Record( const Window1Record& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Window1Record& Window1Record::operator=( const Window1Record& record )
{
    *d = *record.d;
    return *this;
}

unsigned Window1Record::cTabSel() const
{
    return d->cTabSel;
}

void Window1Record::setCTabSel(unsigned cTabSel )
{
    d->cTabSel = cTabSel;
}

int Window1Record::dxWn() const
{
    return d->dxWn;
}

void Window1Record::setDxWn(int dxWn )
{
    d->dxWn = dxWn;
}

int Window1Record::dyWn() const
{
    return d->dyWn;
}

void Window1Record::setDyWn(int dyWn )
{
    d->dyWn = dyWn;
}

bool Window1Record::isFBotAdornment() const
{
    return d->fBotAdornment;
}

void Window1Record::setFBotAdornment(bool fBotAdornment )
{
    d->fBotAdornment = fBotAdornment;
}

bool Window1Record::isFDspHScroll() const
{
    return d->fDspHScroll;
}

void Window1Record::setFDspHScroll(bool fDspHScroll )
{
    d->fDspHScroll = fDspHScroll;
}

bool Window1Record::isFDspVScroll() const
{
    return d->fDspVScroll;
}

void Window1Record::setFDspVScroll(bool fDspVScroll )
{
    d->fDspVScroll = fDspVScroll;
}

bool Window1Record::isFHidden() const
{
    return d->fHidden;
}

void Window1Record::setFHidden(bool fHidden )
{
    d->fHidden = fHidden;
}

bool Window1Record::isFIconic() const
{
    return d->fIconic;
}

void Window1Record::setFIconic(bool fIconic )
{
    d->fIconic = fIconic;
}

bool Window1Record::isFNoAFDateGroup() const
{
    return d->fNoAFDateGroup;
}

void Window1Record::setFNoAFDateGroup(bool fNoAFDateGroup )
{
    d->fNoAFDateGroup = fNoAFDateGroup;
}

bool Window1Record::isFVeryHidden() const
{
    return d->fVeryHidden;
}

void Window1Record::setFVeryHidden(bool fVeryHidden )
{
    d->fVeryHidden = fVeryHidden;
}

unsigned Window1Record::itabCur() const
{
    return d->itabCur;
}

void Window1Record::setItabCur(unsigned itabCur )
{
    d->itabCur = itabCur;
}

unsigned Window1Record::itabFirst() const
{
    return d->itabFirst;
}

void Window1Record::setItabFirst(unsigned itabFirst )
{
    d->itabFirst = itabFirst;
}

unsigned Window1Record::wTabRatio() const
{
    return d->wTabRatio;
}

void Window1Record::setWTabRatio(unsigned wTabRatio )
{
    d->wTabRatio = wTabRatio;
}

int Window1Record::xWn() const
{
    return d->xWn;
}

void Window1Record::setXWn(int xWn )
{
    d->xWn = xWn;
}

int Window1Record::yWn() const
{
    return d->yWn;
}

void Window1Record::setYWn(int yWn )
{
    d->yWn = yWn;
}

void Window1Record::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 18) {
        setIsValid(false);
        return;
    }
    setXWn(readS16(data));
    setYWn(readS16(data + 2));
    setDxWn(readS16(data + 4));
    setDyWn(readS16(data + 6));
    setFHidden(((readU8(data + 8)) & 0x1) != 0);
    setFIconic(((readU8(data + 8) >> 1) & 0x1) != 0);
    setFVeryHidden(((readU8(data + 8) >> 2) & 0x1) != 0);
    setFDspHScroll(((readU8(data + 8) >> 3) & 0x1) != 0);
    setFDspVScroll(((readU8(data + 8) >> 4) & 0x1) != 0);
    setFBotAdornment(((readU8(data + 8) >> 5) & 0x1) != 0);
    setFNoAFDateGroup(((readU8(data + 8) >> 6) & 0x1) != 0);
    setItabCur(readU16(data + 10));
    setItabFirst(readU16(data + 12));
    setCTabSel(readU16(data + 14));
    setWTabRatio(readU16(data + 16));
}

void Window1Record::dump( std::ostream& out ) const
{
    out << "Window1" << std::endl;
    out << "                XWn : " << xWn() << std::endl;
    out << "                YWn : " << yWn() << std::endl;
    out << "               DxWn : " << dxWn() << std::endl;
    out << "               DyWn : " << dyWn() << std::endl;
    out << "            FHidden : " << isFHidden() << std::endl;
    out << "            FIconic : " << isFIconic() << std::endl;
    out << "        FVeryHidden : " << isFVeryHidden() << std::endl;
    out << "        FDspHScroll : " << isFDspHScroll() << std::endl;
    out << "        FDspVScroll : " << isFDspVScroll() << std::endl;
    out << "      FBotAdornment : " << isFBotAdornment() << std::endl;
    out << "     FNoAFDateGroup : " << isFNoAFDateGroup() << std::endl;
    out << "            ItabCur : " << itabCur() << std::endl;
    out << "          ItabFirst : " << itabFirst() << std::endl;
    out << "            CTabSel : " << cTabSel() << std::endl;
    out << "          WTabRatio : " << wTabRatio() << std::endl;
}

static Record* createWindow1Record(Workbook *book)
{
    return new Window1Record(book);
}

// ========== Window2Record ==========

const unsigned Window2Record::id = 0x023E;

class Window2Record::Private
{
public:
    unsigned colLeft;
    bool fDefaultHdr;
    bool fDspFmlaRt;
    bool fDspGridRt;
    bool fDspGuts;
    bool fDspRwColRt;
    bool fDspZerosRt;
    bool fFrozenNoSplit;
    bool fFrozenRt;
    bool fPaged;
    bool fRightToLeft;
    bool fSLV;
    bool fSelected;
    unsigned rwTop;
};

Window2Record::Window2Record(Workbook *book)
    : Record(book), d(new Private)
{
    setColLeft(0);
    setFDefaultHdr(false);
    setFDspFmlaRt(false);
    setFDspGridRt(false);
    setFDspGuts(false);
    setFDspRwColRt(false);
    setFDspZerosRt(false);
    setFFrozenNoSplit(false);
    setFFrozenRt(false);
    setFPaged(false);
    setFRightToLeft(false);
    setFSLV(false);
    setFSelected(false);
    setRwTop(0);
}

Window2Record::~Window2Record()
{
    delete d;
}

Window2Record::Window2Record( const Window2Record& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Window2Record& Window2Record::operator=( const Window2Record& record )
{
    *d = *record.d;
    return *this;
}

unsigned Window2Record::colLeft() const
{
    return d->colLeft;
}

void Window2Record::setColLeft(unsigned colLeft )
{
    d->colLeft = colLeft;
}

bool Window2Record::isFDefaultHdr() const
{
    return d->fDefaultHdr;
}

void Window2Record::setFDefaultHdr(bool fDefaultHdr )
{
    d->fDefaultHdr = fDefaultHdr;
}

bool Window2Record::isFDspFmlaRt() const
{
    return d->fDspFmlaRt;
}

void Window2Record::setFDspFmlaRt(bool fDspFmlaRt )
{
    d->fDspFmlaRt = fDspFmlaRt;
}

bool Window2Record::isFDspGridRt() const
{
    return d->fDspGridRt;
}

void Window2Record::setFDspGridRt(bool fDspGridRt )
{
    d->fDspGridRt = fDspGridRt;
}

bool Window2Record::isFDspGuts() const
{
    return d->fDspGuts;
}

void Window2Record::setFDspGuts(bool fDspGuts )
{
    d->fDspGuts = fDspGuts;
}

bool Window2Record::isFDspRwColRt() const
{
    return d->fDspRwColRt;
}

void Window2Record::setFDspRwColRt(bool fDspRwColRt )
{
    d->fDspRwColRt = fDspRwColRt;
}

bool Window2Record::isFDspZerosRt() const
{
    return d->fDspZerosRt;
}

void Window2Record::setFDspZerosRt(bool fDspZerosRt )
{
    d->fDspZerosRt = fDspZerosRt;
}

bool Window2Record::isFFrozenNoSplit() const
{
    return d->fFrozenNoSplit;
}

void Window2Record::setFFrozenNoSplit(bool fFrozenNoSplit )
{
    d->fFrozenNoSplit = fFrozenNoSplit;
}

bool Window2Record::isFFrozenRt() const
{
    return d->fFrozenRt;
}

void Window2Record::setFFrozenRt(bool fFrozenRt )
{
    d->fFrozenRt = fFrozenRt;
}

bool Window2Record::isFPaged() const
{
    return d->fPaged;
}

void Window2Record::setFPaged(bool fPaged )
{
    d->fPaged = fPaged;
}

bool Window2Record::isFRightToLeft() const
{
    return d->fRightToLeft;
}

void Window2Record::setFRightToLeft(bool fRightToLeft )
{
    d->fRightToLeft = fRightToLeft;
}

bool Window2Record::isFSLV() const
{
    return d->fSLV;
}

void Window2Record::setFSLV(bool fSLV )
{
    d->fSLV = fSLV;
}

bool Window2Record::isFSelected() const
{
    return d->fSelected;
}

void Window2Record::setFSelected(bool fSelected )
{
    d->fSelected = fSelected;
}

unsigned Window2Record::rwTop() const
{
    return d->rwTop;
}

void Window2Record::setRwTop(unsigned rwTop )
{
    d->rwTop = rwTop;
}

void Window2Record::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 6) {
        setIsValid(false);
        return;
    }
    setFDspFmlaRt(((readU8(data)) & 0x1) != 0);
    setFDspGridRt(((readU8(data) >> 1) & 0x1) != 0);
    setFDspRwColRt(((readU8(data) >> 2) & 0x1) != 0);
    setFFrozenRt(((readU8(data) >> 3) & 0x1) != 0);
    setFDspZerosRt(((readU8(data) >> 4) & 0x1) != 0);
    setFDefaultHdr(((readU8(data) >> 5) & 0x1) != 0);
    setFRightToLeft(((readU8(data) >> 6) & 0x1) != 0);
    setFDspGuts(((readU8(data) >> 7) & 0x1) != 0);
    setFFrozenNoSplit(((readU8(data + 1)) & 0x1) != 0);
    setFSelected(((readU8(data + 1) >> 1) & 0x1) != 0);
    setFPaged(((readU8(data + 1) >> 2) & 0x1) != 0);
    setFSLV(((readU8(data + 1) >> 3) & 0x1) != 0);
    setRwTop(readU16(data + 2));
    setColLeft(readU16(data + 4));
}

void Window2Record::dump( std::ostream& out ) const
{
    out << "Window2" << std::endl;
    out << "         FDspFmlaRt : " << isFDspFmlaRt() << std::endl;
    out << "         FDspGridRt : " << isFDspGridRt() << std::endl;
    out << "        FDspRwColRt : " << isFDspRwColRt() << std::endl;
    out << "          FFrozenRt : " << isFFrozenRt() << std::endl;
    out << "        FDspZerosRt : " << isFDspZerosRt() << std::endl;
    out << "        FDefaultHdr : " << isFDefaultHdr() << std::endl;
    out << "       FRightToLeft : " << isFRightToLeft() << std::endl;
    out << "           FDspGuts : " << isFDspGuts() << std::endl;
    out << "     FFrozenNoSplit : " << isFFrozenNoSplit() << std::endl;
    out << "          FSelected : " << isFSelected() << std::endl;
    out << "             FPaged : " << isFPaged() << std::endl;
    out << "               FSLV : " << isFSLV() << std::endl;
    out << "              RwTop : " << rwTop() << std::endl;
    out << "            ColLeft : " << colLeft() << std::endl;
}

static Record* createWindow2Record(Workbook *book)
{
    return new Window2Record(book);
}

// ========== ObjProtectedRecord ==========

const unsigned ObjProtectedRecord::id = 0x0063;

class ObjProtectedRecord::Private
{
public:
    unsigned fLockObj;
};

ObjProtectedRecord::ObjProtectedRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFLockObj(0);
}

ObjProtectedRecord::~ObjProtectedRecord()
{
    delete d;
}

ObjProtectedRecord::ObjProtectedRecord( const ObjProtectedRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ObjProtectedRecord& ObjProtectedRecord::operator=( const ObjProtectedRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned ObjProtectedRecord::fLockObj() const
{
    return d->fLockObj;
}

void ObjProtectedRecord::setFLockObj(unsigned fLockObj )
{
    d->fLockObj = fLockObj;
}

void ObjProtectedRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setFLockObj(readU16(data));
}

void ObjProtectedRecord::dump( std::ostream& out ) const
{
    out << "ObjProtected" << std::endl;
    out << "           FLockObj : " << fLockObj() << std::endl;
}

static Record* createObjProtectedRecord(Workbook *book)
{
    return new ObjProtectedRecord(book);
}

// ========== ScenarioProtectRecord ==========

const unsigned ScenarioProtectRecord::id = 0x00dd;

class ScenarioProtectRecord::Private
{
public:
    unsigned fScenProtect;
};

ScenarioProtectRecord::ScenarioProtectRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFScenProtect(0);
}

ScenarioProtectRecord::~ScenarioProtectRecord()
{
    delete d;
}

ScenarioProtectRecord::ScenarioProtectRecord( const ScenarioProtectRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ScenarioProtectRecord& ScenarioProtectRecord::operator=( const ScenarioProtectRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned ScenarioProtectRecord::fScenProtect() const
{
    return d->fScenProtect;
}

void ScenarioProtectRecord::setFScenProtect(unsigned fScenProtect )
{
    d->fScenProtect = fScenProtect;
}

void ScenarioProtectRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setFScenProtect(readU16(data));
}

void ScenarioProtectRecord::dump( std::ostream& out ) const
{
    out << "ScenarioProtect" << std::endl;
    out << "       FScenProtect : " << fScenProtect() << std::endl;
}

static Record* createScenarioProtectRecord(Workbook *book)
{
    return new ScenarioProtectRecord(book);
}

// ========== WinProtectRecord ==========

const unsigned WinProtectRecord::id = 0x0019;

class WinProtectRecord::Private
{
public:
    unsigned fLockWin;
};

WinProtectRecord::WinProtectRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFLockWin(0);
}

WinProtectRecord::~WinProtectRecord()
{
    delete d;
}

WinProtectRecord::WinProtectRecord( const WinProtectRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

WinProtectRecord& WinProtectRecord::operator=( const WinProtectRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned WinProtectRecord::fLockWin() const
{
    return d->fLockWin;
}

void WinProtectRecord::setFLockWin(unsigned fLockWin )
{
    d->fLockWin = fLockWin;
}

void WinProtectRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setFLockWin(readU16(data));
}

void WinProtectRecord::dump( std::ostream& out ) const
{
    out << "WinProtect" << std::endl;
    out << "           FLockWin : " << fLockWin() << std::endl;
}

static Record* createWinProtectRecord(Workbook *book)
{
    return new WinProtectRecord(book);
}

// ========== Prot4RevRecord ==========

const unsigned Prot4RevRecord::id = 0x01af;

class Prot4RevRecord::Private
{
public:
    unsigned fRevLock;
};

Prot4RevRecord::Prot4RevRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setFRevLock(0);
}

Prot4RevRecord::~Prot4RevRecord()
{
    delete d;
}

Prot4RevRecord::Prot4RevRecord( const Prot4RevRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Prot4RevRecord& Prot4RevRecord::operator=( const Prot4RevRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned Prot4RevRecord::fRevLock() const
{
    return d->fRevLock;
}

void Prot4RevRecord::setFRevLock(unsigned fRevLock )
{
    d->fRevLock = fRevLock;
}

void Prot4RevRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setFRevLock(readU16(data));
}

void Prot4RevRecord::dump( std::ostream& out ) const
{
    out << "Prot4Rev" << std::endl;
    out << "           FRevLock : " << fRevLock() << std::endl;
}

static Record* createProt4RevRecord(Workbook *book)
{
    return new Prot4RevRecord(book);
}

// ========== Prot4RevPassRecord ==========

const unsigned Prot4RevPassRecord::id = 0x01bc;

class Prot4RevPassRecord::Private
{
public:
    unsigned protPwdRev;
};

Prot4RevPassRecord::Prot4RevPassRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setProtPwdRev(0);
}

Prot4RevPassRecord::~Prot4RevPassRecord()
{
    delete d;
}

Prot4RevPassRecord::Prot4RevPassRecord( const Prot4RevPassRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

Prot4RevPassRecord& Prot4RevPassRecord::operator=( const Prot4RevPassRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned Prot4RevPassRecord::protPwdRev() const
{
    return d->protPwdRev;
}

void Prot4RevPassRecord::setProtPwdRev(unsigned protPwdRev )
{
    d->protPwdRev = protPwdRev;
}

void Prot4RevPassRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setProtPwdRev(readU16(data));
}

void Prot4RevPassRecord::dump( std::ostream& out ) const
{
    out << "Prot4RevPass" << std::endl;
    out << "         ProtPwdRev : " << protPwdRev() << std::endl;
}

static Record* createProt4RevPassRecord(Workbook *book)
{
    return new Prot4RevPassRecord(book);
}

// ========== PasswordRecord ==========

const unsigned PasswordRecord::id = 0x0013;

class PasswordRecord::Private
{
public:
    unsigned wPassword;
};

PasswordRecord::PasswordRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setWPassword(0);
}

PasswordRecord::~PasswordRecord()
{
    delete d;
}

PasswordRecord::PasswordRecord( const PasswordRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PasswordRecord& PasswordRecord::operator=( const PasswordRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned PasswordRecord::wPassword() const
{
    return d->wPassword;
}

void PasswordRecord::setWPassword(unsigned wPassword )
{
    d->wPassword = wPassword;
}

void PasswordRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setWPassword(readU16(data));
}

void PasswordRecord::dump( std::ostream& out ) const
{
    out << "Password" << std::endl;
    out << "          WPassword : " << wPassword() << std::endl;
}

static Record* createPasswordRecord(Workbook *book)
{
    return new PasswordRecord(book);
}

// ========== FeatHdrRecord ==========

const unsigned FeatHdrRecord::id = 0x0867;

class FeatHdrRecord::Private
{
public:
};

FeatHdrRecord::FeatHdrRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

FeatHdrRecord::~FeatHdrRecord()
{
    delete d;
}

FeatHdrRecord::FeatHdrRecord( const FeatHdrRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

FeatHdrRecord& FeatHdrRecord::operator=( const FeatHdrRecord& record )
{
    *d = *record.d;
    return *this;
}

void FeatHdrRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void FeatHdrRecord::dump( std::ostream& out ) const
{
    out << "FeatHdr" << std::endl;
}

static Record* createFeatHdrRecord(Workbook *book)
{
    return new FeatHdrRecord(book);
}

// ========== IndexRecord ==========

const unsigned IndexRecord::id = 0x020b;

class IndexRecord::Private
{
public:
};

IndexRecord::IndexRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

IndexRecord::~IndexRecord()
{
    delete d;
}

IndexRecord::IndexRecord( const IndexRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

IndexRecord& IndexRecord::operator=( const IndexRecord& record )
{
    *d = *record.d;
    return *this;
}

void IndexRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void IndexRecord::dump( std::ostream& out ) const
{
    out << "Index" << std::endl;
}

static Record* createIndexRecord(Workbook *book)
{
    return new IndexRecord(book);
}

// ========== GutsRecord ==========

const unsigned GutsRecord::id = 0x0080;

class GutsRecord::Private
{
public:
};

GutsRecord::GutsRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

GutsRecord::~GutsRecord()
{
    delete d;
}

GutsRecord::GutsRecord( const GutsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

GutsRecord& GutsRecord::operator=( const GutsRecord& record )
{
    *d = *record.d;
    return *this;
}

void GutsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void GutsRecord::dump( std::ostream& out ) const
{
    out << "Guts" << std::endl;
}

static Record* createGutsRecord(Workbook *book)
{
    return new GutsRecord(book);
}

// ========== WsBoolRecord ==========

const unsigned WsBoolRecord::id = 0x0081;

class WsBoolRecord::Private
{
public:
};

WsBoolRecord::WsBoolRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

WsBoolRecord::~WsBoolRecord()
{
    delete d;
}

WsBoolRecord::WsBoolRecord( const WsBoolRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

WsBoolRecord& WsBoolRecord::operator=( const WsBoolRecord& record )
{
    *d = *record.d;
    return *this;
}

void WsBoolRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void WsBoolRecord::dump( std::ostream& out ) const
{
    out << "WsBool" << std::endl;
}

static Record* createWsBoolRecord(Workbook *book)
{
    return new WsBoolRecord(book);
}

// ========== HeaderFooterRecord ==========

const unsigned HeaderFooterRecord::id = 0x089c;

class HeaderFooterRecord::Private
{
public:
};

HeaderFooterRecord::HeaderFooterRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

HeaderFooterRecord::~HeaderFooterRecord()
{
    delete d;
}

HeaderFooterRecord::HeaderFooterRecord( const HeaderFooterRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

HeaderFooterRecord& HeaderFooterRecord::operator=( const HeaderFooterRecord& record )
{
    *d = *record.d;
    return *this;
}

void HeaderFooterRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void HeaderFooterRecord::dump( std::ostream& out ) const
{
    out << "HeaderFooter" << std::endl;
}

static Record* createHeaderFooterRecord(Workbook *book)
{
    return new HeaderFooterRecord(book);
}

// ========== PLVRecord ==========

const unsigned PLVRecord::id = 0x088b;

class PLVRecord::Private
{
public:
};

PLVRecord::PLVRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PLVRecord::~PLVRecord()
{
    delete d;
}

PLVRecord::PLVRecord( const PLVRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PLVRecord& PLVRecord::operator=( const PLVRecord& record )
{
    *d = *record.d;
    return *this;
}

void PLVRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PLVRecord::dump( std::ostream& out ) const
{
    out << "PLV" << std::endl;
}

static Record* createPLVRecord(Workbook *book)
{
    return new PLVRecord(book);
}

// ========== CalcSaveRecalcRecord ==========

const unsigned CalcSaveRecalcRecord::id = 0x005f;

class CalcSaveRecalcRecord::Private
{
public:
};

CalcSaveRecalcRecord::CalcSaveRecalcRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CalcSaveRecalcRecord::~CalcSaveRecalcRecord()
{
    delete d;
}

CalcSaveRecalcRecord::CalcSaveRecalcRecord( const CalcSaveRecalcRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CalcSaveRecalcRecord& CalcSaveRecalcRecord::operator=( const CalcSaveRecalcRecord& record )
{
    *d = *record.d;
    return *this;
}

void CalcSaveRecalcRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CalcSaveRecalcRecord::dump( std::ostream& out ) const
{
    out << "CalcSaveRecalc" << std::endl;
}

static Record* createCalcSaveRecalcRecord(Workbook *book)
{
    return new CalcSaveRecalcRecord(book);
}

// ========== InterfaceHdrRecord ==========

const unsigned InterfaceHdrRecord::id = 0x00e1;

class InterfaceHdrRecord::Private
{
public:
};

InterfaceHdrRecord::InterfaceHdrRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

InterfaceHdrRecord::~InterfaceHdrRecord()
{
    delete d;
}

InterfaceHdrRecord::InterfaceHdrRecord( const InterfaceHdrRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

InterfaceHdrRecord& InterfaceHdrRecord::operator=( const InterfaceHdrRecord& record )
{
    *d = *record.d;
    return *this;
}

void InterfaceHdrRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void InterfaceHdrRecord::dump( std::ostream& out ) const
{
    out << "InterfaceHdr" << std::endl;
}

static Record* createInterfaceHdrRecord(Workbook *book)
{
    return new InterfaceHdrRecord(book);
}

// ========== InterfaceEndRecord ==========

const unsigned InterfaceEndRecord::id = 0x00e2;

class InterfaceEndRecord::Private
{
public:
};

InterfaceEndRecord::InterfaceEndRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

InterfaceEndRecord::~InterfaceEndRecord()
{
    delete d;
}

InterfaceEndRecord::InterfaceEndRecord( const InterfaceEndRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

InterfaceEndRecord& InterfaceEndRecord::operator=( const InterfaceEndRecord& record )
{
    *d = *record.d;
    return *this;
}

void InterfaceEndRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void InterfaceEndRecord::dump( std::ostream& out ) const
{
    out << "InterfaceEnd" << std::endl;
}

static Record* createInterfaceEndRecord(Workbook *book)
{
    return new InterfaceEndRecord(book);
}

// ========== RRTabIdRecord ==========

const unsigned RRTabIdRecord::id = 0x013d;

class RRTabIdRecord::Private
{
public:
};

RRTabIdRecord::RRTabIdRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

RRTabIdRecord::~RRTabIdRecord()
{
    delete d;
}

RRTabIdRecord::RRTabIdRecord( const RRTabIdRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

RRTabIdRecord& RRTabIdRecord::operator=( const RRTabIdRecord& record )
{
    *d = *record.d;
    return *this;
}

void RRTabIdRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void RRTabIdRecord::dump( std::ostream& out ) const
{
    out << "RRTabId" << std::endl;
}

static Record* createRRTabIdRecord(Workbook *book)
{
    return new RRTabIdRecord(book);
}

// ========== BuiltInFnGroupCountRecord ==========

const unsigned BuiltInFnGroupCountRecord::id = 0x009c;

class BuiltInFnGroupCountRecord::Private
{
public:
};

BuiltInFnGroupCountRecord::BuiltInFnGroupCountRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BuiltInFnGroupCountRecord::~BuiltInFnGroupCountRecord()
{
    delete d;
}

BuiltInFnGroupCountRecord::BuiltInFnGroupCountRecord( const BuiltInFnGroupCountRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BuiltInFnGroupCountRecord& BuiltInFnGroupCountRecord::operator=( const BuiltInFnGroupCountRecord& record )
{
    *d = *record.d;
    return *this;
}

void BuiltInFnGroupCountRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void BuiltInFnGroupCountRecord::dump( std::ostream& out ) const
{
    out << "BuiltInFnGroupCount" << std::endl;
}

static Record* createBuiltInFnGroupCountRecord(Workbook *book)
{
    return new BuiltInFnGroupCountRecord(book);
}

// ========== HideObjRecord ==========

const unsigned HideObjRecord::id = 0x008d;

class HideObjRecord::Private
{
public:
};

HideObjRecord::HideObjRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

HideObjRecord::~HideObjRecord()
{
    delete d;
}

HideObjRecord::HideObjRecord( const HideObjRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

HideObjRecord& HideObjRecord::operator=( const HideObjRecord& record )
{
    *d = *record.d;
    return *this;
}

void HideObjRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void HideObjRecord::dump( std::ostream& out ) const
{
    out << "HideObj" << std::endl;
}

static Record* createHideObjRecord(Workbook *book)
{
    return new HideObjRecord(book);
}

// ========== RefreshAllRecord ==========

const unsigned RefreshAllRecord::id = 0x01b7;

class RefreshAllRecord::Private
{
public:
};

RefreshAllRecord::RefreshAllRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

RefreshAllRecord::~RefreshAllRecord()
{
    delete d;
}

RefreshAllRecord::RefreshAllRecord( const RefreshAllRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

RefreshAllRecord& RefreshAllRecord::operator=( const RefreshAllRecord& record )
{
    *d = *record.d;
    return *this;
}

void RefreshAllRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void RefreshAllRecord::dump( std::ostream& out ) const
{
    out << "RefreshAll" << std::endl;
}

static Record* createRefreshAllRecord(Workbook *book)
{
    return new RefreshAllRecord(book);
}

// ========== RecalcIdRecord ==========

const unsigned RecalcIdRecord::id = 0x01c1;

class RecalcIdRecord::Private
{
public:
};

RecalcIdRecord::RecalcIdRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

RecalcIdRecord::~RecalcIdRecord()
{
    delete d;
}

RecalcIdRecord::RecalcIdRecord( const RecalcIdRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

RecalcIdRecord& RecalcIdRecord::operator=( const RecalcIdRecord& record )
{
    *d = *record.d;
    return *this;
}

void RecalcIdRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void RecalcIdRecord::dump( std::ostream& out ) const
{
    out << "RecalcId" << std::endl;
}

static Record* createRecalcIdRecord(Workbook *book)
{
    return new RecalcIdRecord(book);
}

// ========== BookBoolRecord ==========

const unsigned BookBoolRecord::id = 0x00da;

class BookBoolRecord::Private
{
public:
};

BookBoolRecord::BookBoolRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BookBoolRecord::~BookBoolRecord()
{
    delete d;
}

BookBoolRecord::BookBoolRecord( const BookBoolRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BookBoolRecord& BookBoolRecord::operator=( const BookBoolRecord& record )
{
    *d = *record.d;
    return *this;
}

void BookBoolRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void BookBoolRecord::dump( std::ostream& out ) const
{
    out << "BookBool" << std::endl;
}

static Record* createBookBoolRecord(Workbook *book)
{
    return new BookBoolRecord(book);
}

// ========== StyleRecord ==========

const unsigned StyleRecord::id = 0x0293;

class StyleRecord::Private
{
public:
};

StyleRecord::StyleRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

StyleRecord::~StyleRecord()
{
    delete d;
}

StyleRecord::StyleRecord( const StyleRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

StyleRecord& StyleRecord::operator=( const StyleRecord& record )
{
    *d = *record.d;
    return *this;
}

void StyleRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void StyleRecord::dump( std::ostream& out ) const
{
    out << "Style" << std::endl;
}

static Record* createStyleRecord(Workbook *book)
{
    return new StyleRecord(book);
}

// ========== StyleExtRecord ==========

const unsigned StyleExtRecord::id = 0x0892;

class StyleExtRecord::Private
{
public:
};

StyleExtRecord::StyleExtRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

StyleExtRecord::~StyleExtRecord()
{
    delete d;
}

StyleExtRecord::StyleExtRecord( const StyleExtRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

StyleExtRecord& StyleExtRecord::operator=( const StyleExtRecord& record )
{
    *d = *record.d;
    return *this;
}

void StyleExtRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void StyleExtRecord::dump( std::ostream& out ) const
{
    out << "StyleExt" << std::endl;
}

static Record* createStyleExtRecord(Workbook *book)
{
    return new StyleExtRecord(book);
}

// ========== ExtSSTRecord ==========

const unsigned ExtSSTRecord::id = 0x00ff;

class ExtSSTRecord::Private
{
public:
};

ExtSSTRecord::ExtSSTRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ExtSSTRecord::~ExtSSTRecord()
{
    delete d;
}

ExtSSTRecord::ExtSSTRecord( const ExtSSTRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ExtSSTRecord& ExtSSTRecord::operator=( const ExtSSTRecord& record )
{
    *d = *record.d;
    return *this;
}

void ExtSSTRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ExtSSTRecord::dump( std::ostream& out ) const
{
    out << "ExtSST" << std::endl;
}

static Record* createExtSSTRecord(Workbook *book)
{
    return new ExtSSTRecord(book);
}

// ========== UsesELFsRecord ==========

const unsigned UsesELFsRecord::id = 0x0160;

class UsesELFsRecord::Private
{
public:
};

UsesELFsRecord::UsesELFsRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

UsesELFsRecord::~UsesELFsRecord()
{
    delete d;
}

UsesELFsRecord::UsesELFsRecord( const UsesELFsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

UsesELFsRecord& UsesELFsRecord::operator=( const UsesELFsRecord& record )
{
    *d = *record.d;
    return *this;
}

void UsesELFsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void UsesELFsRecord::dump( std::ostream& out ) const
{
    out << "UsesELFs" << std::endl;
}

static Record* createUsesELFsRecord(Workbook *book)
{
    return new UsesELFsRecord(book);
}

// ========== DBCellRecord ==========

const unsigned DBCellRecord::id = 0x00d7;

class DBCellRecord::Private
{
public:
};

DBCellRecord::DBCellRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

DBCellRecord::~DBCellRecord()
{
    delete d;
}

DBCellRecord::DBCellRecord( const DBCellRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

DBCellRecord& DBCellRecord::operator=( const DBCellRecord& record )
{
    *d = *record.d;
    return *this;
}

void DBCellRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void DBCellRecord::dump( std::ostream& out ) const
{
    out << "DBCell" << std::endl;
}

static Record* createDBCellRecord(Workbook *book)
{
    return new DBCellRecord(book);
}

// ========== PhoneticInfoRecord ==========

const unsigned PhoneticInfoRecord::id = 0x00ef;

class PhoneticInfoRecord::Private
{
public:
};

PhoneticInfoRecord::PhoneticInfoRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PhoneticInfoRecord::~PhoneticInfoRecord()
{
    delete d;
}

PhoneticInfoRecord::PhoneticInfoRecord( const PhoneticInfoRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PhoneticInfoRecord& PhoneticInfoRecord::operator=( const PhoneticInfoRecord& record )
{
    *d = *record.d;
    return *this;
}

void PhoneticInfoRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PhoneticInfoRecord::dump( std::ostream& out ) const
{
    out << "PhoneticInfo" << std::endl;
}

static Record* createPhoneticInfoRecord(Workbook *book)
{
    return new PhoneticInfoRecord(book);
}

// ========== CalcCountRecord ==========

const unsigned CalcCountRecord::id = 0x000c;

class CalcCountRecord::Private
{
public:
};

CalcCountRecord::CalcCountRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CalcCountRecord::~CalcCountRecord()
{
    delete d;
}

CalcCountRecord::CalcCountRecord( const CalcCountRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CalcCountRecord& CalcCountRecord::operator=( const CalcCountRecord& record )
{
    *d = *record.d;
    return *this;
}

void CalcCountRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CalcCountRecord::dump( std::ostream& out ) const
{
    out << "CalcCount" << std::endl;
}

static Record* createCalcCountRecord(Workbook *book)
{
    return new CalcCountRecord(book);
}

// ========== CalcPrecisionRecord ==========

const unsigned CalcPrecisionRecord::id = 0x000e;

class CalcPrecisionRecord::Private
{
public:
};

CalcPrecisionRecord::CalcPrecisionRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CalcPrecisionRecord::~CalcPrecisionRecord()
{
    delete d;
}

CalcPrecisionRecord::CalcPrecisionRecord( const CalcPrecisionRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CalcPrecisionRecord& CalcPrecisionRecord::operator=( const CalcPrecisionRecord& record )
{
    *d = *record.d;
    return *this;
}

void CalcPrecisionRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CalcPrecisionRecord::dump( std::ostream& out ) const
{
    out << "CalcPrecision" << std::endl;
}

static Record* createCalcPrecisionRecord(Workbook *book)
{
    return new CalcPrecisionRecord(book);
}

// ========== CalcDeltaRecord ==========

const unsigned CalcDeltaRecord::id = 0x0010;

class CalcDeltaRecord::Private
{
public:
};

CalcDeltaRecord::CalcDeltaRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CalcDeltaRecord::~CalcDeltaRecord()
{
    delete d;
}

CalcDeltaRecord::CalcDeltaRecord( const CalcDeltaRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CalcDeltaRecord& CalcDeltaRecord::operator=( const CalcDeltaRecord& record )
{
    *d = *record.d;
    return *this;
}

void CalcDeltaRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CalcDeltaRecord::dump( std::ostream& out ) const
{
    out << "CalcDelta" << std::endl;
}

static Record* createCalcDeltaRecord(Workbook *book)
{
    return new CalcDeltaRecord(book);
}

// ========== CalcIterRecord ==========

const unsigned CalcIterRecord::id = 0x0011;

class CalcIterRecord::Private
{
public:
};

CalcIterRecord::CalcIterRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CalcIterRecord::~CalcIterRecord()
{
    delete d;
}

CalcIterRecord::CalcIterRecord( const CalcIterRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CalcIterRecord& CalcIterRecord::operator=( const CalcIterRecord& record )
{
    *d = *record.d;
    return *this;
}

void CalcIterRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CalcIterRecord::dump( std::ostream& out ) const
{
    out << "CalcIter" << std::endl;
}

static Record* createCalcIterRecord(Workbook *book)
{
    return new CalcIterRecord(book);
}

// ========== CalcRefModeRecord ==========

const unsigned CalcRefModeRecord::id = 0x000f;

class CalcRefModeRecord::Private
{
public:
};

CalcRefModeRecord::CalcRefModeRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CalcRefModeRecord::~CalcRefModeRecord()
{
    delete d;
}

CalcRefModeRecord::CalcRefModeRecord( const CalcRefModeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CalcRefModeRecord& CalcRefModeRecord::operator=( const CalcRefModeRecord& record )
{
    *d = *record.d;
    return *this;
}

void CalcRefModeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CalcRefModeRecord::dump( std::ostream& out ) const
{
    out << "CalcRefMode" << std::endl;
}

static Record* createCalcRefModeRecord(Workbook *book)
{
    return new CalcRefModeRecord(book);
}

// ========== CodePageRecord ==========

const unsigned CodePageRecord::id = 0x0042;

class CodePageRecord::Private
{
public:
};

CodePageRecord::CodePageRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CodePageRecord::~CodePageRecord()
{
    delete d;
}

CodePageRecord::CodePageRecord( const CodePageRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CodePageRecord& CodePageRecord::operator=( const CodePageRecord& record )
{
    *d = *record.d;
    return *this;
}

void CodePageRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CodePageRecord::dump( std::ostream& out ) const
{
    out << "CodePage" << std::endl;
}

static Record* createCodePageRecord(Workbook *book)
{
    return new CodePageRecord(book);
}

// ========== CountryRecord ==========

const unsigned CountryRecord::id = 0x008c;

class CountryRecord::Private
{
public:
};

CountryRecord::CountryRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CountryRecord::~CountryRecord()
{
    delete d;
}

CountryRecord::CountryRecord( const CountryRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CountryRecord& CountryRecord::operator=( const CountryRecord& record )
{
    *d = *record.d;
    return *this;
}

void CountryRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CountryRecord::dump( std::ostream& out ) const
{
    out << "Country" << std::endl;
}

static Record* createCountryRecord(Workbook *book)
{
    return new CountryRecord(book);
}

// ========== SelectionRecord ==========

const unsigned SelectionRecord::id = 0x001d;

class SelectionRecord::Private
{
public:
    unsigned colAct;
    unsigned cref;
    unsigned irefAct;
    unsigned pnn;
    unsigned rwAct;
};

SelectionRecord::SelectionRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setColAct(0);
    setCref(0);
    setIrefAct(0);
    setPnn(0);
    setRwAct(0);
}

SelectionRecord::~SelectionRecord()
{
    delete d;
}

SelectionRecord::SelectionRecord( const SelectionRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

SelectionRecord& SelectionRecord::operator=( const SelectionRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned SelectionRecord::colAct() const
{
    return d->colAct;
}

void SelectionRecord::setColAct(unsigned colAct )
{
    d->colAct = colAct;
}

unsigned SelectionRecord::cref() const
{
    return d->cref;
}

void SelectionRecord::setCref(unsigned cref )
{
    d->cref = cref;
}

unsigned SelectionRecord::irefAct() const
{
    return d->irefAct;
}

void SelectionRecord::setIrefAct(unsigned irefAct )
{
    d->irefAct = irefAct;
}

unsigned SelectionRecord::pnn() const
{
    return d->pnn;
}

void SelectionRecord::setPnn(unsigned pnn )
{
    d->pnn = pnn;
}

unsigned SelectionRecord::rwAct() const
{
    return d->rwAct;
}

void SelectionRecord::setRwAct(unsigned rwAct )
{
    d->rwAct = rwAct;
}

void SelectionRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 9) {
        setIsValid(false);
        return;
    }
    setPnn(readU8(data));
    setRwAct(readU16(data + 1));
    setColAct(readU16(data + 3));
    setIrefAct(readU16(data + 5));
    setCref(readU16(data + 7));
}

void SelectionRecord::dump( std::ostream& out ) const
{
    out << "Selection" << std::endl;
    out << "                Pnn : " << pnn() << std::endl;
    out << "              RwAct : " << rwAct() << std::endl;
    out << "             ColAct : " << colAct() << std::endl;
    out << "            IrefAct : " << irefAct() << std::endl;
    out << "               Cref : " << cref() << std::endl;
}

static Record* createSelectionRecord(Workbook *book)
{
    return new SelectionRecord(book);
}

// ========== PrinterSettingsRecord ==========

const unsigned PrinterSettingsRecord::id = 0x004d;

class PrinterSettingsRecord::Private
{
public:
};

PrinterSettingsRecord::PrinterSettingsRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PrinterSettingsRecord::~PrinterSettingsRecord()
{
    delete d;
}

PrinterSettingsRecord::PrinterSettingsRecord( const PrinterSettingsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PrinterSettingsRecord& PrinterSettingsRecord::operator=( const PrinterSettingsRecord& record )
{
    *d = *record.d;
    return *this;
}

void PrinterSettingsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PrinterSettingsRecord::dump( std::ostream& out ) const
{
    out << "PrinterSettings" << std::endl;
}

static Record* createPrinterSettingsRecord(Workbook *book)
{
    return new PrinterSettingsRecord(book);
}

// ========== PrintRowColRecord ==========

const unsigned PrintRowColRecord::id = 0x002a;

class PrintRowColRecord::Private
{
public:
};

PrintRowColRecord::PrintRowColRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PrintRowColRecord::~PrintRowColRecord()
{
    delete d;
}

PrintRowColRecord::PrintRowColRecord( const PrintRowColRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PrintRowColRecord& PrintRowColRecord::operator=( const PrintRowColRecord& record )
{
    *d = *record.d;
    return *this;
}

void PrintRowColRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PrintRowColRecord::dump( std::ostream& out ) const
{
    out << "PrintRowCol" << std::endl;
}

static Record* createPrintRowColRecord(Workbook *book)
{
    return new PrintRowColRecord(book);
}

// ========== PrintGridRecord ==========

const unsigned PrintGridRecord::id = 0x002b;

class PrintGridRecord::Private
{
public:
};

PrintGridRecord::PrintGridRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

PrintGridRecord::~PrintGridRecord()
{
    delete d;
}

PrintGridRecord::PrintGridRecord( const PrintGridRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

PrintGridRecord& PrintGridRecord::operator=( const PrintGridRecord& record )
{
    *d = *record.d;
    return *this;
}

void PrintGridRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void PrintGridRecord::dump( std::ostream& out ) const
{
    out << "PrintGrid" << std::endl;
}

static Record* createPrintGridRecord(Workbook *book)
{
    return new PrintGridRecord(book);
}

// ========== LastWriteAccessRecord ==========

const unsigned LastWriteAccessRecord::id = 0x005c;

class LastWriteAccessRecord::Private
{
public:
};

LastWriteAccessRecord::LastWriteAccessRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

LastWriteAccessRecord::~LastWriteAccessRecord()
{
    delete d;
}

LastWriteAccessRecord::LastWriteAccessRecord( const LastWriteAccessRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

LastWriteAccessRecord& LastWriteAccessRecord::operator=( const LastWriteAccessRecord& record )
{
    *d = *record.d;
    return *this;
}

void LastWriteAccessRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void LastWriteAccessRecord::dump( std::ostream& out ) const
{
    out << "LastWriteAccess" << std::endl;
}

static Record* createLastWriteAccessRecord(Workbook *book)
{
    return new LastWriteAccessRecord(book);
}

// ========== XFExtRecord ==========

const unsigned XFExtRecord::id = 0x087d;

class XFExtRecord::Private
{
public:
};

XFExtRecord::XFExtRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

XFExtRecord::~XFExtRecord()
{
    delete d;
}

XFExtRecord::XFExtRecord( const XFExtRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

XFExtRecord& XFExtRecord::operator=( const XFExtRecord& record )
{
    *d = *record.d;
    return *this;
}

void XFExtRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void XFExtRecord::dump( std::ostream& out ) const
{
    out << "XFExt" << std::endl;
}

static Record* createXFExtRecord(Workbook *book)
{
    return new XFExtRecord(book);
}

// ========== XFCRCRecord ==========

const unsigned XFCRCRecord::id = 0x087c;

class XFCRCRecord::Private
{
public:
};

XFCRCRecord::XFCRCRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

XFCRCRecord::~XFCRCRecord()
{
    delete d;
}

XFCRCRecord::XFCRCRecord( const XFCRCRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

XFCRCRecord& XFCRCRecord::operator=( const XFCRCRecord& record )
{
    *d = *record.d;
    return *this;
}

void XFCRCRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void XFCRCRecord::dump( std::ostream& out ) const
{
    out << "XFCRC" << std::endl;
}

static Record* createXFCRCRecord(Workbook *book)
{
    return new XFCRCRecord(book);
}

// ========== TableStylesRecord ==========

const unsigned TableStylesRecord::id = 0x088e;

class TableStylesRecord::Private
{
public:
};

TableStylesRecord::TableStylesRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

TableStylesRecord::~TableStylesRecord()
{
    delete d;
}

TableStylesRecord::TableStylesRecord( const TableStylesRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

TableStylesRecord& TableStylesRecord::operator=( const TableStylesRecord& record )
{
    *d = *record.d;
    return *this;
}

void TableStylesRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void TableStylesRecord::dump( std::ostream& out ) const
{
    out << "TableStyles" << std::endl;
}

static Record* createTableStylesRecord(Workbook *book)
{
    return new TableStylesRecord(book);
}

// ========== MTRSettingsRecord ==========

const unsigned MTRSettingsRecord::id = 0x089a;

class MTRSettingsRecord::Private
{
public:
};

MTRSettingsRecord::MTRSettingsRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

MTRSettingsRecord::~MTRSettingsRecord()
{
    delete d;
}

MTRSettingsRecord::MTRSettingsRecord( const MTRSettingsRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

MTRSettingsRecord& MTRSettingsRecord::operator=( const MTRSettingsRecord& record )
{
    *d = *record.d;
    return *this;
}

void MTRSettingsRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void MTRSettingsRecord::dump( std::ostream& out ) const
{
    out << "MTRSettings" << std::endl;
}

static Record* createMTRSettingsRecord(Workbook *book)
{
    return new MTRSettingsRecord(book);
}

// ========== ForceFullCalculationRecord ==========

const unsigned ForceFullCalculationRecord::id = 0x08a3;

class ForceFullCalculationRecord::Private
{
public:
};

ForceFullCalculationRecord::ForceFullCalculationRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ForceFullCalculationRecord::~ForceFullCalculationRecord()
{
    delete d;
}

ForceFullCalculationRecord::ForceFullCalculationRecord( const ForceFullCalculationRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ForceFullCalculationRecord& ForceFullCalculationRecord::operator=( const ForceFullCalculationRecord& record )
{
    *d = *record.d;
    return *this;
}

void ForceFullCalculationRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ForceFullCalculationRecord::dump( std::ostream& out ) const
{
    out << "ForceFullCalculation" << std::endl;
}

static Record* createForceFullCalculationRecord(Workbook *book)
{
    return new ForceFullCalculationRecord(book);
}

// ========== BookExtRecord ==========

const unsigned BookExtRecord::id = 0x0863;

class BookExtRecord::Private
{
public:
};

BookExtRecord::BookExtRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

BookExtRecord::~BookExtRecord()
{
    delete d;
}

BookExtRecord::BookExtRecord( const BookExtRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

BookExtRecord& BookExtRecord::operator=( const BookExtRecord& record )
{
    *d = *record.d;
    return *this;
}

void BookExtRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void BookExtRecord::dump( std::ostream& out ) const
{
    out << "BookExt" << std::endl;
}

static Record* createBookExtRecord(Workbook *book)
{
    return new BookExtRecord(book);
}

// ========== ThemeRecord ==========

const unsigned ThemeRecord::id = 0x0896;

class ThemeRecord::Private
{
public:
};

ThemeRecord::ThemeRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

ThemeRecord::~ThemeRecord()
{
    delete d;
}

ThemeRecord::ThemeRecord( const ThemeRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

ThemeRecord& ThemeRecord::operator=( const ThemeRecord& record )
{
    *d = *record.d;
    return *this;
}

void ThemeRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void ThemeRecord::dump( std::ostream& out ) const
{
    out << "Theme" << std::endl;
}

static Record* createThemeRecord(Workbook *book)
{
    return new ThemeRecord(book);
}

// ========== CompressPicturesRecord ==========

const unsigned CompressPicturesRecord::id = 0x089b;

class CompressPicturesRecord::Private
{
public:
};

CompressPicturesRecord::CompressPicturesRecord(Workbook *book)
    : Record(book), d(new Private)
{
}

CompressPicturesRecord::~CompressPicturesRecord()
{
    delete d;
}

CompressPicturesRecord::CompressPicturesRecord( const CompressPicturesRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

CompressPicturesRecord& CompressPicturesRecord::operator=( const CompressPicturesRecord& record )
{
    *d = *record.d;
    return *this;
}

void CompressPicturesRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void CompressPicturesRecord::dump( std::ostream& out ) const
{
    out << "CompressPictures" << std::endl;
}

static Record* createCompressPicturesRecord(Workbook *book)
{
    return new CompressPicturesRecord(book);
}

// ========== AutoFilterRecord ==========

const unsigned AutoFilterRecord::id = 0x009e;

class AutoFilterRecord::Private
{
public:
    unsigned entry;
    Join join;
    bool simple1;
    bool simple2;
    TopDirection topDirection;
    bool topN;
    unsigned topNCount;
    bool topPercentage;
};

AutoFilterRecord::AutoFilterRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setEntry(0);
    setSimple1(false);
    setSimple2(false);
    setTopN(false);
    setTopNCount(0);
    setTopPercentage(false);
}

AutoFilterRecord::~AutoFilterRecord()
{
    delete d;
}

AutoFilterRecord::AutoFilterRecord( const AutoFilterRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AutoFilterRecord& AutoFilterRecord::operator=( const AutoFilterRecord& record )
{
    *d = *record.d;
    return *this;
}

UString AutoFilterRecord::joinToString(Join join)
{
    switch (join) {
        case JoinAnd: return UString("JoinAnd");
        case JoinOr: return UString("JoinOr");
        default: return UString("Unknown: ") + UString::from(join);
    }
}

UString AutoFilterRecord::topDirectionToString(TopDirection topDirection)
{
    switch (topDirection) {
        case TopNBottom: return UString("TopNBottom");
        case TopNTop: return UString("TopNTop");
        default: return UString("Unknown: ") + UString::from(topDirection);
    }
}

unsigned AutoFilterRecord::entry() const
{
    return d->entry;
}

void AutoFilterRecord::setEntry(unsigned entry )
{
    d->entry = entry;
}

AutoFilterRecord::Join AutoFilterRecord::join() const
{
    return d->join;
}

void AutoFilterRecord::setJoin(Join join )
{
    d->join = join;
}

bool AutoFilterRecord::isSimple1() const
{
    return d->simple1;
}

void AutoFilterRecord::setSimple1(bool simple1 )
{
    d->simple1 = simple1;
}

bool AutoFilterRecord::isSimple2() const
{
    return d->simple2;
}

void AutoFilterRecord::setSimple2(bool simple2 )
{
    d->simple2 = simple2;
}

AutoFilterRecord::TopDirection AutoFilterRecord::topDirection() const
{
    return d->topDirection;
}

void AutoFilterRecord::setTopDirection(TopDirection topDirection )
{
    d->topDirection = topDirection;
}

bool AutoFilterRecord::isTopN() const
{
    return d->topN;
}

void AutoFilterRecord::setTopN(bool topN )
{
    d->topN = topN;
}

unsigned AutoFilterRecord::topNCount() const
{
    return d->topNCount;
}

void AutoFilterRecord::setTopNCount(unsigned topNCount )
{
    d->topNCount = topNCount;
}

bool AutoFilterRecord::isTopPercentage() const
{
    return d->topPercentage;
}

void AutoFilterRecord::setTopPercentage(bool topPercentage )
{
    d->topPercentage = topPercentage;
}

void AutoFilterRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 4) {
        setIsValid(false);
        return;
    }
    setEntry(readU16(data));
    setJoin(static_cast<Join>(((readU8(data + 2)) & 0x3)));
    setSimple1(((readU8(data + 2) >> 2) & 0x1) != 0);
    setSimple2(((readU8(data + 2) >> 3) & 0x1) != 0);
    setTopN(((readU8(data + 2) >> 4) & 0x1) != 0);
    setTopDirection(static_cast<TopDirection>(((readU8(data + 2) >> 5) & 0x1)));
    setTopPercentage(((readU8(data + 2) >> 6) & 0x1) != 0);
    setTopNCount(((readU16(data + 2) >> 7) & 0x1ff));
}

void AutoFilterRecord::dump( std::ostream& out ) const
{
    out << "AutoFilter" << std::endl;
    out << "              Entry : " << entry() << std::endl;
    out << "               Join : " << joinToString(join()) << std::endl;
    out << "            Simple1 : " << isSimple1() << std::endl;
    out << "            Simple2 : " << isSimple2() << std::endl;
    out << "               TopN : " << isTopN() << std::endl;
    out << "       TopDirection : " << topDirectionToString(topDirection()) << std::endl;
    out << "      TopPercentage : " << isTopPercentage() << std::endl;
    out << "          TopNCount : " << topNCount() << std::endl;
}

static Record* createAutoFilterRecord(Workbook *book)
{
    return new AutoFilterRecord(book);
}

// ========== AutoFilter12Record ==========

const unsigned AutoFilter12Record::id = 0x087e;

class AutoFilter12Record::Private
{
public:
};

AutoFilter12Record::AutoFilter12Record(Workbook *book)
    : Record(book), d(new Private)
{
}

AutoFilter12Record::~AutoFilter12Record()
{
    delete d;
}

AutoFilter12Record::AutoFilter12Record( const AutoFilter12Record& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AutoFilter12Record& AutoFilter12Record::operator=( const AutoFilter12Record& record )
{
    *d = *record.d;
    return *this;
}

void AutoFilter12Record::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
}

void AutoFilter12Record::dump( std::ostream& out ) const
{
    out << "AutoFilter12" << std::endl;
}

static Record* createAutoFilter12Record(Workbook *book)
{
    return new AutoFilter12Record(book);
}

// ========== AutoFilterInfoRecord ==========

const unsigned AutoFilterInfoRecord::id = 0x009d;

class AutoFilterInfoRecord::Private
{
public:
    unsigned entries;
};

AutoFilterInfoRecord::AutoFilterInfoRecord(Workbook *book)
    : Record(book), d(new Private)
{
    setEntries(0);
}

AutoFilterInfoRecord::~AutoFilterInfoRecord()
{
    delete d;
}

AutoFilterInfoRecord::AutoFilterInfoRecord( const AutoFilterInfoRecord& record )
    : Record(record), d(new Private)
{
    *this = record;
}

AutoFilterInfoRecord& AutoFilterInfoRecord::operator=( const AutoFilterInfoRecord& record )
{
    *d = *record.d;
    return *this;
}

unsigned AutoFilterInfoRecord::entries() const
{
    return d->entries;
}

void AutoFilterInfoRecord::setEntries(unsigned entries )
{
    d->entries = entries;
}

void AutoFilterInfoRecord::setData( unsigned size, const unsigned char* data, const unsigned int* )
{
    if (size < 2) {
        setIsValid(false);
        return;
    }
    setEntries(readU16(data));
}

void AutoFilterInfoRecord::dump( std::ostream& out ) const
{
    out << "AutoFilterInfo" << std::endl;
    out << "            Entries : " << entries() << std::endl;
}

static Record* createAutoFilterInfoRecord(Workbook *book)
{
    return new AutoFilterInfoRecord(book);
}

void registerRecordClasses()
{
    RecordRegistry::registerRecordClass(BackupRecord::id, createBackupRecord);
    RecordRegistry::registerRecordClass(EOFRecord::id, createEOFRecord);
    RecordRegistry::registerRecordClass(BlankRecord::id, createBlankRecord);
    RecordRegistry::registerRecordClass(BoolErrRecord::id, createBoolErrRecord);
    RecordRegistry::registerRecordClass(StartBlockRecord::id, createStartBlockRecord);
    RecordRegistry::registerRecordClass(EndBlockRecord::id, createEndBlockRecord);
    RecordRegistry::registerRecordClass(ShapePropsStreamRecord::id, createShapePropsStreamRecord);
    RecordRegistry::registerRecordClass(TextPropsStreamRecord::id, createTextPropsStreamRecord);
    RecordRegistry::registerRecordClass(ChartFrtInfoRecord::id, createChartFrtInfoRecord);
    RecordRegistry::registerRecordClass(ChartLayout12ARecord::id, createChartLayout12ARecord);
    RecordRegistry::registerRecordClass(LeftMarginRecord::id, createLeftMarginRecord);
    RecordRegistry::registerRecordClass(RightMarginRecord::id, createRightMarginRecord);
    RecordRegistry::registerRecordClass(TopMarginRecord::id, createTopMarginRecord);
    RecordRegistry::registerRecordClass(BottomMarginRecord::id, createBottomMarginRecord);
    RecordRegistry::registerRecordClass(BoundSheetRecord::id, createBoundSheetRecord);
    RecordRegistry::registerRecordClass(CalcModeRecord::id, createCalcModeRecord);
    RecordRegistry::registerRecordClass(ColInfoRecord::id, createColInfoRecord);
    RecordRegistry::registerRecordClass(DataTableRecord::id, createDataTableRecord);
    RecordRegistry::registerRecordClass(DateModeRecord::id, createDateModeRecord);
    RecordRegistry::registerRecordClass(DimensionRecord::id, createDimensionRecord);
    RecordRegistry::registerRecordClass(ExternSheetRecord::id, createExternSheetRecord);
    RecordRegistry::registerRecordClass(FontRecord::id, createFontRecord);
    RecordRegistry::registerRecordClass(HeaderRecord::id, createHeaderRecord);
    RecordRegistry::registerRecordClass(FooterRecord::id, createFooterRecord);
    RecordRegistry::registerRecordClass(FormatRecord::id, createFormatRecord);
    RecordRegistry::registerRecordClass(LabelRecord::id, createLabelRecord);
    RecordRegistry::registerRecordClass(LabelSSTRecord::id, createLabelSSTRecord);
    RecordRegistry::registerRecordClass(MergedCellsRecord::id, createMergedCellsRecord);
    RecordRegistry::registerRecordClass(MulBlankRecord::id, createMulBlankRecord);
    RecordRegistry::registerRecordClass(NumberRecord::id, createNumberRecord);
    RecordRegistry::registerRecordClass(PaletteRecord::id, createPaletteRecord);
    RecordRegistry::registerRecordClass(RowRecord::id, createRowRecord);
    RecordRegistry::registerRecordClass(StringRecord::id, createStringRecord);
    RecordRegistry::registerRecordClass(NoteRecord::id, createNoteRecord);
    RecordRegistry::registerRecordClass(ProtectRecord::id, createProtectRecord);
    RecordRegistry::registerRecordClass(DefaultRowHeightRecord::id, createDefaultRowHeightRecord);
    RecordRegistry::registerRecordClass(DefaultColWidthRecord::id, createDefaultColWidthRecord);
    RecordRegistry::registerRecordClass(HCenterRecord::id, createHCenterRecord);
    RecordRegistry::registerRecordClass(VCenterRecord::id, createVCenterRecord);
    RecordRegistry::registerRecordClass(SetupRecord::id, createSetupRecord);
    RecordRegistry::registerRecordClass(CrtLineRecord::id, createCrtLineRecord);
    RecordRegistry::registerRecordClass(UnitsRecord::id, createUnitsRecord);
    RecordRegistry::registerRecordClass(PrintSizeRecord::id, createPrintSizeRecord);
    RecordRegistry::registerRecordClass(ChartRecord::id, createChartRecord);
    RecordRegistry::registerRecordClass(SheetExtRecord::id, createSheetExtRecord);
    RecordRegistry::registerRecordClass(BeginRecord::id, createBeginRecord);
    RecordRegistry::registerRecordClass(EndRecord::id, createEndRecord);
    RecordRegistry::registerRecordClass(DataFormatRecord::id, createDataFormatRecord);
    RecordRegistry::registerRecordClass(LineFormatRecord::id, createLineFormatRecord);
    RecordRegistry::registerRecordClass(MarkerFormatRecord::id, createMarkerFormatRecord);
    RecordRegistry::registerRecordClass(AreaFormatRecord::id, createAreaFormatRecord);
    RecordRegistry::registerRecordClass(PieRecord::id, createPieRecord);
    RecordRegistry::registerRecordClass(PieFormatRecord::id, createPieFormatRecord);
    RecordRegistry::registerRecordClass(AttachedLabelRecord::id, createAttachedLabelRecord);
    RecordRegistry::registerRecordClass(ChartFormatRecord::id, createChartFormatRecord);
    RecordRegistry::registerRecordClass(LegendRecord::id, createLegendRecord);
    RecordRegistry::registerRecordClass(SeriesListRecord::id, createSeriesListRecord);
    RecordRegistry::registerRecordClass(BarRecord::id, createBarRecord);
    RecordRegistry::registerRecordClass(LineRecord::id, createLineRecord);
    RecordRegistry::registerRecordClass(AreaRecord::id, createAreaRecord);
    RecordRegistry::registerRecordClass(ScatterRecord::id, createScatterRecord);
    RecordRegistry::registerRecordClass(AxisRecord::id, createAxisRecord);
    RecordRegistry::registerRecordClass(TickRecord::id, createTickRecord);
    RecordRegistry::registerRecordClass(ValueRangeRecord::id, createValueRangeRecord);
    RecordRegistry::registerRecordClass(CatSerRangeRecord::id, createCatSerRangeRecord);
    RecordRegistry::registerRecordClass(AxisLineRecord::id, createAxisLineRecord);
    RecordRegistry::registerRecordClass(CrtLinkRecord::id, createCrtLinkRecord);
    RecordRegistry::registerRecordClass(DefaultTextRecord::id, createDefaultTextRecord);
    RecordRegistry::registerRecordClass(TextRecord::id, createTextRecord);
    RecordRegistry::registerRecordClass(FontXRecord::id, createFontXRecord);
    RecordRegistry::registerRecordClass(ObjectLinkRecord::id, createObjectLinkRecord);
    RecordRegistry::registerRecordClass(PlotAreaRecord::id, createPlotAreaRecord);
    RecordRegistry::registerRecordClass(MsoDrawingSelectionRecord::id, createMsoDrawingSelectionRecord);
    RecordRegistry::registerRecordClass(Chart3dRecord::id, createChart3dRecord);
    RecordRegistry::registerRecordClass(PicFRecord::id, createPicFRecord);
    RecordRegistry::registerRecordClass(DropBarRecord::id, createDropBarRecord);
    RecordRegistry::registerRecordClass(RadarRecord::id, createRadarRecord);
    RecordRegistry::registerRecordClass(SurfRecord::id, createSurfRecord);
    RecordRegistry::registerRecordClass(RadarAreaRecord::id, createRadarAreaRecord);
    RecordRegistry::registerRecordClass(AxisParentRecord::id, createAxisParentRecord);
    RecordRegistry::registerRecordClass(LegendExceptionRecord::id, createLegendExceptionRecord);
    RecordRegistry::registerRecordClass(ShtPropsRecord::id, createShtPropsRecord);
    RecordRegistry::registerRecordClass(SerToCrtRecord::id, createSerToCrtRecord);
    RecordRegistry::registerRecordClass(AxesUsedRecord::id, createAxesUsedRecord);
    RecordRegistry::registerRecordClass(SBaseRefRecord::id, createSBaseRefRecord);
    RecordRegistry::registerRecordClass(SerParentRecord::id, createSerParentRecord);
    RecordRegistry::registerRecordClass(SerAuxTrendRecord::id, createSerAuxTrendRecord);
    RecordRegistry::registerRecordClass(IFmtRecordRecord::id, createIFmtRecordRecord);
    RecordRegistry::registerRecordClass(PosRecord::id, createPosRecord);
    RecordRegistry::registerRecordClass(AIRunsRecord::id, createAIRunsRecord);
    RecordRegistry::registerRecordClass(SerAuxErrBarRecord::id, createSerAuxErrBarRecord);
    RecordRegistry::registerRecordClass(ClrtClientRecord::id, createClrtClientRecord);
    RecordRegistry::registerRecordClass(SerFmtRecord::id, createSerFmtRecord);
    RecordRegistry::registerRecordClass(Chart3DBarShapeRecord::id, createChart3DBarShapeRecord);
    RecordRegistry::registerRecordClass(FbiRecord::id, createFbiRecord);
    RecordRegistry::registerRecordClass(BopPopRecord::id, createBopPopRecord);
    RecordRegistry::registerRecordClass(AxcExtRecord::id, createAxcExtRecord);
    RecordRegistry::registerRecordClass(DatRecord::id, createDatRecord);
    RecordRegistry::registerRecordClass(PlotGrowthRecord::id, createPlotGrowthRecord);
    RecordRegistry::registerRecordClass(SIIndexRecord::id, createSIIndexRecord);
    RecordRegistry::registerRecordClass(GelFrameRecord::id, createGelFrameRecord);
    RecordRegistry::registerRecordClass(BobPopCustomRecord::id, createBobPopCustomRecord);
    RecordRegistry::registerRecordClass(Fbi2Record::id, createFbi2Record);
    RecordRegistry::registerRecordClass(ZoomLevelRecord::id, createZoomLevelRecord);
    RecordRegistry::registerRecordClass(FrameRecord::id, createFrameRecord);
    RecordRegistry::registerRecordClass(SeriesRecord::id, createSeriesRecord);
    RecordRegistry::registerRecordClass(SeriesTextRecord::id, createSeriesTextRecord);
    RecordRegistry::registerRecordClass(GridSetReservedRecord::id, createGridSetReservedRecord);
    RecordRegistry::registerRecordClass(DSFReservedRecord::id, createDSFReservedRecord);
    RecordRegistry::registerRecordClass(Excel9FileRecord::id, createExcel9FileRecord);
    RecordRegistry::registerRecordClass(MmsReservedRecord::id, createMmsReservedRecord);
    RecordRegistry::registerRecordClass(Compat12Record::id, createCompat12Record);
    RecordRegistry::registerRecordClass(Window1Record::id, createWindow1Record);
    RecordRegistry::registerRecordClass(Window2Record::id, createWindow2Record);
    RecordRegistry::registerRecordClass(ObjProtectedRecord::id, createObjProtectedRecord);
    RecordRegistry::registerRecordClass(ScenarioProtectRecord::id, createScenarioProtectRecord);
    RecordRegistry::registerRecordClass(WinProtectRecord::id, createWinProtectRecord);
    RecordRegistry::registerRecordClass(Prot4RevRecord::id, createProt4RevRecord);
    RecordRegistry::registerRecordClass(Prot4RevPassRecord::id, createProt4RevPassRecord);
    RecordRegistry::registerRecordClass(PasswordRecord::id, createPasswordRecord);
    RecordRegistry::registerRecordClass(FeatHdrRecord::id, createFeatHdrRecord);
    RecordRegistry::registerRecordClass(IndexRecord::id, createIndexRecord);
    RecordRegistry::registerRecordClass(GutsRecord::id, createGutsRecord);
    RecordRegistry::registerRecordClass(WsBoolRecord::id, createWsBoolRecord);
    RecordRegistry::registerRecordClass(HeaderFooterRecord::id, createHeaderFooterRecord);
    RecordRegistry::registerRecordClass(PLVRecord::id, createPLVRecord);
    RecordRegistry::registerRecordClass(CalcSaveRecalcRecord::id, createCalcSaveRecalcRecord);
    RecordRegistry::registerRecordClass(InterfaceHdrRecord::id, createInterfaceHdrRecord);
    RecordRegistry::registerRecordClass(InterfaceEndRecord::id, createInterfaceEndRecord);
    RecordRegistry::registerRecordClass(RRTabIdRecord::id, createRRTabIdRecord);
    RecordRegistry::registerRecordClass(BuiltInFnGroupCountRecord::id, createBuiltInFnGroupCountRecord);
    RecordRegistry::registerRecordClass(HideObjRecord::id, createHideObjRecord);
    RecordRegistry::registerRecordClass(RefreshAllRecord::id, createRefreshAllRecord);
    RecordRegistry::registerRecordClass(RecalcIdRecord::id, createRecalcIdRecord);
    RecordRegistry::registerRecordClass(BookBoolRecord::id, createBookBoolRecord);
    RecordRegistry::registerRecordClass(StyleRecord::id, createStyleRecord);
    RecordRegistry::registerRecordClass(StyleExtRecord::id, createStyleExtRecord);
    RecordRegistry::registerRecordClass(ExtSSTRecord::id, createExtSSTRecord);
    RecordRegistry::registerRecordClass(UsesELFsRecord::id, createUsesELFsRecord);
    RecordRegistry::registerRecordClass(DBCellRecord::id, createDBCellRecord);
    RecordRegistry::registerRecordClass(PhoneticInfoRecord::id, createPhoneticInfoRecord);
    RecordRegistry::registerRecordClass(CalcCountRecord::id, createCalcCountRecord);
    RecordRegistry::registerRecordClass(CalcPrecisionRecord::id, createCalcPrecisionRecord);
    RecordRegistry::registerRecordClass(CalcDeltaRecord::id, createCalcDeltaRecord);
    RecordRegistry::registerRecordClass(CalcIterRecord::id, createCalcIterRecord);
    RecordRegistry::registerRecordClass(CalcRefModeRecord::id, createCalcRefModeRecord);
    RecordRegistry::registerRecordClass(CodePageRecord::id, createCodePageRecord);
    RecordRegistry::registerRecordClass(CountryRecord::id, createCountryRecord);
    RecordRegistry::registerRecordClass(SelectionRecord::id, createSelectionRecord);
    RecordRegistry::registerRecordClass(PrinterSettingsRecord::id, createPrinterSettingsRecord);
    RecordRegistry::registerRecordClass(PrintRowColRecord::id, createPrintRowColRecord);
    RecordRegistry::registerRecordClass(PrintGridRecord::id, createPrintGridRecord);
    RecordRegistry::registerRecordClass(LastWriteAccessRecord::id, createLastWriteAccessRecord);
    RecordRegistry::registerRecordClass(XFExtRecord::id, createXFExtRecord);
    RecordRegistry::registerRecordClass(XFCRCRecord::id, createXFCRCRecord);
    RecordRegistry::registerRecordClass(TableStylesRecord::id, createTableStylesRecord);
    RecordRegistry::registerRecordClass(MTRSettingsRecord::id, createMTRSettingsRecord);
    RecordRegistry::registerRecordClass(ForceFullCalculationRecord::id, createForceFullCalculationRecord);
    RecordRegistry::registerRecordClass(BookExtRecord::id, createBookExtRecord);
    RecordRegistry::registerRecordClass(ThemeRecord::id, createThemeRecord);
    RecordRegistry::registerRecordClass(CompressPicturesRecord::id, createCompressPicturesRecord);
    RecordRegistry::registerRecordClass(AutoFilterRecord::id, createAutoFilterRecord);
    RecordRegistry::registerRecordClass(AutoFilter12Record::id, createAutoFilter12Record);
    RecordRegistry::registerRecordClass(AutoFilterInfoRecord::id, createAutoFilterInfoRecord);
}

} // namespace Swinder
